/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************

NAME            MPP_LinuxUpper.c
SUMMARY         %description%
VERSION         %version: 47 %
UPDATE DATE     %date_modified: Tue Feb 03 10:49:23 2004 %
PROGRAMMER      %created_by:    yqi %

        Copyright 2003 by LSI Logic Inc.

DESCRIPTION: This is the main file that contains all the functionality of the linux Upper
Level driver. 

INCLUDE FILES:

NOTES: MPP upper level driver uses error number 700-900

RESTRICTIONS:

SEE ALSO:

REFERENCE: 

IMPLEMENTATION:

MODIFICATION HISTORY:

*******************************************************************************/


/***  MACRO DEFINITIONS  ***/
#define		WWN_LENGTH 16

#define __SRCMPP_Target

/***  INCLUDES  ***/  


#include "MPP_Sysdep.h"
#include "MPP_Common.h"
#include "MPP_RdacInfo.h"
#include "MPP_ProtoTypes.h"
#include "copyright.h"
#include "mppLnx_shared.h"
#include "mppLnx_prototypes.h"
#include <linux/list.h>

/*** LOCALS  ***/

/* The following function pointers stores the old function pointers for sd/sg driver */
static sdtpntFuncPtr sd_old_detect=NULL,sd_old_attach=NULL, sg_old_detect=NULL, sg_old_attach=NULL;
/* variable used to mark if the initial discovery scan has been finished */
static int initialDiscoveryFinished = 0;
/*
* The /proc/mpp/array/controller/path directory needs the physical HBA driver's identification.
* But at the time of building mpp_CreateControllerVertex, the physical device is not in 
* RdacDeviceInformation struct yet. The following two variables are used to set 
* mppLnx_host_procName that are used in the MPP_Sysdep.c function.
*/
static int              is_mppLnx_host_procName_set = 0;
extern char             * mppLnx_host_procName;  

/*  MPP Scsi_Device_Template structure declaration */
static struct Scsi_Device_Template mpp_template;
/* MPP configuration parameter structure declaration */
static struct mpp_Config_Parameters mppLnx_Config_Parameters;
/* MPP proc entry declaration */
extern struct proc_dir_entry *proc_mpp;
static struct proc_dir_entry *proc_vbus = NULL;
/* file operation structure used for module initialization */
static struct file_operations mppLnx_fops = {
    .owner = THIS_MODULE,
    .read = NULL,
    .write = NULL,
    .open = mppLnx_open,
    .ioctl = mppLnx_ioctl,
    .release = mppLnx_release
};

extern wait_queue_head_t	mppLnx_wait_queue; /* Wait queue for thread delay function */

static spinlock_t unconfiguedDevice_lock = SPIN_LOCK_UNLOCKED;
static mppLnx_UnconfiguredDeviceList_t  unconfiguredDeviceList;


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_init
* SUMMARY:  MPP upper level driver initialization routine
* SCOPE:    local
*
* DESCRIPTION: This function is called when loading MPP upper level driver. This function will:
*              1. register with Scsi middle layer  as an upper level driver
*              2. register with Linux kernelas a character device using dynamically allocated major number
*              3. read configurable parameter file, persistent binding file and translation key.
*              4. replace sd/sg driver function pointers
*              5. create virtual bus node
*              6. allocate mpp_ModuleArray memory resources
*
* RESTRICTIONS:
* 
* RETURNS: 0 if sucess, 1 if error
*
* ERRNO:
*
*
*/

static int mppLnx_init (void)
{
	static int mpp_major_no;
	int scsiRegisterStatus = 0;
	struct Scsi_Device_Template* SDTpnt = scsi_devicelist;
	BOOL sdExists = FALSE, sgExists = FALSE;

	/* assign default values to MPP driver configurable parameters */
        mppLnx_Config_Parameters.mpp_NotReadyWaitTime = mpp_NotReadyWaitTime;
        mppLnx_Config_Parameters.mpp_BusyWaitTime =     mpp_BusyWaitTime;
        mppLnx_Config_Parameters.mpp_QuiescenceWaitTime =       mpp_QuiescenceWaitTime;
        mppLnx_Config_Parameters.mpp_InquiryWaitTime =  mpp_InquiryWaitTime;
        mppLnx_Config_Parameters.mpp_MaxLunsPerArray =  mpp_MaxLunsPerArray;
        mppLnx_Config_Parameters.mpp_MaxPathsPerController =    mpp_MaxPathsPerController = 4;
        mppLnx_Config_Parameters.mpp_ScanInterval =     mpp_ScanInterval;
        mppLnx_Config_Parameters.mpp_InquiryInterval =  mpp_InquiryInterval;
        mppLnx_Config_Parameters.mpp_MaxArrayModules =  mpp_MaxArrayModules;
        mppLnx_Config_Parameters.mpp_ErrorLevel =       mpp_ErrorLevel;
        mppLnx_Config_Parameters.mpp_Debug =       mpp_Debug;
        mppLnx_Config_Parameters.mpp_SelectionTimeoutRetryCount =       mpp_SelectionTimeoutRetryCount;
        mppLnx_Config_Parameters.mpp_CommandTimeoutRetryCount =         mpp_CommandTimeoutRetryCount;
        mppLnx_Config_Parameters.mpp_UaRetryCount =     mpp_UaRetryCount;
        mppLnx_Config_Parameters.mpp_SyncRetryCount =   mpp_SyncRetryCount;
        mppLnx_Config_Parameters.mpp_SynchTimeout =     mpp_SynchTimeout = 60;
        mppLnx_Config_Parameters.mpp_FailOverQuiescenceTime =   mpp_FailOverQuiescenceTime;
        mppLnx_Config_Parameters.mpp_FailoverTimeout =  mpp_FailoverTimeout;
        mppLnx_Config_Parameters.mpp_FailBackToCurrentAllowed =         mpp_FailBackToCurrentAllowed;
        mppLnx_Config_Parameters.mpp_HoldAltInReset =   mpp_HoldAltInReset;
        mppLnx_Config_Parameters.mpp_DoUARetry =        mpp_DoUARetry = 1;
        mppLnx_Config_Parameters.mpp_ControllerIoWaitTime = mpp_ControllerIoWaitTime;
        mppLnx_Config_Parameters.mpp_ArrayIoWaitTime = mpp_ArrayIoWaitTime;
        mppLnx_Config_Parameters.mpp_DisableLUNRebalance = mpp_DisableLUNRebalance;

	// assigned default values to parameters of type char array
        strcpy (mppLnx_Config_Parameters.mpp_VendorId, "LSI     \0");
        strcpy (mppLnx_Config_Parameters.mpp_ProductId,"VirtualDisk     \0");
        strcpy (mpp_VendorId, "LSI     \0");
        strcpy (mpp_ProductId,"VirtualDisk     \0");
	strcpy (mppLnx_Config_Parameters.mpp_Version,RAID_MANAGER_VERSION);

	// reset mpp_S2ToS3Key memory
	memset (mpp_S2ToS3Key, 0, 8);
	memset (mppLnx_Config_Parameters.mpp_S2ToS3Key, 0, 8);

	// read configurable parameter file
	mppLnx_readConfigData();

	// allocate and reset mpp_ModuleArray[] data structure
	mpp_ModuleArray = (ArrayModuleEntry_t*) MPP_KMEM_ALLOC(sizeof(ArrayModuleEntry_t)*mpp_MaxArrayModules);
	if (mpp_ModuleArray == NULL)
	{
		MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 700, 
			"Failed Allocating mpp_ModuleArray[]\n"));
		goto errorOut;
	}

	MPP_DEBUGPRINT((MPP_INIT_DEBUG, "mpp_ModuleArray allocated starting at address 0x%p\n", mpp_ModuleArray));

	// read persistent binding data
	mppLnx_readBindingData();

	// Now register with the kernel as a char device using dynamically allocated major number.
	mpp_major_no = register_chrdev (0, "mppUpper", &mppLnx_fops);
	if (mpp_major_no < 0) {
		MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 701, 
			"Failed registering mppLnx, returned %d\n", mpp_major_no));
		goto errorOut;
	}
	MPP_DEBUGPRINT((MPP_INIT_DEBUG, "device successfully   registered, Major No. = %d\n", mpp_major_no));

	/*** Now assign proper values to the fields in mpp_template data structure ***/
	mpp_template.name= "mppUpper";
	mpp_template.tag= "mpp";
	mpp_template.nr_dev = mpp_template.dev_noticed = 0;
	mpp_template.dev_max = mppLnx_Config_Parameters.mpp_MaxLunsPerArray 
	* mppLnx_Config_Parameters.mpp_MaxPathsPerController *2 * 
	mppLnx_Config_Parameters.mpp_MaxArrayModules;
	mpp_template.module = THIS_MODULE;
	mpp_template.scsi_type=TYPE_DISK;
	mpp_template.major= mpp_major_no;
	mpp_template.blk=0;
	mpp_template.detect= mppLnx_target_detect;
	mpp_template.init= mppLnx_target_init;
	mpp_template.finish= mppLnx_target_finish;
	mpp_template.attach= mppLnx_target_attach;
	mpp_template.detach= mppLnx_target_detach;
	mpp_template.init_command= mppLnx_target_init_command;
	
	// replace detect() and attach() functions for sd and sg drivers
	while (SDTpnt != NULL) {
		if (strcmp("sd", SDTpnt->tag) == 0)
		{
			sd_old_detect = SDTpnt->detect;
			sd_old_attach = SDTpnt->attach;
			sdExists = TRUE;
		}
		if (strcmp("sg", SDTpnt->tag) == 0)
		{
			sg_old_detect = SDTpnt->detect;
			sg_old_attach = SDTpnt->attach;
			sgExists = TRUE;
		}
		if (sdExists && sgExists)
			break;
		SDTpnt = SDTpnt->next;
	}
	if (!sdExists || !sgExists)
	{
		MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 702, 
			"sd driver or sg driver is not present on the system\n" ));
		goto errorOut;
	}
	if (mppLnx_replaceFuncPtr("sd",mppLnx_sd_detect, mppLnx_sd_attach) &&
	mppLnx_replaceFuncPtr("sg",mppLnx_sg_detect, mppLnx_sg_attach) ) 		{
	 	MPP_DEBUGPRINT((MPP_INIT_DEBUG, "Sd and Sg interception finished\n"));
	}
	else
	{
		MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 703, 
			"Failed to intercept sd/sg function\n" ));
		goto errorOut;
	}

	/* create the /proc/mpp directory */
	proc_mpp = proc_mkdir(MPPLNX_PROC_NAME, 0);
	if (NULL == proc_mpp)
	{
		MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 704, 
			"Failed to create MPP driver proc entry %s\n",MPPLNX_PROC_NAME));
	}
	else
	{
		proc_mpp->owner = THIS_MODULE;

		/* create proc entry for virtual bus node */		
		proc_vbus = proc_mknod(MPP_VBUSNODE, 0777+S_IFCHR, proc_mpp, (kdev_t)MKDEV(mpp_template.major, 0)); 
		if (NULL == proc_vbus)
		{
			MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 705, 
				"Failed to create virtual bus node proc entry %s\n",MPP_VBUSNODE));
		}
		else
		{
			proc_vbus->owner = THIS_MODULE;
		}
	}

   /*
    * Initialize wait queue 
    */
    init_waitqueue_head( &mppLnx_wait_queue );

   /*
   * initialize the unconfigured device list and its lock
   */
   unconfiguredDeviceList.unconfiguredDeviceLock = &unconfiguedDevice_lock;
   unconfiguredDeviceList.deviceEntryList = 
      (mppLnx_UnconfiguredDeviceEntry_t *) MPP_INT_KMEM_ALLOC(sizeof(mppLnx_UnconfiguredDeviceEntry_t));
   if(!unconfiguredDeviceList.deviceEntryList)
   {
      /* log error*/
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 808,
            "Cannot allocate memory for unconfigured device entry\n"));
      goto errorOut;
   }

   INIT_LIST_HEAD(&(unconfiguredDeviceList.deviceEntryList->list));
     
	/*** register with Scsi middle layer as an upper level driver and return status back ***/
	scsiRegisterStatus =  scsi_register_module(MODULE_SCSI_DEV,&mpp_template);
	if (scsiRegisterStatus)
		goto errorOut;
	else
		return 0;
errorOut:
	/* roll back the resources that we allocated if any error occurs */
	if (mpp_ModuleArray != NULL)
	{
		/* Free the allocated memory for mpp_ModuleArray[] data structure */
		MPP_FREE(mpp_ModuleArray, sizeof(ArrayModuleEntry_t)*mppLnx_Config_Parameters.mpp_MaxArrayModules);
	}
	if (mpp_template.major != 0)
	{
		/* unregister with the kernel */
		unregister_chrdev (mpp_template.major, "mppUpper");		
	}
	
	if (proc_vbus != NULL)
	{
		/* remove virtual bus node */
		remove_proc_entry(MPP_VBUSNODE, proc_mpp);
	}
	if (proc_mpp != NULL)
	{
		/* remove MPP proc entry */
		remove_proc_entry(MPPLNX_PROC_NAME, 0);
	}
	/* restore function pointers */
	mppLnx_replaceFuncPtr("sd",sd_old_detect, sd_old_attach);
	mppLnx_replaceFuncPtr("sg",sg_old_detect, sg_old_attach);
	return 1;

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_exit
* SUMMARY:  MPP upper level driver clean up routine
* SCOPE:    local
*
* DESCRIPTION: This function is called when unloading MPP upper level driver. This function will:
*              1. unregister with Scsi middle layer and Linux kernel
*              2. Restore sd/sg driver function pointers
*              3. Remove virtual bus node
*              4. Free mpp_ModuleArray memory resources
*
* RESTRICTIONS:
* 
* RETURNS: none
*
* ERRNO:
*
*
*/
static void mppLnx_exit (void)
{
	int                              kernelUnregisterStatus = 0;
	int                              scsiUnregisterStatus = 0;
        mppLnx_UnconfiguredDeviceEntry_t * deviceEntry = NULL;
        struct list_head                 *listp;
        struct list_head                 *temp;

	scsiUnregisterStatus = scsi_unregister_module(MODULE_SCSI_DEV, &mpp_template) ;
	if (scsiUnregisterStatus != 0)
	{
		MPP_ERRORLOGPRINT ((MPP_ERR_FATAL, 706, 
			"Error status %d returned while unregistering module with Scsi middle layer\n", scsiUnregisterStatus));
	}
		
	mpp_template.dev_max=0;

	MPP_DEBUGPRINT((MPP_DETACH_DEBUG, "unregistered with scsi middle layer\n"));

	// restore detect() and attach() functions for sd and sg drivers
	if (mppLnx_replaceFuncPtr("sd",sd_old_detect, sd_old_attach) &&
	mppLnx_replaceFuncPtr("sg",sg_old_detect, sg_old_attach) ) 		{
	 	MPP_DEBUGPRINT((MPP_INIT_DEBUG, "Sd and Sg restore finished\n"));
	}
	else
	{
		MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 707, 
			"Failed to restore sd/sg function\n" ));
	}
	/* restore intercepted function pointers */
	mppLnx_replaceFuncPtr("sd",sd_old_detect, sd_old_attach);
	mppLnx_replaceFuncPtr("sg",sg_old_detect, sg_old_attach);

	kernelUnregisterStatus = unregister_chrdev (mpp_template.major, "mppUpper");
	if (kernelUnregisterStatus != 0)
	{
		MPP_ERRORLOGPRINT ((MPP_ERR_FATAL, 708, 
			"Error status %d returned while unregistering with kernel\n", kernelUnregisterStatus));
	}
	MPP_DEBUGPRINT((MPP_DETACH_DEBUG, "unregistered with Linux kernel\n"));

	/*** Free the allocated memory for mpp_ModuleArray[] data structure ***/
	MPP_FREE(mpp_ModuleArray, sizeof(ArrayModuleEntry_t)*mppLnx_Config_Parameters.mpp_MaxArrayModules);

	/*** Remove MPP driver virtual bus node ***/
	if (proc_vbus != NULL)
	{
		remove_proc_entry(MPP_VBUSNODE, proc_mpp);
		MPP_DEBUGPRINT((MPP_DETACH_DEBUG, "virtual bus node %s removed\n",
			MPP_VBUSNODE));
	}
	
	/*** Remove proc entry directory for MPP driver ***/
	if (proc_mpp != NULL)
	{
		remove_proc_entry(MPPLNX_PROC_NAME, 0);
		MPP_DEBUGPRINT((MPP_DETACH_DEBUG, "MPP proc entry /proc/%s removed\n",
			MPPLNX_PROC_NAME));
	}

   /*
   * clean-up the un-configured device list
   */

   list_for_each_safe( listp, temp, &(unconfiguredDeviceList.deviceEntryList->list))
   {
      deviceEntry = list_entry(listp, mppLnx_UnconfiguredDeviceEntry_t, list);
      mppLnx_deleteUnconfiugredDevice(deviceEntry->physical_device);
      
   }
   MPP_FREE(unconfiguredDeviceList.deviceEntryList,sizeof(mppLnx_UnconfiguredDeviceEntry_t));

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_replaceFuncPtr
* SUMMARY:  replace s function pointer in Scsi_Device_Template with a new
*           fucntion pointer
* SCOPE:    local
*
* DESCRIPTION: This function is called by mppLnx_init() and mppLnx_exit to replace function pointers in sd and sg drivers
		@1: string containing driver name
		@2: The function pointer of detect function in Scsi_Device_Template
		@3: The function pointer of attach function in Scsi_Device_Template
*
* RETURNS: 1 on success, 0 if the driver is not present or if any of the replacement fucntion pointer is NULL
*/
static int mppLnx_replaceFuncPtr 
   (char* driverName, sdtpntFuncPtr detectFuncPtr, sdtpntFuncPtr attachFuncPtr)
{
	struct Scsi_Device_Template* SDTpnt = scsi_devicelist;
	BOOL replaceFinished = FALSE;
	if ( detectFuncPtr == NULL && attachFuncPtr == NULL)
		return 0;
	while (SDTpnt != NULL) {
		if (strcmp(SDTpnt->tag ,driverName) == 0) {
			SDTpnt->detect = detectFuncPtr;
			SDTpnt->attach = attachFuncPtr;
			replaceFinished = TRUE;
			break;
		}
		SDTpnt=SDTpnt->next;
	}
	if (replaceFinished)
		return 1;
	else
		return 0;
}
	
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_readConfigData
* SUMMARY:  read configurable parameter file
* SCOPE:    local
*
*
* DESCRIPTION: This function is called in MPP upper level driver initialization routine to read configurable parameter values.
*              The file has the following format:
*              <ParameterName1>=<ParameterValue1>'\n'
*              <ParameterName1>=<ParameterValue1>'\n'
*              ...
*              <ParameterName1>=<ParameterValue1>'\n'
*	       Leading or trailing spaces are allowed around parameter name and parameter value for more flexibility.
*	       User can insert blank lines.
*	       User can have comment lines as long as they do not have '='
*
* RESTRICTIONS:
* 
* RETURNS: 0 if suceeds, 1 if error
*
* ERRNO: 1
*
* NOTES: if error occurs, the parameter values are set back to their default values.
*
*/
static int mppLnx_readConfigData(void)
{
	struct file* filp;
	char* buf = NULL, *bufBeginPtr = NULL;  
	char* tag = NULL;
	char* value = NULL;
	LWORD bytesRead = 0;
	LWORD fileSize = 0;

	/*** Now open the file /etc/mpp.conf to read ***/
	filp = filp_open(CONFIG_FILE_NAME,O_RDONLY,0);
	if(!filp || IS_ERR(filp)) {
		MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 709, 
			"mppLnx_readConfigData: Error opening /etc/mpp.conf\n"));
		return 1;
	}
	fileSize = filp->f_dentry->d_inode->i_size;
	/*** allocate the buffer to hold file contents ***/
	buf = MPP_KMEM_ALLOC(fileSize);
	if(!buf) {
		MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 710, 
			"mppLnx_readConfigData: Error allocating buffer of size %d\n",fileSize));
		filp_close(filp,0);
		return 1;
	}
	bufBeginPtr = buf;

	/*** read the whole contents of the file into the buffer ***/	
	bytesRead = kernel_read(filp,0,buf,fileSize);
	if(bytesRead != fileSize) {
		MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 711, 
			"mppLnx_readConfigData: Only get %d bytes out of %d bytes.\n", 
			bytesRead, fileSize));
		filp_close(filp,0);
		kfree(buf);
		return 1;
	}
	/*** close file handler ***/
	filp_close(filp,0);

 	/*** Now parse the buffer ***/
	while (buf < bufBeginPtr+fileSize && buf != NULL && *buf != 0)
	{
		mppLnx_parseLine(&buf, "=", &tag, &value);	
		mppLnx_modifyConfigParameter(tag, value);
		tag = value = NULL;
	}
	/*** Now free the buffer and other memory resources that we allocated ***/
	MPP_FREE(bufBeginPtr,fileSize);
	return 0;

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_modifyConfigParameter
* SUMMARY:  Modify the value of MPP driver configurable parameters according to /etc/mpp.conf
* SCOPE:    local
*
*
* DESCRIPTION: Modify the value of MPP driver configurable parameters according to /etc/mpp.conf
*	       @1: string containing parameter name
*	       @2: string containing parameter value
*
* RESTRICTIONS:
* 
* RETURNS: void
*
* NOTES: if error occurs, the parameter values are set back to their default values.
*
*/
static void mppLnx_modifyConfigParameter(char* tag, char* value)
{

	LWORD index = 0;
	BOOL S2ToS3KeyInvalid = FALSE; 
	char* endOfString;
	unsigned long parseValue = 0;
	if (tag == NULL || value == NULL || *tag == 0 || *value == 0)
		return;
	/*** now parse the parameter value and assign it to proper parameters ***/
	if(strncmp(tag,MPP_DEBUG_LEVEL,strlen(MPP_DEBUG_LEVEL))== 0) 
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_Debug = mppLnx_Config_Parameters.mpp_Debug = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", MPP_DEBUG_LEVEL, 
			mpp_Debug));
		}
	}
	else if(strncmp(tag,MPP_NOTREADYWAITTIME, strlen(MPP_NOTREADYWAITTIME))== 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_NotReadyWaitTime = mppLnx_Config_Parameters.mpp_NotReadyWaitTime = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", MPP_NOTREADYWAITTIME,
			mpp_NotReadyWaitTime));
		}
	}
	else if(strncmp(tag,MPP_BUSYWAITTIME, strlen(MPP_BUSYWAITTIME)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_BusyWaitTime = mppLnx_Config_Parameters.mpp_BusyWaitTime = parseValue; 
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", MPP_BUSYWAITTIME,
				mpp_BusyWaitTime));
		}
	}
	else if(strncmp(tag,MPP_QUIESCENCEWAITTIME,strlen(MPP_QUIESCENCEWAITTIME)) == 0) 
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_QuiescenceWaitTime = mppLnx_Config_Parameters.mpp_QuiescenceWaitTime = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", MPP_QUIESCENCEWAITTIME,
				mpp_QuiescenceWaitTime));
		}
	}
	else if (strncmp(tag,MPP_INQUIRYWAITTIME,strlen(MPP_INQUIRYWAITTIME)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_InquiryWaitTime = mppLnx_Config_Parameters.mpp_InquiryWaitTime = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", MPP_INQUIRYWAITTIME,
				mpp_InquiryWaitTime));
		}
	}
	else if (strncmp(tag,MPP_MAXLUNSPERARRAY,strlen(MPP_MAXLUNSPERARRAY)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_MaxLunsPerArray = mppLnx_Config_Parameters.mpp_MaxLunsPerArray =  simple_strtoul(value,&endOfString,0);
			if(	(mpp_MaxLunsPerArray > 256)	|| (mpp_MaxLunsPerArray	== 0)) {
				MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 712, 
					"The mpp driver	configuration parameter	%s(%d) is incorrect\n",
				MPP_MAXLUNSPERARRAY, mpp_MaxLunsPerArray));
				mpp_MaxLunsPerArray = mppLnx_Config_Parameters.mpp_MaxLunsPerArray = 32;
			}
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", MPP_MAXLUNSPERARRAY,
				mpp_MaxLunsPerArray));
		}
	}
	else if (strncmp(tag,MPP_MAXPATHSPERCONTROLLER,strlen(MPP_MAXPATHSPERCONTROLLER)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_MaxPathsPerController = mppLnx_Config_Parameters.mpp_MaxPathsPerController = parseValue;
			if(	(mpp_MaxPathsPerController > 255) || (mpp_MaxPathsPerController	== 0) )	{
				MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 713, 
					"The mpp driver	configuration parameter	%s(%d) is incorrect\n",
					MPP_MAXPATHSPERCONTROLLER, mpp_MaxPathsPerController));
				mpp_MaxPathsPerController = mppLnx_Config_Parameters.mpp_MaxPathsPerController = 4;
			}
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", 
				MPP_MAXPATHSPERCONTROLLER, mpp_MaxPathsPerController));
		}
	}
	else if (strncmp(tag,MPP_SCANINTERVAL,strlen(MPP_SCANINTERVAL)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_ScanInterval = mppLnx_Config_Parameters.mpp_ScanInterval = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", MPP_SCANINTERVAL,
				mpp_ScanInterval));
		}
	}
	else if (strncmp(tag,MPP_INQUIRYINTERVAL,strlen(MPP_INQUIRYINTERVAL)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_InquiryInterval = mppLnx_Config_Parameters.mpp_InquiryInterval = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", MPP_INQUIRYINTERVAL,
				mpp_InquiryInterval));
		}
	}
	else if (strncmp(tag,MPP_MAXARRAYMODULES,strlen(MPP_MAXARRAYMODULES)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_MaxArrayModules = mppLnx_Config_Parameters.mpp_MaxArrayModules = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", "MaxArrayModules",
				mpp_MaxArrayModules));
		}
	}
	else if (strncmp(tag,MPP_ERRORLEVEL,strlen(MPP_ERRORLEVEL)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_ErrorLevel = mppLnx_Config_Parameters.mpp_ErrorLevel = parseValue;
			if(	mpp_ErrorLevel <= 5	) {
				MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	0x%x\n", MPP_ERRORLEVEL,
					mpp_ErrorLevel));
			} else {
				mpp_ErrorLevel = mppLnx_Config_Parameters.mpp_ErrorLevel = 0;
				MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 715, 
				"The mpp driver	configuration parameter	mpp_ErrorLevel is invalid\n	"));
			}
		}
	}
	else if (strncmp(tag,MPP_SELECTIONTIMEOUTRETRYCOUNT,strlen(MPP_SELECTIONTIMEOUTRETRYCOUNT)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_SelectionTimeoutRetryCount = mppLnx_Config_Parameters.mpp_SelectionTimeoutRetryCount = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", 
				MPP_SELECTIONTIMEOUTRETRYCOUNT,	mpp_SelectionTimeoutRetryCount));
		}
	}
	else if (strncmp(tag,MPP_COMMANDTIMEOUTRETRYCOUNT,strlen(MPP_COMMANDTIMEOUTRETRYCOUNT)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_CommandTimeoutRetryCount = mppLnx_Config_Parameters.mpp_CommandTimeoutRetryCount = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", 
				MPP_COMMANDTIMEOUTRETRYCOUNT, mpp_CommandTimeoutRetryCount));
		}
	}
	else if (strncmp(tag,MPP_UARETRYCOUNT,strlen(MPP_UARETRYCOUNT)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_UaRetryCount = mppLnx_Config_Parameters.mpp_UaRetryCount = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", MPP_UARETRYCOUNT,
				mpp_UaRetryCount));
		}
	}
	else if (strncmp(tag,MPP_RETRYCOUNT,strlen(MPP_RETRYCOUNT)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_SyncRetryCount = mppLnx_Config_Parameters.mpp_SyncRetryCount = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", MPP_RETRYCOUNT,
				mpp_SyncRetryCount));
		}
	}
	else if (strncmp(tag,MPP_SYNCHTIMEOUT,strlen(MPP_SYNCHTIMEOUT)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_SynchTimeout = mppLnx_Config_Parameters.mpp_SynchTimeout= parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", MPP_SYNCHTIMEOUT,
				mpp_SynchTimeout));
		}
	}
	else if (strncmp(tag,MPP_FAILOVERQUIESCETIME,strlen(MPP_FAILOVERQUIESCETIME)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_FailOverQuiescenceTime = mppLnx_Config_Parameters.mpp_FailOverQuiescenceTime = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", 
				MPP_FAILOVERQUIESCETIME, mpp_FailOverQuiescenceTime));
		}
	}
	else if (strncmp(tag,MPP_FAILOVERTIMEOUT,strlen(MPP_FAILOVERTIMEOUT)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_FailoverTimeout = mppLnx_Config_Parameters.mpp_FailoverTimeout = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", 
				MPP_FAILOVERTIMEOUT, mpp_FailoverTimeout));
		}
	}
	else if (strncmp(tag,MPP_FAILBACKTOCURRENTALLOWED,strlen(MPP_FAILBACKTOCURRENTALLOWED)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_FailBackToCurrentAllowed = mppLnx_Config_Parameters.mpp_FailBackToCurrentAllowed = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", 
				MPP_FAILBACKTOCURRENTALLOWED,mpp_FailBackToCurrentAllowed));
		}
	}
	else if (strncmp(tag,MPP_HOLDALTINRESET,strlen(MPP_HOLDALTINRESET)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_HoldAltInReset = mppLnx_Config_Parameters.mpp_HoldAltInReset = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", 
			MPP_HOLDALTINRESET,mpp_HoldAltInReset));
		}
	}
	else if (strncmp(tag,MPP_DOUARETRY,strlen(MPP_DOUARETRY)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_DoUARetry = mppLnx_Config_Parameters.mpp_DoUARetry = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", 
				MPP_DOUARETRY,mpp_DoUARetry));
		}
	}
	else if (strncmp(tag,MPP_VENDORID,strlen(MPP_VENDORID)) == 0) 
	{
		memset (mppLnx_Config_Parameters.mpp_VendorId,'\0',sizeof(mpp_VendorId));
		memset (mpp_VendorId,'\0',sizeof(mpp_VendorId));
		strncpy(mppLnx_Config_Parameters.mpp_VendorId,value,sizeof(mpp_VendorId)-1);
		strncpy(mpp_VendorId,value,sizeof(mpp_VendorId)-1);
		mppLnx_Config_Parameters.mpp_VendorId[sizeof(mpp_VendorId)-1] = '\0';
		mpp_VendorId[sizeof(mpp_VendorId)-1] = '\0';
		MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s is %s\n",	
			MPP_VENDORID,	mpp_VendorId));
	}
	else if (strncmp(tag, MPP_PRODUCTID,strlen( MPP_PRODUCTID)) == 0) 
	{
		memset (mppLnx_Config_Parameters.mpp_ProductId,'\0',sizeof(mpp_ProductId));
		memset (mpp_ProductId,'\0',sizeof(mpp_ProductId));
		strncpy(mppLnx_Config_Parameters.mpp_ProductId,value,sizeof(mpp_ProductId)-1);
		strncpy(mpp_ProductId,value,sizeof(mpp_ProductId)-1);
		mppLnx_Config_Parameters.mpp_VendorId[sizeof(mpp_ProductId)-1] = '\0';
		mpp_ProductId[sizeof(mpp_ProductId)-1] = '\0';
		MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s is %s\n",	
			MPP_PRODUCTID,	mpp_ProductId));
	}
	else if (strncmp(tag, MPP_S2TOS3KEY,strlen( MPP_S2TOS3KEY)) == 0) 
	{
		/*** translate character into digit number ***/
		for (index = 0; index < 16; index++)
		{
			if (value[index] >= '0' && value[index] <= '9')
			{
				value[index] = value[index] + 0 - '0';
			}
			else if (value[index] >= 'a' && value[index] <= 'f')
			{
				value[index] = value[index] + 10 - 'a';
			}
			else if (value[index] >= 'A' && value[index] <= 'F')
			{
				value[index] = value[index] + 10 - 'A';
			}
			else
			{
				MPP_ERRORLOGPRINT ((MPP_ERR_RECOVERED,716,
				"mpp_ReadConfigData: unrecognized char %c in parsing mpp_S2ToS3Key\n",
					value[index]));
				S2ToS3KeyInvalid = TRUE;
				break;
			}
		}
		/*** if the key is invalid, discard the key ***/
		if (S2ToS3KeyInvalid)
			return;
		/*** combine the adjacent two characters into one BYTE ***/
		for (index = 0; index < 8; index++)
		{
			mpp_S2ToS3Key[index] = mppLnx_Config_Parameters.mpp_S2ToS3Key[index] = 
			(((BYTE)(value[2*index])) << 4) | ((BYTE)(value[2*index+1]));
		}
		MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3, "mpp_readConfigData: %s is 0x", MPP_S2TOS3KEY));
		for (index = 0; index < 8; index++)
		{
			MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,"%02x", mpp_S2ToS3Key[index]));
		}
		MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,"\n"));
	}

	else if (strstr(tag,MPP_CONTROLLERIOWAITTIME)!= NULL)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_ControllerIoWaitTime = mppLnx_Config_Parameters.mpp_ControllerIoWaitTime = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", 
				MPP_CONTROLLERIOWAITTIME,mpp_ControllerIoWaitTime));
		}
	}
	else if (strncmp(tag,MPP_ARRAYIOWAITTIME,strlen(MPP_ARRAYIOWAITTIME)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_ArrayIoWaitTime = mppLnx_Config_Parameters.mpp_ArrayIoWaitTime = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", 
				MPP_ARRAYIOWAITTIME,mpp_ArrayIoWaitTime));
		}
	}
	else if (strncmp(tag, MPP_DISABLELUNREBALANCE, strlen(MPP_DISABLELUNREBALANCE)) == 0)
	{
		if (mppLnx_simple_strtoul(value,&endOfString,0,&parseValue))
		{
			mpp_DisableLUNRebalance = mppLnx_Config_Parameters.mpp_DisableLUNRebalance = parseValue;
			MPP_DEBUGPRINT((MPP_INIT_DEBUG,	"mpp_readConfigData: Configuration Parameter %s =	%d\n", 
				MPP_DISABLELUNREBALANCE,mpp_DisableLUNRebalance));
		}
	}
		
	else {
		// unrecognized parameter
		MPP_ERRORLOGPRINT ((MPP_ERR_RECOVERED,717, "mpp_readConfigData: Unrecognized parameter tag %s\n", tag));
	}

}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_simple_strtoul
* SUMMARY:  convert a string to an unsigned long integer
* SCOPE:    local
*
*
* DESCRIPTION: This function convert a string to an unsigned long integer. It does sanity checking for the string, also.
* RESTRICTIONS:
* 
*
* ERRNO: 1
*
*/
BOOL mppLnx_simple_strtoul(const char* cp, char** endp, unsigned int base, unsigned long *parseValue)
{
	*parseValue = simple_strtoul(cp, endp, base);
	if (**endp !='\0')
	{
		MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 730, 
			"The mpp driver	configuration parameter	%d is incorrect\n", 
			*parseValue));
		**endp = '\0';
		return FALSE;
	}
	else
	{
		return TRUE;
	}
}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_readBindingData
* SUMMARY:  read persistent binding file
* SCOPE:    local
*
*
* DESCRIPTION: This function is called in MPP upper level driver initialization routine to read persistent binding information. 
*              This information is used to fill in the array name information in the proper slot in mpp_ModuleArray 
*              according to the virtual target ID of the array.
*              The file has the following format:
*              <VirtualId1>:<arrayName1>'\n'
*              <VirtualId2>:<arrayName2>'\n'
*              ...
*              <VirtualIdN>:<arrayNameN>'\n'
*	       leading or trailing spaces are allowed.
*              Entries that exceed the maximum number of array modules that MPP driver can support will be discarded.
*              Comment lines are allowed as long as they do not have '='
*
* RESTRICTIONS:
* 
* RETURNS: 0 if suceeds, 1 if error (file open error, kernel read error, etc)
*
* ERRNO: 1
*
* NOTES: if error occurs, no persistent binding information is used.
*
*/
static int mppLnx_readBindingData(void)
{
	struct file* filp = NULL;
	char* buf = NULL;  
	char* bufBeginPtr = NULL;
	char* virtualId = NULL;
	char* arrayName = NULL;
	int bytesRead = 0;
	int numberOfArrays = 0;
	LWORD fileSize = 0;
	/*** open parameter file /var/mpp/devicemapping ***/
	filp = filp_open(BINDING_FILE_NAME,O_RDONLY,0);
	if(!filp || IS_ERR(filp)) {
		MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 718, 
			"mppLnx_readBindingData: Error opening /var/mpp/devicemapping\n"));
		return 1;
	}
	fileSize = filp->f_dentry->d_inode->i_size;
	buf = MPP_KMEM_ALLOC(fileSize);
	if(!buf) {
		MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 719, 
			"mppLnx_readBindingData: Error allocating buffer of size %d\n", fileSize ));
		filp_close(filp,0);
		return 1;
	}
	bufBeginPtr = buf;

	/*** Now read the persistent binding file into the buffer ***/
	bytesRead = kernel_read(filp,0,buf,fileSize);

	/*** If read is incomplete, return error status without parsing the buffer ***/
	if(bytesRead != fileSize) {
		MPP_ERRORLOGPRINT ((MPP_ERR_RECOVERED,720,"mpp_ReadBindingData: need to read %d bytes. Only get %d bytes.\n", 
			filp->f_dentry->d_inode->i_size,bytesRead));
		filp_close(filp,0);
		kfree(buf);
		return 1;
	}
	filp_close(filp,0);

 	/*** Now parse the buffer ***/
	while (buf < bufBeginPtr+fileSize && numberOfArrays < mpp_MaxArrayModules && buf != NULL && *buf != 0)
	{
		mppLnx_parseLine(&buf, ":",&virtualId, &arrayName);
		mppLnx_FillInPersistentArray(virtualId, arrayName);
		numberOfArrays++;
		virtualId = arrayName = NULL;
	}
	/*** Now free the buffer and other memory resources that we allocated ***/
	MPP_FREE(bufBeginPtr, fileSize);
	return 0;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_FillInPersistentArray
* SUMMARY:  Fill in module name information for persistent binding arrays
* SCOPE:    local
*
*
* DESCRIPTION: Fill in module name information for persistent binding arrays
*              @1: The virtual target id of the array
*              @2: The array name
*
* RESTRICTIONS:
* 
* RETURNS: void
*
* NOTES: 1. discard any array information with virtual target id exceeding the number that we support
*        2. discard any array information if there is conflicting array name in the slot
*
*/
static void mppLnx_FillInPersistentArray(char* virtualId, char* arrayName)
{
	struct ArrayModuleEntry* pArray = NULL;
	LWORD arrayVirtualId = 0;
	if (virtualId == NULL || arrayName == NULL || *virtualId == 0 || *arrayName == 0)
		return;
	arrayVirtualId = simple_strtoul(virtualId,NULL,0);
	 
	if(arrayVirtualId >= mpp_MaxArrayModules)
	{
		MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 721, 
			"mppLnx_readBindingData: The mpp driver	virtual ID %d is incorrect\n", arrayVirtualId));
	}
	else
	{
		/*** Now assign the array name to the mpp_ModuleArray slot 
		with proper virtual target ID.***/
		pArray = mpp_ModuleArray + arrayVirtualId;
		if (pArray->ModuleNameInConfig)
		{
			MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 722, 
				"mppLnx_FillInPersistentArray: Array slot %d has duplicate information\n", arrayVirtualId));
		}
			
		strncpy(pArray->ModuleName, arrayName, DEVNM_LEN-1);
		pArray->ModuleName[DEVNM_LEN-1] = '\0';
		pArray->ModuleNameInConfig = TRUE;
		MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3, 
			"Assign array name %s to virtual target ID %d\n", 
			arrayName, arrayVirtualId));
	}

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_parseLine
* SUMMARY:  Line parsing function to parse a line into two fields
* SCOPE:    local
*
* DESCRIPTION: This function is called to parse a buffer into two fields delimited by 
*              delimiter and strip off all spaces around the field.
*              Lines without '=' (delimiter) are treated as comment lines and are thus ignored.
*              @1: Address of the buffer to be parsed
*              @2: delimiter character
*              @3: address of the string buffer that is to hold the first part
*              @3: address of the secoond buffer that is to hold the first part
*
* RESTRICTIONS:
* 
* RETURNS: void
*
* ERRNO: 
*
* NOTES: 
*
*/
static void mppLnx_parseLine(char** buf, char* delimiter, char** part1, char** part2)
{
	/* In the following cases:
	   1. The buf contains NULL content
	   2. The buffer pointer is invalid (NULL)
	   return immediately */
	if (*buf == NULL || **buf == 0)
	{
		return ;
	}
	/* devide the buffer into two fields seperated with delimiter */
	*part2 = strsep(buf,"\n");
	*part1 = strsep(part2, delimiter);
	if (part2 == NULL || *part2 == 0)
		return ;
	/* strip off all spaces around the fields */
	*part1 = mppLnx_strstrip(*part1, " \t");
	*part2 = mppLnx_strstrip(*part2, " \t");
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_strstrip
* SUMMARY:  strip leading and trailing characters in string s off string str
* SCOPE:    local
*
*
* DESCRIPTION:  strip leading and trailing characters in string s off string str
*		@1: string to be processed
*		@2: the characters to be stripped off
*
* RESTRICTIONS:
* 
* RETURNS: char *
*/
static char* mppLnx_strstrip (char *str, char* s)
{
        char *eos = str + strlen(str);          /* -> end of string */

        while (*str && strchr(s, *str))
                ++str                           /* strip leading spaces */
        ;
        while ((eos > str) && strchr(s, *(eos - 1)))
                --eos                           /* strip trailing spaces */
        ;
        *eos = '\0';
        return str;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_target_init
* SUMMARY:  MPP upper level driver initialization
* SCOPE:    local
*
*
* DESCRIPTION: This function is called to register the upper level driver with Scsi middle layer.
*
* RESTRICTIONS:
* 
* RETURNS: 0
*
* ERRNO:
*
* NOTES: This function always returns sccess
*
*/
static int mppLnx_target_init(void) {
   return 0;
}


/****************************************************************************
* PROCEDURE
* NAME:    mppLnx_failbackrescan
* SUMMARY: This function starts the failbackscan thread
* SCOPE:   public 
* DESCRIPTION: This function just calls the failbackrescan function provided by the virtual HBA driver. This function is called in response to an ioctl call to the Upper Level driver.
* RESTRICTIONS:
* RETURNS: void
* ERRNO:
* NOTES:
* EXAMPLES:
* SEE ALSO:  
****************************************************************************/

static void 
mppLnx_failBackRescan(void )
{
	void (*failBackRescan_func) (void );

	failBackRescan_func = inter_module_get(MPPLNX_FAILBACKRESCAN_KEY);

	if(failBackRescan_func) {
		(*failBackRescan_func) ();
		inter_module_put(MPPLNX_FAILBACKRESCAN_KEY);
	}
	else
		MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 806, 
			"mppLnx_failBackRescan: inter_module_get did not return a function pointer\n"));

}




/****************************************************************************
* PROCEDURE
* NAME:    mppLnx_newDeviceAdded
* SUMMARY: notify MPP virtual HBA driver of the newly discovered devices
* SCOPE:   public 
* DESCRIPTION: During bus rescan time, new physical devices may be added. The middle level calls MPP 
*              upper level driver's attach() function to notify the mpp upper level driver of the newly 
*              added device. The upper level driver needs to notify the virtual HBA driver to create 
*              /proc device node for the newly added device. 
* RESTRICTIONS:
* RETURNS: void
* ERRNO:
* NOTES:
* EXAMPLES:
* SEE ALSO:  
****************************************************************************/

static void 
mppLnx_newDeviceAdded (Scsi_Device* sdp)
{
	void (*newDeviceAdded_func) ( Scsi_Device* );

	newDeviceAdded_func = inter_module_get(MPPLNX_NEWDENOTIFICATION_KEY);

	if(newDeviceAdded_func) {
		(*newDeviceAdded_func) (sdp);
		inter_module_put(MPPLNX_NEWDENOTIFICATION_KEY);
	}
	else
		MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 723, 
			"mppLnx_newDeviceAdded: inter_module_get did not return a function pointer\n"));

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_target_finish
* SUMMARY:  MPP upper level driver finish routine
* SCOPE:    local
*
*
* DESCRIPTION: This function is called to do the following things:
	       1. dump mpp_ModuleArray data structure for debugging purpose
                  MPP vitual HBA driver.

*
* RESTRICTIONS:
* 
* RETURNS: 0
*
* ERRNO:
*
* NOTES: This function always returns sccess
*
*/
static void mppLnx_target_finish() {
	initialDiscoveryFinished = 1;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_target_attach
* SUMMARY:  MPP upper level driver attaches to the devices using this routine
* SCOPE:    local
*
*
* DESCRIPTION: This function does the following things:
               1. Claim LSI physical device
               2. Notify MPP virtual HBA driver of the newly-discovered devices if they
                  are added after the initial scan of the low level HBA driver.
*
* RESTRICTIONS:
* 
* RETURNS: 0 if the device is successfully processed (no matter if it is attached or not),
*
*/
static int mppLnx_target_attach(Scsi_Device *sdp) 
{
   mppLnx_UnconfiguredDeviceEntry_t * unConfiguredDeviceEntry = NULL;

	//printk ("mppLnx_target_attach invoked for H%dT%dL%d\n", sdp->host->host_no, sdp->id, sdp->lun);
	/* check if the Scsi_Device is a virtual device */
	if (strcmp(sdp->host->hostt->proc_name, MPPLNX_PROC_NAME) == 0)
	{
		return 0;
	}

	/*
	* workaround for setting mppLnx_host_procName
	*/
	if(!is_mppLnx_host_procName_set)
	{
		mppLnx_host_procName = sdp->host->hostt->proc_name;
	}

        if ( mppLnx_IsArrayDevice((MPP_HANDLE)sdp, sdp->lun))
	{
		/* mppLnx_IsArrayDevice returns true if the device has already been discovered and is in the RdacInfo data structure */
		is_mppLnx_host_procName_set = 1;

      unConfiguredDeviceEntry = mppLnx_getUnconfiguredDevice(sdp);

      /* we increase the dev_noticed only if the device is a configured device
      * or a un-configured device is first-time passed to this function
      */
      if(unConfiguredDeviceEntry == NULL )
      {
         mpp_template.nr_dev++;
      }
      else if(unConfiguredDeviceEntry->isUnconfigured == TRUE && 
         unConfiguredDeviceEntry->isAttached == FALSE )
      {
		   mpp_template.nr_dev++;
         unConfiguredDeviceEntry->isAttached = TRUE;
      }

		/* notify virtual HBA driver of the newly discovered device 
		if this is not the initial scan */
		if (initialDiscoveryFinished
            && (unConfiguredDeviceEntry == NULL || 
                unConfiguredDeviceEntry->isUnconfigured == FALSE) )
      {
			mppLnx_newDeviceAdded(sdp);
      }
	}
	return 0;
    
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_target_detect
* SUMMARY:  MPP upper level driver device detection routine
* SCOPE:    local
*
* DESCRIPTION: This function is called when a new device is discovered to see
               if the upper level driver is interested in claiming this device
*
* RESTRICTIONS:
* 
* RETURNS: 1 for LSI physical device, 0 otherwise
* NOTES:
*       This function may be invoked multiple times with the same "sdp" object 
*       because of the un-configured device case.
*
*/
static int mppLnx_target_detect(Scsi_Device *sdp) 
{
	ControllerAddress_t              ctlAddress;
   mppLnx_UnconfiguredDeviceEntry_t * unConfiguredDeviceEntry = NULL;
	
	//printk ("mppLnx_target_detect invoked for H%dT%dL%d\n", sdp->host->host_no, sdp->id, sdp->lun);
	/* check if the Scsi_Device is a virtual device */
	if (strcmp(sdp->host->hostt->proc_name, MPPLNX_PROC_NAME) == 0)
	{
		return 0;
	}

	ctlAddress.ChannelId = sdp->channel;
	ctlAddress.TargetId  = sdp->id;
	ctlAddress.HostId    = sdp->host->host_no;

   /*
   * workaround for setting mppLnx_host_procName
   */
   if(!is_mppLnx_host_procName_set)
   {
      mppLnx_host_procName = sdp->host->hostt->proc_name;
   }

	
	/* insert the Scsi_Device into mpp_ModuleArray */
	if(mppLnx_BuildRdacinfo((MPP_HANDLE)sdp, &ctlAddress, sdp->lun)) 
   {
      is_mppLnx_host_procName_set = 1;
      unConfiguredDeviceEntry = mppLnx_getUnconfiguredDevice(sdp);

      /* we increase the dev_noticed only if the device is a configured device
      * or a un-configured device is first-time passed to this function
      */
      if(unConfiguredDeviceEntry == NULL)
      {
         mpp_template.dev_noticed++;
		   return 1;
      }
      else if (unConfiguredDeviceEntry->isUnconfigured == TRUE && 
         unConfiguredDeviceEntry->isDetected == FALSE )
      {
		   mpp_template.dev_noticed++;
         unConfiguredDeviceEntry->isDetected = TRUE;
		   return 1;
      }
      else
      {
         return 0;
      }

	}
	else  /* the Scsi_Device is a non-LSI device */
	{
		return 0;
	}
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_target_detach
* SUMMARY:  MPP upper level driver detach routine
* SCOPE:    local
*
* DESCRIPTION: This function is invoked when a physical device is going to be removed.
               It does the following things:
               1. check if it is a LSI physical device
               2. If it is, then:
                  a) Tear down associated data structure in rdacInfo for this physical device.
                  b) decrement device attach counter and device number in mpp_template.
*
* RESTRICTIONS:
* 
* RETURNS: none
*
* ERRNO:
*
* NOTES: 
*
*/
static void mppLnx_target_detach(Scsi_Device *sdp) 
{
	ControllerAddress_t ctlAddress;
	ArrayModuleEntry_t *ModuleArray;
	RdacDeviceInformation_t* RdacInfo;
	RdacLunInformation_t *pRdacLun;
	LunPathObjects_t *pCtlPath;
	LunPathInfo_t *pLunPathInfo;
	BOOL pathAlreadyRemoved = FALSE;
	BYTE ctlIndex;
	BYTE pathIndex;
	LWORD count;
   mppLnx_UnconfiguredDeviceEntry_t * deviceEntry = NULL;

	MPP_DEBUGPRINT ((MPP_DETACH_DEBUG,
	"mppLnx_target_detach invoked for h%dt%dl%d, device vertex 0x%p, attached=%d\n", 
	sdp->host->host_no, sdp->id, sdp->lun, sdp, sdp->attached));

	/* Need to check if the Lun number exceeds the configured Lun number that MPP
	   driver can support */
	if (sdp->lun >= mpp_MaxLunsPerArray)
	{
		MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 730, 
			"mppLnx_target_detach: Lun number %d exceeds mpp_MaxLunsPerArray (%d)\n",
			sdp->lun, mpp_MaxLunsPerArray));
		return;
	}

   /*
   * if it is a unconfigured device, remove it from the unconfigured device list
   */
   deviceEntry = mppLnx_getUnconfiguredDevice(sdp);
   if( (deviceEntry != NULL) && (deviceEntry->isUnconfigured == TRUE) )
   {
      sdp->attached--;
		mpp_template.dev_noticed--;
		mpp_template.nr_dev--;
      mppLnx_deleteUnconfiugredDevice(sdp);
      return;
   }
		

	ctlAddress.ChannelId = sdp->channel;
	ctlAddress.TargetId  = sdp->id;
	ctlAddress.HostId    = sdp->host->host_no;

	/* check if virtual device. If it is, return */
	if (strcmp(sdp->host->hostt->proc_name, MPPLNX_PROC_NAME) == 0)
		return;

	/* Found the device information in RdacInfo. If it is in RdacInfo data structure,
       remove the asssociated structure from RdacInfo
	   and decrement the attach counter for that device if it is a physical device.*/
	for(count=0; count<mpp_MaxArrayModules; ++count) 
	{
		ModuleArray = mpp_ModuleArray+count;
		if( (RdacInfo = ModuleArray->RdacInfo) ) 
		{
			if(RdacInfo->RdacLuns[sdp->lun].Present)
			{
				pRdacLun = &RdacInfo->RdacLuns[sdp->lun];
				for (ctlIndex = 0; ctlIndex < 2; ctlIndex++)
				{
					pCtlPath = pRdacLun->LunControllers[ctlIndex];
					/* We need to loop through 0..MaxLunsPerArray to real with the following scenarios:
					   We have Lun 1,2,3, the user removes a single lun(e.g., Lun0 and leave Lun1, Lun2,
					   if we loop through 0.. NumberofLunObjects(2), we can not find Lun2
					   in RdacInfo data structure. */
					for (pathIndex = 0; pathIndex < mpp_MaxPathsPerController; pathIndex++)
					{ 
						pLunPathInfo = &pCtlPath->LunPaths[pathIndex];
						if (!pLunPathInfo->LunPathDevice) 
						{
							continue;
						}
						else 
						{
							if ((Scsi_Device*)pLunPathInfo->LunPathDevice == sdp) 
							{
								mpp_RemoveLun((MPP_HANDLE)sdp, RdacInfo, ctlIndex, pathIndex, sdp->lun);
								pathAlreadyRemoved = TRUE;
								/* This is a physical disk, we need to decrement 
								the attach counter and other relevant data*/
								sdp->attached--;
								mpp_template.dev_noticed--;
								mpp_template.nr_dev--;
								MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_2,
								"mppLnx_target_detach: Removed physical dev 0x%p, Array %s, Ctl %c, Lun %d, Path %d, attached=%d\n", 
								sdp, ModuleArray->ModuleName, ctlIndex+'A'-0, sdp->lun, pathIndex, sdp->attached)); 
								break;
							}                               
							else
							{
								continue;
							}
						}
					}
					if (pathAlreadyRemoved)
						break;
				}
				if(pathAlreadyRemoved)
					break;
			}       
		}
		if(pathAlreadyRemoved)
			break;
	}

}

		
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_target_init_command
* SUMMARY:  MPP upper level driver command initialization routine
* SCOPE:    local
*
* DESCRIPTION: This function is a dummy function. It should never be invoked.
*
* RESTRICTIONS:
* 
* RETURNS: 1
*
* ERRNO:
*
* NOTES: This function log an error message in MPP error log
*
*/
static int mppLnx_target_init_command(Scsi_Cmnd *sdp) {
	MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 724, "mpp_target_init_command: invoked"));
        return 1;
}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_open
* SUMMARY:  MPP upper level driver open function
* SCOPE:    local
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION: This function is called by the kernel whenever some program tries to open the 
* upper level character driver. Need to do a __MOD_INC_USE_COUNT to increment the usage counter
* of the module. This is done to make sure the module cannot be unloaded. 
*
* RESTRICTIONS:
* 
* RETURNS: 0
*
* ERRNO:
*
* NOTES: This function always returns success
*
*/
static int mppLnx_open (struct inode *inode, struct file *file)
{
	/* Increment the usage counter of the module */
	if(mpp_template.module)
		__MOD_INC_USE_COUNT(mpp_template.module);

	return 0;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_release
* SUMMARY:  MPP upper level driver release function
* SCOPE:    local
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION: This function is called by the kernel whenever some program tries to close the 
* upper level character driver. Need to do a __MOD_DEC_USE_COUNT to decrement the usage counter
* of the module. This is done to make sure the module can be unloaded. 
*
* RESTRICTIONS:
* 
* RETURNS: 0
*
* ERRNO:
*
* NOTES: This function always returns success
*
*/
static int mppLnx_release (struct inode *inode, struct file *file)
{

	/* Decrement the usage counter for the module */
	if(mpp_template.module)
		__MOD_DEC_USE_COUNT(mpp_template.module);

	return 0;
}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_ioctl
* SUMMARY:  MPP upper level driver ioctl function
* SCOPE:    local
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION: This function is called by the kernel whenever some program tries to use the ioctl 
* interface of the upper level character driver. 
*
* RESTRICTIONS:
* 
* RETURNS: 0
*
* ERRNO:
*
* NOTES: This function returns success or EFAULT or ENXIO depending on the outcome of the 
* copy-to-user and copy-from-user operations
* 
*/

static int mppLnx_ioctl(struct inode *inode, struct file *file, unsigned int cmd_in, unsigned long arg)
{
	getRdacInfoIoctl_t UserArg;
	ArrayModuleEntry_t *ModPtr;
	RdacDeviceInformation_t *RdacInfo; 
	BOOL bSilent = FALSE;
	struct Scsi_Host *shpnt;
	LINT ioctl_status = 0;
	struct host_target_t host_target_arg;

	MPP_DEBUGPRINT((MPP_IOCTL_DEBUG+MPP_DEBUG_LEVEL_4, "mpp_ioctl: called for dev cmd = 0x%lx, arg = 0x%lx\n", cmd_in, arg));
   
	
	switch (cmd_in) {
		/* Return the MaxPathsPerController parameter */
	       	case MPP_IOCTL_GETPATHS:
			ioctl_status = copy_to_user((void *)arg, &mpp_MaxPathsPerController, sizeof(LWORD));
			if( ioctl_status )
				ioctl_status = -EFAULT;
			break;
		/* Return the MaxLunsPerArray parameter */ 
       		case MPP_IOCTL_GETLUNS:
			ioctl_status = copy_to_user((void *)arg, &mpp_MaxLunsPerArray,  sizeof(LWORD));
			if( ioctl_status )
				ioctl_status = -EFAULT;
			break;
		/* Return the MaxArrayModules parameter */ 
       		case MPP_IOCTL_GETMODULES:
			ioctl_status = copy_to_user((void *)arg, &mpp_MaxArrayModules,  sizeof(LWORD));
			if( ioctl_status )
				ioctl_status = -EFAULT;
			break;
		/* Return the MaxTargetPerPhysicalHost parameter after walking through the scsi_hostlist */ 
       		case MPP_IOCTL_GET_PHYSICALHOST_MAX_TARGET:
		
			ioctl_status = copy_from_user(&host_target_arg, (void *)arg, sizeof(struct host_target_t));
			if( ioctl_status )
				ioctl_status = -EFAULT;
			for(shpnt = scsi_hostlist; shpnt; shpnt = shpnt->next) {
				if(shpnt->host_no == host_target_arg.host_id ) {
					if(strcmp(shpnt->hostt->proc_name, "mpp") == 0) 
						host_target_arg.max_target_id =	-1;
					else
						host_target_arg.max_target_id = shpnt->max_id;	
				}
			}
			ioctl_status = copy_to_user((void *)arg, &host_target_arg, sizeof(struct host_target_t));
			if( ioctl_status )
				ioctl_status = -EFAULT;
		
			break;

		/* Return the RdacInfo corresponding to the virtual target ID supplied */
       		case MPP_IOCTL_GETRDACINFO:
		{
			ioctl_status = copy_from_user(&UserArg, (void *)arg, sizeof(getRdacInfoIoctl_t));
			if( ioctl_status ) {
				ioctl_status = -EFAULT;
				break;
			}

			ModPtr = mpp_ModuleArray;
			if( UserArg.target > (mpp_MaxArrayModules-1) ) {
				ioctl_status = -ENXIO;
				break;		       
			}
			ModPtr += UserArg.target;
			RdacInfo = ModPtr->RdacInfo;
			if( !RdacInfo ) {
				ioctl_status = -ENXIO;
				break;		       
			}
			if( UserArg.BufferSize < MPP_RDACINFOSIZE(mpp_MaxLunsPerArray, mpp_MaxPathsPerController) ) {
				MPP_ERRORLOGPRINT((MPP_ERR_RETRYABLE, 800, "mpp_ioctl: ioctl failed on device input Buffer size %d, RdacInfo size %d\n", UserArg.BufferSize, MPP_RDACINFOSIZE(mpp_MaxLunsPerArray, mpp_MaxPathsPerController)));
				ioctl_status = -ENOMEM;
				break;
			}
			ioctl_status = copy_to_user(UserArg.RdacInfoAdrs, RdacInfo, MPP_RDACINFOSIZE(mpp_MaxLunsPerArray, mpp_MaxPathsPerController));
			if( ioctl_status ) {
				ioctl_status = -EFAULT;
				break;
			}

			ioctl_status = copy_to_user( UserArg.RdacName, RdacInfo->ModuleName, sizeof(DevName_t) );
			if( ioctl_status ) {
				ioctl_status = -EFAULT;
				break;
			}
			break;
		}
       		case MPP_IOCTL_FAILBACKSCAN:
			/* Call the FailBackScan function */
			mppLnx_failBackRescan();
			break;
       		case MPP_IOCTL_AVTCHECKSCAN:
			/* Call the CheckAVTState function */
			mpp_CheckAVTState();
			break;

        	case MPP_IOCTL_BUSSCAN:
    			mppLnx_RescanUnconfiguredDevices();
      			break;
		case MPP_IOCTL_SETDEBUG:
			/* Set the debug value to the value supplied by the user */
			ioctl_status = copy_from_user( &mpp_Debug, (void *)arg, sizeof(LWORD));
			if( ioctl_status ) 
				ioctl_status = -EFAULT;
			break;

		case MPP_IOCTL_SETERROR:
			/* Set the error level to the value supplied by the user */
			ioctl_status = copy_from_user( &mpp_ErrorLevel, (void *)arg, sizeof(LWORD));
			if( ioctl_status ) 
				ioctl_status = -EFAULT;
			break;
		
		default:
			ioctl_status = -EINVAL;
			bSilent = TRUE;   // remain silent if we have no idea what to do 
	}

	if( ioctl_status  && !bSilent)
	{
		MPP_ERRORLOGPRINT((MPP_ERR_ALL, 801, "mpp_ioctl: ioctl failed on device 0x%lx, status %d, device type %d, cmd 0x%lx\n", 
		 ioctl_status, cmd_in));
	}

    	return(ioctl_status);


}


 /*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_Translate_Linux_Status 
* SUMMARY:  Returns an equivalent MPP status code based on status returned by the target and low level driver .
* SCOPE:    public
*
*
* DESCRIPTION:
* This function has to translate the SCSI command status codes returned by the SCSI middle level 
* layer to an appropriate MPP driver status code. It takes a SCSI command pointer as argument. The 
* function has to translate the status of this SCSI command.
* There are four bytes of data in the result field of the SCSI command that is returned. The first 
* byte is the SCSI_STATUS. If this field returns a valid status, it means the connection to the 
* target is fine, only an error occurred in executing the SCSI command. It also means that we have 
* the sense keys to make further sense out of any error that occurred. The second byte is the SCSI 
* message. This gives more information about the status of the SCSI command, in which it failed. 
* The third byte is the status returned by the low-level driver. The last byte is the error code 
* inserted by the linux SCSI middle-layer code. If a SCSI error occurs, a SCSI sense key is returned
* and this also has to be checked so that more information about the errior may be obtained. This is
* done in the common code mpp_AnalyseIOError function. All these conditions have to be converted to
* MPP specific status codes.
 *      status byte = set by target device
 *      msg_byte    = return status from host adapter itself.
 *      host_byte   = set by low-level driver to indicate status.
 *      driver_byte = set by mid-level.
*
* RESTRICTIONS:
* 
* RETURNS: An LWORD which is the equivalent MPP status, which are listed in MPP_Common.h
*
* ERRNO:
*
* NOTES: Called from the function mpp_SysdepSynchronousIo
*
* EXAMPLES:
*
* SEE ALSO:
*
*/


static LWORD mppLnx_Translate_Linux_Status (int result) {
	LWORD  retVal =  MPP_UNRECOGNIZED_OS_STATUS;

	MPP_DEBUGPRINT((MPP_SYNC_IO_DEBUG+MPP_DEBUG_LEVEL_2, "mppLnx_Translate_Linux_Status: %d %d %d\n", host_byte(result), status_byte(result), msg_byte(result)));

	switch ( host_byte( result ) ) {
		/* No error */
		case DID_OK:
			switch ( status_byte( result ) ) {
				case GOOD: 
					retVal = MPP_GOOD_STATUS;
					break;
				case COMMAND_TERMINATED:
					retVal =  MPP_UNRECOGNIZED_OS_STATUS;
	         			break;
				case CHECK_CONDITION:
					retVal = MPP_CHECK_CONDITION;
					break;
				case BUSY:
					retVal = MPP_SCSI_BUSY;
					break;
				case RESERVATION_CONFLICT:
					retVal = MPP_SCSI_RESERVATION;
					break;
				default:
					retVal = MPP_UNSUPPORTED_SCSI_STATUS;
			}
			break;
		/* SCSI SELECTION failed because there was no device at the address specified */
		/* DID_NO_CONNECT will abort the command */
		case DID_NO_CONNECT:  
			retVal = MPP_SELECTION_TIMEOUT;
			break;
		/* SCSI ARBITRATION failed */
		/* DID_BUS_BUSY will force the command to be retried
       * This is case of command->result 0x20000 case. QLA driver returns this status even
       * in the case of command time-out. We translate this to MPP_SCSI_BUSY. This mpp status
       * will cause normal IO retry, sync IO retry and failover command retry.
       */
		case DID_BUS_BUSY:
			retVal = MPP_SCSI_BUSY;
			break;
		/* A time-out occurred for some unknown reason, probably during SELECTION or while waiting for RESELECTION */
		case DID_TIME_OUT: 
			retVal = MPP_SELECTION_TIMEOUT;
			break;
		/* The SCSI ID of the target was the same as the SCSI ID of the host adapter */
		case DID_BAD_TARGET:
			retVal = MPP_INVALID_REQUEST;
			break;
		/* The high-level code called the low-level abort() function */
		case DID_ABORT:
			retVal = MPP_COMMAND_ABORTED;
			break;
		/* A SCSI PARITY error was detected */
		case DID_PARITY:
			retVal = MPP_HARDWARE_ERROR;
			break;
		/* An error occurred which lacks a more appropriate error code (for example, an internal host adapter error) */

		case DID_ERROR:
			retVal = MPP_HARDWARE_ERROR;
			break;
		/* The high-level code called the low-level reset() function */
		case DID_RESET:
			retVal = MPP_HARDWARE_ERROR;
			break;
		/* An unexpected interrupt occurred and there is no appropriate way to handle this interrupt */
		case DID_BAD_INTR:
			retVal = MPP_HARDWARE_ERROR;
			break;
		default:
			retVal = MPP_UNRECOGNIZED_OS_STATUS;
	}
	/* Here the middle layer is telling us that a timeout has occurred on the SCSI command. */
	if (msg_byte(result)  == DRIVER_TIMEOUT) 
		retVal = MPP_COMMAND_TIMEOUT;
	MPP_DEBUGPRINT((MPP_SYNC_IO_DEBUG+MPP_DEBUG_LEVEL_2, "mpp_mppLnx_Translate_Linux_Status: %d \n", retVal));

	return retVal;

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mpp_SysdepSynchronousIo
* SUMMARY:  This is the system dependent part of the Synchronous IO function .
* SCOPE:    public
*
*
*DESCRIPTION   
* mpp_SysdepSynchronousIO starts the I/O indicated by the cdb on the device scsiDevice
* and waits for it to complete. bufferAddress and bufferLength define the data buffer, which will be
* returned by or sent to the scsi target. senseBuffer and senseLength define the sense data buffer.
* DataDirection indicates the direction of the data transfer - zero for output and non-zero for 
* input. On return from this function, dataLengthXfered and senseLengthReturned will indicate the 
* amount of data transferred and amount of sense data transferred.
* So this function takes the SCSI command, which is in the form of a cdb and converts it to a 
* platform-based SCSI request and sends it down to the linux SCSI middle layer. The SCSI data, which
* is returned, in return is then passed back to the calling function. If the SCSI command results in
* an error, then the sense key data is sent back so that sense can be made of the error.
* For this we will use the scsi_request structure, which is essentially a slimmed down version of 
* Scsi_Cmnd. Here we set its fields, like the actual SCSI command and the data buffer and the length
* of the data buffer.  Then we use the scsi_wait_req() function, exported by the middle layer, and 
* pass the scsi command to the middle layer. To this function we also pass the MPP_SYNCTIMEOUT value
* as the timeout value for the command and mpp_SyncRetryCount for the number of times this command 
* should be retried, both of which are defined in MPP_Common.h. Depending on whether the SCSI 
* command was successful or not, we do or do not send any sense data back to the calling function.
* We also need to call mppLnx_Translate_Linux_Status to translate the status retuned by 
* scsi_wait_req to an MPP specific status. If the status returned is MPP_CHECK_CONDITION, sense data
* would also have to be sent back.
*
* RESTRICTIONS:
* 
* RETURNS: An LWORD which is the equivalent MPP status, which are listed in MPP_Common.h
*
* ERRNO:
*
* NOTES: Called from the function mpp_SynchronousIo
*
* EXAMPLES:
*
* SEE ALSO:
*
*/


LWORD mpp_SysdepSynchronousIo(MPP_HANDLE ScsiDevice, BYTE *cdb,	BYTE cdblen, void *BufferAddress, 
								LWORD BufferLength,	void *SenseBuffer, 
								LWORD SenseLength, TINY	DataDirection, 
								LWORD *SenseLengthReturned,
								LWORD *DataLengthXfered)
{
	
	LWORD max_retrys = 0;
	Scsi_Request *SRpnt = NULL;
	Scsi_Device  *dev;
	LWORD mppStatus = 0;
	unsigned char cmd[cdblen];
	int count = MPPLNX_SCSI_ALLOCREQ_RETRY_COUNT ;/*SCSI_ALLOC_RETRY_COUNT what value would be best? */
	dev = ScsiDevice;

	/* Allocate the scsi request data structure. We have to request the middle layer to do this for us */
	while(count != 0) {
	        SRpnt = scsi_allocate_request(dev);
       		if (!SRpnt) {
			MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 802,"mpp_SysdepSynchronousIo: scsi_alloc_consistent_buf failed\n")); 
			mpp_ThreadDelay( MPPLNX_SCSI_ALLOCREQ_WAIT, MPPCMN_DELAY_MILLISECONDS );
			count--;

		}
		if ( SRpnt)
			break;

        }
	
	if(count == 0)
                return (MPP_SCSI_BUSY); /* need to return the right value here */

	/* The length of the cdb is determined by the SCSI spec, just an extra check to make sure we are not 
	   presented with a cdb greater than what the linux kernel can handle */
	if(cdblen > MAX_COMMAND_SIZE )  {
		MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 803,"mpp_SysdepSynchronousIo: cdb length greater than can be assigned to the Scsi Request \n")); 

		return (MPP_SCSI_BUSY); /* need to return the right value here */
	}


	/* we copy the contents of the cdb to the command array in the Scsi Request data structure */
	memset (cmd, 0, cdblen);
	memcpy(cmd, cdb, cdblen);

	/* Make the Buffer pointer in the Scsi request point to the Buffer argument to this function, as memory 
	   has already been alocated for it */
	SRpnt->sr_buffer = BufferAddress;	
	
	/* We choose a value of 0 becuase the failed command is tried mpp_SyncRetryCount number of times by the 
           common code */
	max_retrys = 0; 


	if (DataDirection == MPPCMN_XFER_DEV_HOST)
	{
		SRpnt->sr_data_direction =  SCSI_DATA_READ;
	}
	else if(DataDirection == MPPCMN_XFER_HOST_DEV)
	{
		SRpnt->sr_data_direction =  SCSI_DATA_WRITE;
   }
   else if(DataDirection == MPPCMN_XFER_NONE)
   {
      SRpnt->sr_data_direction = SCSI_DATA_NONE;
   }else
   {
      /*
      * Log an error message
      */
      SRpnt->sr_data_direction = SCSI_DATA_UNKNOWN;
      MPP_ERRORLOGPRINT((MPP_ERR_RECOVERED, 807, 
			"Synchronous IO request with data transfer direction unknown:%d\n",
         DataDirection));
   }

	/* More initializations */
	SRpnt->sr_cmd_len = 0;
	SRpnt->sr_sense_buffer[0] = 0;
	SRpnt->sr_underflow = 0;


	/* This function is provided by the middle layer and exported by it too. We send the Scsi Request down to the middle layer 
	   which will in turn pass it on to the low level driver */

	scsi_wait_req (SRpnt, (void *) cmd, (void *) BufferAddress, BufferLength, mpp_SynchTimeout * HZ  , max_retrys);

	MPP_DEBUGPRINT((MPP_SYNC_IO_DEBUG+MPP_DEBUG_LEVEL_2, "mpp_SysdepSynchronousIo: Returned from scsi_wait_req \n"));

	/* Next we need to translate the status returned by the target, low level driver and middle layer to an equivalent MPP  status */
	mppStatus = mppLnx_Translate_Linux_Status(SRpnt->sr_result);


	/* Need to check how much data is actually transfered */
	if(	mppStatus == MPP_GOOD_STATUS ) 
	{
		*DataLengthXfered = 1; /* some positive value right now, so that the code works */
	}
	else 
		*DataLengthXfered = 0; 


	/* We need to return the sense data back if the error condition is MPP_CHECK_CONDITION */
	if(	mppStatus == MPP_CHECK_CONDITION ) 
	{
		*SenseLengthReturned = MIN(SenseLength,	SCSI_SENSE_BUFFERSIZE); 
		bcopy ( SRpnt->sr_sense_buffer, SenseBuffer, *SenseLengthReturned ); 
	} 
	else
	{
		*SenseLengthReturned = 0; 
	}

	/* Release the Scsi Request we allocated as we are doen with it */
	scsi_release_request(SRpnt);

	MPP_DEBUGPRINT((MPP_SYNC_IO_DEBUG+MPP_DEBUG_LEVEL_2, "mpp_SysdepSynchronousIo: sync i/o done,  mppStatus %d, sense len %d, data len %d\n", mppStatus,*SenseLengthReturned,*DataLengthXfered)); 


	return (mppStatus);

}



/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_sd_detect
* SUMMARY:  MPP replacement detection routine for sd driver
* SCOPE:    local
*
* DESCRIPTION: This replacement function is called by the scsi middle layer when a new device is 
* discovered to see if the sd driver is interested in claiming this device. We replace the standard
* detect function for the sd driver with our replacement function.
* RESTRICTIONS:
* 
* RETURNS: calls original sd_detect for LSI virtual device and root/boot devices, and all other 
* non-LSI devices. Returns a 0 on all LSI physical devices and UTM devices so that sd does not 
* know of these devices. It calls mppLnx_BuildRdacinfo to determine if a device is LSI or Non-LSI.
*/

static int mppLnx_sd_detect(Scsi_Device* SDpnt)
{
	ControllerAddress_t ctlAddress;
	
	/* check if virtual device  */
	if (strcmp(SDpnt->host->hostt->proc_name, MPPLNX_PROC_NAME) == 0) {
		return sd_old_detect(SDpnt);
	}

	/* Fill in values to be sent to mppLnx_BuildRdacinfo */
	ctlAddress.ChannelId = SDpnt->channel;
	ctlAddress.TargetId  = SDpnt->id;
	ctlAddress.HostId    = SDpnt->host->host_no;
	

   /*
   * workaround for setting mppLnx_host_procName
   */
   if(!is_mppLnx_host_procName_set)
   {
      mppLnx_host_procName = SDpnt->host->hostt->proc_name;
   }

	/* Call mppLnx_BuildRdacinfo */
	if(mppLnx_BuildRdacinfo((MPP_HANDLE)SDpnt, &ctlAddress, SDpnt->lun))
	{
      is_mppLnx_host_procName_set = 1;
		/* mppLnx_BuildRdacinfo returns true if it inserts the lun into the RdacInfo data structure or if the device has already been discovered */

		/* So if we come here it means that this Scsi_Device is a physical LUN and is part of an LSI array */

	
			return 0;

	}

	/* Not an LSI Device */
	return sd_old_detect(SDpnt);   /* call original function */

}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_sg_detect
* SUMMARY:  MPP replacement detection routine for sg driver
* SCOPE:    local
*
* DESCRIPTION: This replacement function is called by the scsi middle layer when a new device is 
* discovered to see if the sg driver is interested in claiming this device. We replace the standard
* detect function for the sg driver with our replacement function.

* RESTRICTIONS:
* 
* RETURNS: calls original sg_detect for LSI virtual devices and UTM and Root/Boot devices, and all 
* other non-LSI devices. Returns a 0 on all LSI physical devices so that sg does not know of these 
* devices. It calls mppLnx_BuildRdacinfo to determine if a device is LSI or Non-LSI. 
*/


static int mppLnx_sg_detect(Scsi_Device* SDpnt)
{
	ControllerAddress_t ctlAddress;
	LWORD device_type;
	
	
	/* check if virtual device  */
	if (strcmp(SDpnt->host->hostt->proc_name, MPPLNX_PROC_NAME) == 0) {
		return sg_old_detect(SDpnt);
	}

	/* Fill in values to be sent to mppLnx_BuildRdacinfo */
	ctlAddress.ChannelId = SDpnt->channel;
	ctlAddress.TargetId  = SDpnt->id;
	ctlAddress.HostId    = SDpnt->host->host_no;
   /*
   * workaround for setting mppLnx_host_procName
   */
   if(!is_mppLnx_host_procName_set)
   {
      mppLnx_host_procName = SDpnt->host->hostt->proc_name;
   }
	

	/* Call mppLnx_BuildRdacinfo */
	if(mppLnx_BuildRdacinfo(SDpnt, &ctlAddress, SDpnt->lun))
	{
		/* mppLnx_BuildRdacinfo returns true if it inserts the lun into the RdacInfo data structure or if the device has already been discovered */
		/* So if we come here it means that this Scsi_Device is a physical LUN and is part of an LSI array */

	   is_mppLnx_host_procName_set = 1;
		/* Find out what kind of LSI device it is */
		if (( device_type = mppLnx_DeviceType(SDpnt)) ) {

			/* if it is a UTMLUN */
			if ( device_type == MPP_UTM_LUN_DEVICE ) {
				return sg_old_detect(SDpnt);		     /* we want sg to claim this device */
			}
          
			/* It is a normal LSI physical lun */
			return 0;
		}
		else 
		{

			MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 804, "Is a LSI physical device but is an Unconfigured Device \n")); 
			return 0;

		}
	}

	/* Not an LSI Device */
	return sg_old_detect(SDpnt);   /* call original function */

}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_sd_attach
* SUMMARY:  MPP replacement attach routine for sd driver
* SCOPE:    local
*
* DESCRIPTION: This replacement function is called by the scsi middle layer when a new device has 
* to be claimed to see if the sd driver is interested in claiming this device. We replace the 
* standard attach function for the sd driver with our replacement function.
* RESTRICTIONS:
* 
* RETURNS: calls original sd_attach for LSI virtual devices and root/boot devices, and all other 
* non-LSI devices. Returns a 0 on all LSI physical devices and UTM devices so that sd does not know
* of these devices. It calls mppLnx_BuildRdacinfo to determine if a device is LSI or Non-LSI. 
*/

int mppLnx_sd_attach(Scsi_Device* SDpnt)
{
	ControllerAddress_t ctlAddress;
	RdacDeviceInformation_t	*pRdacDevice;
	RdacLunInformation_t	*pRdacLun;
	LINT LockLevel;
	
	
	/* Fill in values to be sent to mppLnx_BuildRdacinfo */
	ctlAddress.ChannelId = SDpnt->channel;
	ctlAddress.TargetId  = SDpnt->id;
	ctlAddress.HostId    = SDpnt->host->host_no;
	
	/* check if virtual device  */
	if (strcmp(SDpnt->host->hostt->proc_name, MPPLNX_PROC_NAME) == 0) {

		//make sure you point to the right RdacInfo while locking and unlocking
		pRdacDevice = mpp_ModuleArray[(SDpnt->id)].RdacInfo;
		pRdacLun = &pRdacDevice->RdacLuns[SDpnt->lun];
		if ( !pRdacLun->PseudoLunObject ) {
			MPP_LOCK(pRdacDevice->Lock, LockLevel);
			pRdacLun->PseudoLunObject = SDpnt;
			MPP_UNLOCK(pRdacDevice->Lock, LockLevel);
		}

		return sd_old_attach(SDpnt);
	}

	/*
	* workaround for setting mppLnx_host_procName
	*/
	if(!is_mppLnx_host_procName_set)
	{
		mppLnx_host_procName = SDpnt->host->hostt->proc_name;
	}


	/* Call mppLnx_IsArrayDevice  */
        if ( mppLnx_IsArrayDevice((MPP_HANDLE)SDpnt, SDpnt->lun))
	{
		is_mppLnx_host_procName_set = 1;
		/* mppLnx_IsArrayDevice returns true if the device has already been discovered and is in the RdacInfo data structure */

	
		return 0;

	}

	/* Not an LSI Device */
	return sd_old_attach(SDpnt);   // call original function

}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_sg_attach
* SUMMARY:  MPP replacement attach routine for sg driver
* SCOPE:    local
*
* DESCRIPTION: This replacement function is called by the scsi middle layer when a new device is to
* be claimed to see if the sg driver is interested in claiming this device. We replace the standardi
* attach function for the sg driver with our replacement function.

* RESTRICTIONS:
* 
* RETURNS: calls original sg_attach for LSI virtual devices and UTM and Root/Boot devices, and all 
* other non-LSI devices. Returns a 0 on all LSI physical devices so that sg does not know of these 
* devices. It calls mppLnx_BuildRdacinfo to determine if a device is LSI or Non-LSI.
*/


int mppLnx_sg_attach(Scsi_Device* SDpnt)
{
	ControllerAddress_t ctlAddress;
	LWORD device_type;
	RdacDeviceInformation_t	*pRdacDevice;
	RdacLunInformation_t	*pRdacLun;
	LINT LockLevel;

	
	/* Fill in values to be sent to mppLnx_BuildRdacinfo */
	ctlAddress.ChannelId = SDpnt->channel;
	ctlAddress.TargetId  = SDpnt->id;
	ctlAddress.HostId    = SDpnt->host->host_no;
	
	/* check if virtual device */
	if (strcmp(SDpnt->host->hostt->proc_name, MPPLNX_PROC_NAME) == 0) {

		//make sure you point to the right RdacINfo while locking and unlocking
		pRdacDevice = mpp_ModuleArray[(SDpnt->id)].RdacInfo;
		pRdacLun = &pRdacDevice->RdacLuns[SDpnt->lun];
		if ( !pRdacLun->PseudoLunObject ) {
			MPP_LOCK(pRdacDevice->Lock, LockLevel);
			pRdacLun->PseudoLunObject = SDpnt;
			MPP_UNLOCK(pRdacDevice->Lock, LockLevel);
		}

		return 	sg_old_attach(SDpnt);
	}

	/*
	* workaround for setting mppLnx_host_procName
	*/
	if(!is_mppLnx_host_procName_set)
	{
		mppLnx_host_procName = SDpnt->host->hostt->proc_name;
	}

	/* Call mppLnx_IsArrayDevice*/
        if ( mppLnx_IsArrayDevice((MPP_HANDLE)SDpnt, SDpnt->lun))
	{
		is_mppLnx_host_procName_set = 1;
		/* mppLnx_IsArrayDevice returns true if the device has already been discovered and is in the RdacInfo data structure */

		/* So if we come here it means that this Scsi_Device is a physical LUN and is part of an LSI array */

	
		/* Find out what kind of LSI device it is */
		if (( device_type = mppLnx_DeviceType(SDpnt)) ) {

			/* if it is a UTMLUN */
			if ( device_type == MPP_UTM_LUN_DEVICE )
				return sg_old_attach(SDpnt);		     // we want sg to claim this device
          
			/* It is a normal LSI physical lun*/
			return 0;
		}
                else
                {

			MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 805, "Is a LSI physical device but is an Unconfigured Device\n"));
			return 0;
                }

	}

	/* Not an LSI Device */
	return sg_old_attach(SDpnt);   // call original function

}



/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_DeviceType
* SUMMARY:  MPP function that goes through RdacInfo data structure to find if a device has already been discovered
* and to return the type of the device whether it is a UTM LUN or Root/Boot device or just a LSI physical device.
* SCOPE:    local
*
* DESCRIPTION: We need to send VPD Inquiry pages to the storage array to determine the type of a LUN presented to 
* the detect function of an Upper level driver. A Scsi_Device could have already been discovered by another device
* driver. In this case we do not want to send a VPD Page Inquiry as this would have already been done and is 
* expensive. So the solution is to go through the RdacInfo data structure and look for the device and check and see
* if it has already been discovered. The device can only be a Physical Device. We do not go through the entire list
* of luns on an array but instead index on the lun number. This Lun number is passed to mpp BuildRdacInfo from with* in which this function will be called. This function also returns the Device type of each device.

*
* RESTRICTIONS:
* 
* RETURNS: TRUE on finding a match for the device in the RdacInfo data structure, returns MPP_UTM_LUN_DEVICE or 
* MPP_ROOT_BOOT_DEVICE depending on devicetype else returns FALSE
*/

LWORD mppLnx_DeviceType (MPP_HANDLE SDpnt)  {
	int numModules = 0;
	ArrayModuleEntry_t *mppModule = NULL;
        RdacDeviceInformation_t *pRdacDevice = NULL;
        RdacLunInformation_t    *pRdacLun = NULL;
	int  LunCont = 0;
	int  LunPath = 0;
	LunPathInfo_t  *pLunPathInfo = NULL;
	LunPathObjects_t        *pLunPath;
	unsigned char			nullWWN[WWN_LENGTH];

	memset(nullWWN, 0, WWN_LENGTH);

	/*should we lock here ?  */
	/*PP_LOCK(RdacInfo->Lock, LockLevel);  */
	for ( mppModule     = mpp_ModuleArray, numModules = 0; 
                 numModules < mpp_MaxArrayModules; mppModule++, numModules++)
	{
	     
		if (! (pRdacDevice = mppModule->RdacInfo))
		{
			/* There is no array at this location as yet */
			continue;
		}
	      
		/* index on the Lun number passed to this function  */
		pRdacLun = &pRdacDevice->RdacLuns[SDpnt->lun];

		if ( !bcmp((BYTE *) pRdacLun->WWN, (BYTE *) nullWWN, 16) )
		{
			/* That means that memory for the struct has been allocated and bzeroed but the lun has not been inserted as yet */
			continue;   

		}
			       
		/* Go through each controller */
		for( LunCont = 0; LunCont < 2; LunCont++)
		{

			pLunPath = pRdacLun->LunControllers[LunCont];

			/* Go through each path */
			for( LunPath = 0; LunPath <  mpp_MaxPathsPerController; LunPath++)
			{
				pLunPathInfo = &pLunPath->LunPaths[LunPath];
				if (!pLunPathInfo->LunPathDevice ||     
					pLunPathInfo->PathRemoveState)
				{

					continue;
				}

				if ( pLunPathInfo->LunPathDevice == SDpnt)
				{
				   
					/* This is a physical device */

					MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, "mppLnx_DeviceType : Physical device 0x%x \n",  SDpnt));

					/* if it is a UTMLUN */
					if (pRdacLun->UTMLun )
					{
						/*MPP_UNLOCK(RdacInfo->Lock, LockLevel);*/

						return MPP_UTM_LUN_DEVICE;
					}
					
					return TRUE;

				}
			}
                 
		}

	}

	/* MPP_UNLOCK(RdacInfo->Lock, LockLevel); */
	
	return FALSE;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     LWORD mppLnx_IsArrayDevice(MPP_HANDLE DeviceVertex, LWORD Lun)
* SUMMARY:  Check whether or not the specified physical device is a storage array device without performing scsi io
* SCOPE:    local
*
* DESCRIPTION: This function checks whether or not the specified physical device is a storage array device based 
    RdacInfo data model or the un-configured device list. 
* @DeviceVertex-- A Scsi_Device object that represents the un-configured device
  @Lun -- the LUN number of the physical device object
* RESTRICTIONS:
* 
* RETURNS: TRUE if the physical device is a storage array device (either configured or un-configured device) or FALSE 
                   otherwise.
*/

LWORD 
mppLnx_IsArrayDevice(MPP_HANDLE DeviceVertex, LWORD Lun)
{
   LWORD  retValue = FALSE;
   mppLnx_UnconfiguredDeviceEntry_t * deviceEntry = NULL;

   if( mppSys_DeviceAlreadyDiscovered(DeviceVertex,Lun) )
   {
      retValue = TRUE;
   }
   else
   {
      deviceEntry = mppLnx_getUnconfiguredDevice(DeviceVertex);

      if(!deviceEntry)
      {
         retValue = FALSE;
      }
      else
      {
         retValue = TRUE;
      }
   }

   return retValue;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     LWORD mppLnx_BuildRdacinfo(MPP_HANDLE DeviceVertex, ControllerAddress_t *DeviceAddress, 
                     LWORD Lun)
* SUMMARY:  Inserts a physical device to the RdacInfo data model
* SCOPE:    local
*
* DESCRIPTION: This function is a wrap of the mpp_BuildRdacinfo() common function. This function calls 
mpp_BuildRdacinfo() to determine a device is a storage array device, a un-configured storage array device or not a 
storage array device. For a un-configured storage array device, it will add the device to un-configured device list and manages its state changes
*
* @DeviceVertex-- A Scsi_Device object that represents the un-configured device
  @DeviceAddress -- A ControllerAddress_t object which represents the physical connection of the path
  @Lun -- the LUN number of the physical device object
* RESTRICTIONS:
* 
* RETURNS: TRUE if the physical device is a storage array device (either configured or un-configured device) or FALSE 
                   otherwise.
*/

LWORD 
mppLnx_BuildRdacinfo(MPP_HANDLE DeviceVertex, ControllerAddress_t *DeviceAddress, 
                     LWORD Lun)
{
   mppLnx_UnconfiguredDeviceEntry_t * deviceEntry = NULL;
   LWORD  deviceType;
   LWORD  retValue = FALSE;

   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "Entering mppLnx_BuildRdacinfo H:%dC:%dT:%dL:%d\n",
      DeviceVertex->host->host_no,DeviceVertex->channel,DeviceVertex->id,DeviceVertex->lun));

   deviceEntry = mppLnx_getUnconfiguredDevice(DeviceVertex);


   deviceType = mpp_BuildRdacinfo(DeviceVertex,DeviceAddress, Lun);

   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "mppLnx_BuildRdacinfo H:%dC:%dT:%dL:%d type:%d\n",
      DeviceVertex->host->host_no,DeviceVertex->channel,
      DeviceVertex->id,DeviceVertex->lun,deviceType));

   switch (deviceType)
   {
      case TRUE:
         MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3,
            "mppLnx_BuildRdacinfo TRUE\n"));
         retValue = TRUE;
         /*it is a storage array device */
         if(deviceEntry)
         {
            /* it is in the unconfigured device list. We get here by hot-add */
            deviceEntry->isUnconfigured = FALSE;
         }
         break;
      case FALSE:
         MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3,
            "mppLnx_BuildRdacinfo FALSE\n"));
         /* it is not a storage array device */
         retValue = FALSE;
         break;
      case MPPCMN_UNCONFIGURED_LUN:
         /* it is a unconfigured device */
         MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3,
            "mppLnx_BuildRdacinfo MPPCMN_UNCONFIGURED_LUN\n"));
         if(!deviceEntry)
         {
            /* it is not in the un-configured device lis. Add it into the list */
            mppLnx_addUnconfiguredDevice(DeviceVertex);
         }
         retValue = TRUE;
         break;
	case MPPCMN_CONFIG_PARAMETER_OUT_OF_LIMIT:
         /* Max config parameter value exceeded add it to unconfigured list */
         MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3,
            "mppLnx_BuildRdacinfo MPPCMN_CONFIG_PARAMETER_OUT_OF_LIMIT\n"));
         if(!deviceEntry)
         {
            /* it is not in the un-configured device lis. Add it into the list */
            mppLnx_addUnconfiguredDevice(DeviceVertex);
         }
         retValue = TRUE;
         break;

      default:
         /* should not get here */
         break;
   }

   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "Exiting mppLnx_BuildRdacinfo H:%dC:%dT:%dL:%d  return %d\n",
      DeviceVertex->host->host_no,DeviceVertex->channel,
      DeviceVertex->id,DeviceVertex->lun,retValue));
   return retValue;

}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_getUnconfiguredDevice()
* SUMMARY:  Gets a un-configured device from the global un-configured device list
* SCOPE:    local
*
* DESCRIPTION: A un-configured device is a un-configured LUN that either reported as
 a configured device by the controller firmware or a dummy LUN created by a physical 
 HBA driver. The Linux MPP driver keeps a un-configured device list and treats 
 a un-configured device differently than a real-configured device. This function 
 queries a given Scsi_Deivce object from  the global un-configured device list
*
* @device -- A Scsi_Device object that represents the un-configured device
* RESTRICTIONS:
* 
* RETURNS: An mppLnx_UnconfiguredDeviceEntry_t object if the specified device is in
  the global un-configured device list or NULL otherwise.
*/
static mppLnx_UnconfiguredDeviceEntry_t * mppLnx_getUnconfiguredDevice(Scsi_Device * device)
{
   unsigned long iflages;
   mppLnx_UnconfiguredDeviceEntry_t * deviceEntry = NULL;
   mppLnx_UnconfiguredDeviceEntry_t * foundDevice = NULL;
   struct list_head *listp;

   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "Entering mppLnx_getUnconfiguredDevice H:%dC:%dT:%dL:%d\n",
      device->host->host_no,device->channel,device->id,device->lun));


   spin_lock_irqsave(unconfiguredDeviceList.unconfiguredDeviceLock, iflages);
   for(listp = unconfiguredDeviceList.deviceEntryList->list.next;
      listp != &(unconfiguredDeviceList.deviceEntryList->list);
      listp = listp->next)
   {
      deviceEntry = list_entry(listp, mppLnx_UnconfiguredDeviceEntry_t, list);
      if(deviceEntry->physical_device == device)
      {
         foundDevice = deviceEntry;
         break;
      }
   }
   spin_unlock_irqrestore(unconfiguredDeviceList.unconfiguredDeviceLock,iflages);
   
   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "Exiting mppLnx_getUnconfiguredDevice H:%dC:%dT:%dL:%d found?%d\n",
      device->host->host_no,device->channel,device->id,device->lun,
      foundDevice==NULL? 0:1));
   return foundDevice;
        
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_deleteUnconfiugredDevice()
* SUMMARY:  Deletes a un-configured device from the global un-configured device list
* SCOPE:    local
*
* DESCRIPTION: A un-configured device is a un-configured LUN that either reported as
 a configured device by the controller firmware or a dummy LUN created by a physical 
 HBA driver. The Linux MPP driver keeps a un-configured device list and treats 
 a un-configured device differently than a real-configured device. This function 
 deletes a given Scsi_Deivce object from  the global un-configured device list
*
* @device -- A Scsi_Device object that represents the un-configured device
* RESTRICTIONS:
* 
* RETURNS: 
*/

static void mppLnx_deleteUnconfiugredDevice(Scsi_Device * device)
{
   unsigned long iflages;
   mppLnx_UnconfiguredDeviceEntry_t * deviceEntry = NULL;

   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "Entering mppLnx_deleteUnconfiugredDevice H:%dC:%dT:%dL:%d\n",
      device->host->host_no,device->channel,device->id,device->lun));

   deviceEntry = mppLnx_getUnconfiguredDevice(device);
   if(deviceEntry)
   {
      spin_lock_irqsave(unconfiguredDeviceList.unconfiguredDeviceLock, iflages);
      list_del(&(deviceEntry->list));
      spin_unlock_irqrestore(unconfiguredDeviceList.unconfiguredDeviceLock,iflages);
      MPP_FREE(deviceEntry,sizeof(mppLnx_UnconfiguredDeviceEntry_t));
   }
   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "Exiting mppLnx_deleteUnconfiugredDevice H:%dC:%dT:%dL:%d\n",
      device->host->host_no,device->channel,device->id,device->lun));

   return;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_addUnconfiguredDevice
* SUMMARY:  Adds a un-configured device to the global un-configured device list
* SCOPE:    local
*
* DESCRIPTION: A un-configured device is a un-configured LUN that either reported as
 a configured device by the controller firmware or a dummy LUN created by a physical 
 HBA driver. The Linux MPP driver keeps a un-configured device list and treats 
 a un-configured device differently than a real-configured device. This function 
 adds a given Scsi_Deivce object to the global un-configured device list
*
* @device -- A Scsi_Device object that represents the un-configured device
* RESTRICTIONS:
* 
* RETURNS: 
*/
static void mppLnx_addUnconfiguredDevice(Scsi_Device * device)
{
   mppLnx_UnconfiguredDeviceEntry_t * deviceEntry = NULL;
   unsigned long iflages;
   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "Entering mppLnx_addUnconfiguredDevice H:%dC:%dT:%dL:%d\n",
      device->host->host_no,device->channel,device->id,device->lun));

   deviceEntry = mppLnx_getUnconfiguredDevice(device);
   if(!deviceEntry)
   {
      deviceEntry = (mppLnx_UnconfiguredDeviceEntry_t *) MPP_INT_KMEM_ALLOC(sizeof(mppLnx_UnconfiguredDeviceEntry_t));
      
      if(NULL == deviceEntry)
      {
         MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 807,
            "Cannot allocate memory for unconfigured device entry\n"));
         return;
      }

      deviceEntry->isUnconfigured = TRUE;
      deviceEntry->isAttached = FALSE;
      deviceEntry->isDetected = FALSE;
      deviceEntry->physical_device = device;
      spin_lock_irqsave(unconfiguredDeviceList.unconfiguredDeviceLock, iflages);
      list_add_tail(&(deviceEntry->list), &(unconfiguredDeviceList.deviceEntryList->list));
      spin_unlock_irqrestore(unconfiguredDeviceList.unconfiguredDeviceLock,iflages);
   }
   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "Exiting mppLnx_addUnconfiguredDevice H:%dC:%dT:%dL:%d\n",
      device->host->host_no,device->channel,device->id,device->lun));

   return;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_RescanUnconfiguredDevices
* SUMMARY:  Rescan un-configured devices in the global un-configured device list
* SCOPE:    local
*
* DESCRIPTION: A un-configured device is a un-configured LUN that either reported as
 a configured device by the controller firmware or a dummy LUN created by a physical 
 HBA driver. The Linux MPP driver keeps a un-configured device list and treats 
 a un-configured device differently than a real-configured device. This function 
 is invoked when an end user issues "mppUtil -s busscan". This function will go through
 each un-configured device in the un-configured device list to check whether or not
 it become configured. If it is, the device will be added to RdacInfo data model.
*
* @device -- A Scsi_Device object that represents the un-configured device
* RESTRICTIONS:
* 
* RETURNS: 
*/
void mppLnx_RescanUnconfiguredDevices()
{
   unsigned long iflages;
   mppLnx_UnconfiguredDeviceEntry_t * deviceEntry = NULL;
   struct list_head *listp;

   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "Entering mppLnx_RescanUnconfiguredDevices()\n"));

   listp = unconfiguredDeviceList.deviceEntryList->list.next;
   while(1)
   {
      spin_lock_irqsave(unconfiguredDeviceList.unconfiguredDeviceLock, iflages);
      if( listp != &(unconfiguredDeviceList.deviceEntryList->list) )
      {
         deviceEntry = list_entry(listp, mppLnx_UnconfiguredDeviceEntry_t, list);
         listp = listp->next;
         spin_unlock_irqrestore(unconfiguredDeviceList.unconfiguredDeviceLock,iflages);
         if(deviceEntry->isUnconfigured == TRUE)
         {
            /*
            * we found a un-configured device, feed it to all detect/attach functions
            */
            mppLnx_sd_detect(deviceEntry->physical_device);
            mppLnx_sg_detect(deviceEntry->physical_device);
            mppLnx_target_detect(deviceEntry->physical_device);
            mppLnx_sd_attach(deviceEntry->physical_device);
            mppLnx_sg_attach(deviceEntry->physical_device);
            mppLnx_target_attach(deviceEntry->physical_device);
         }
      }
      else
      {
         spin_unlock_irqrestore(unconfiguredDeviceList.unconfiguredDeviceLock,iflages);
         break;
      }
   }

   MPP_DEBUGPRINT((MPP_ATTACH_DEBUG+MPP_DEBUG_LEVEL_3, 
      "Exiting mppLnx_RescanUnconfiguredDevices\n"));
}
module_init (mppLnx_init);
module_exit (mppLnx_exit);
MODULE_AUTHOR("IBM");
MODULE_DESCRIPTION ("MPP upper level driver");
MODULE_LICENSE("GPL");

