/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************

NAME            mppLnx_vhba.c
SUMMARY         %description%
VERSION         %version: 58.1.1 %
UPDATE DATE     %date_modified: Fri Feb 27 08:58:52 2004 %
PROGRAMMER      %created_by:    yqi %

        Copyright 2003, LSI Logic Storage Systems, Inc.

DESCRIPTION: 
   The Linux MPP virtual HBA driver is a lower level driver in the Linux SCSI subsystem. 
   A lower level driver must implement Scsi_Device_Template data struct. This source 
   file contains all functions that are required by the Linux Scsi middle level. The 
   responsibilities of the functions in this file include:
    1. Initialization of the module
    2. Registration with the SCSI middle level
    3. Receive SCSI commands for any virtual devices from the middle level
    4. Provides all required function interfaces to the middle level
    5. Clearly clean system resources and unregister with the Kernel at the module unloading time.

INCLUDE FILES:

NOTES:

RESTRICTIONS:

SEE ALSO:

REFERENCE: 

  349-1033500 Linux MPP Driver Feature Architecture Specification
  349-1034610 Linux MPP Virtual HBA Driver Feature Design Document
  349-1034750 Linux MPP Upper Level Driver Feature Design Document
  349-1034730 Linux MPP Utility Feature Design Document
  349-1034620 Linux MPP Installation Feature Design Document

IMPLEMENTATION:

MODIFICATION HISTORY:

*******************************************************************************/

#define __SRCmppLnx_vhba_c

/***  Linux kernel INCLUDES  ***/  
#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/genhd.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/interrupt.h>

#include <asm/io.h>

#include <linux/blk.h>
#include "scsi.h"
#include "hosts.h"
#include "sd.h"

#include <linux/stat.h>

/***  MPP commomn API INCLUDES  ***/  
#include "MPP_Common.h"
#include "MPP_RdacInfo.h"
#include "MPP_ProtoTypes.h"
#include "MPP_Sysdep.h"
#include "mppLnx_version.h"
#include "mppCmn_s2tos3.h"

/***  Linux MPP Virtual HBA INCLUDES  ***/  
#include "mppLnx_vhba.h"
#include "mppLnx_shared.h"
/***  CONSTANT DEFINITIONS  ***/

/***  MACRO DEFINITIONS  ***/
#define         YES_NO(a)               a == 1 ? 'Y' : 'N'
#define         A_OR_B(a)               a == 0 ? 'A' : 'B'
/*
* A macro used in mppLnx_proc_info for maintaining write-buffer position.
*/
#define  VHBA_PROC_INFO_POS  {   \
         pos = pos + len;  \
         if(pos > (MPPLNX_PROC_INFO_BUFFSIZE -256)) \
         { \
            isTruncated = 1; \
            goto proc_info_output_stop; \
         } \
   };
/*
* Mpp status string used for debug-print
*/
char * mppStatusStrings[] = 
{
      "MPP_GOOD_STATUS",
      "MPP_CHECK_CONDITION",
      "MPP_SCSI_BUSY",
      "MPP_SCSI_RESERVATION",
      "MPP_UNSUPPORTED_SCSI_STATUS",
      "MPP_INVALID_REQUEST",
      "MPP_SELECTION_TIMEOUT",
      "MPP_HARDWARE_ERROR",
      "MPP_COMMAND_TIMEOUT",
      "MPP_COMMAND_ABORTED",
      "MPP_UNRECOGNIZED_OS_STATUS",
      "MPP_REQUEST_SENSE_FAILED"
};
/*
* mpp action string used for debug-print
*/
char * mppActionStrings[] = 
{
      "MPP_NO_ERROR",
      "MPP_RETURN_ERROR",
      "MPP_RETRY_DEC_COUNT",
      "MPP_RETRY_NO_DEC",
      "MPP_RETRY_ALTERNATE_CONTROLLER",
      "MPP_FAILOVER",
      "MPP_FAILOVER_CONTROLLER",
      "MPP_FAILOVER_TO_CURRENT"
};
/***  TYPE DEFINITIONS  ***/
/*
* scsi middle level's public API but not prototyped in a .h file
*/
extern int scsicam_bios_param(Disk * disk,	/* SCSI disk */
		       kdev_t dev,	/* Device major, minor */
		       int *ip /* Heads, sectors, cylinders in that order */ ); 
/***  LOCALS  ***/
/*
* proxy command timeout
* the value of a proxy command timeout is dependent on physical HBAs.
* We set the default value to 25 seconds.
* The value will be reset based on a physical HBA
*/
int mpLnx_proxy_cmnd_timeout;
static char * QLA_DRIVER_PROC_NAME_PREFIX     = "qla";

extern mppLnx_VHBAContext_t mppLnxVHBAContext;
/*
* the saved physical HBA error handlers
*/
static mppLnx_saved_PHBA_eh_handlers_t mppLnxSavedPHBAEhHandlers;

/*
* the tasklet queue related declaration
*/
#if USE_THREAD_FOR_DPC
static mppLnx_dpcQueueContext_t   mppLnxDpcQueueContext;
#else
static mppLnx_taskletQueue_t mppLnxTaskletQueue;
#endif

static spinlock_t taskQueue_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t syncIOQueue_lock = SPIN_LOCK_UNLOCKED;

#if USE_THREAD_FOR_DPC
DECLARE_MUTEX_LOCKED(mppLnx_dpc_sem);
#else
DECLARE_TASKLET (mppLnxTasklet, mppLnx_tasklet_func, (unsigned long) &mppLnxTaskletQueue);
#endif
/*
* fail back scan thread
*/
DECLARE_MUTEX_LOCKED(mppLnx_detect_sem);
DECLARE_MUTEX_LOCKED(mppLnx_failback_sem);
mppLnx_failbackScanContext_t mppLnxFailbackScanContext;

/*
* syncIO worker thread and its queue
*/
DECLARE_MUTEX_LOCKED(mppLnx_syncIOWorker_sem);
mppLnx_workerQueueContext_t mppLnxWorkerQueueContextContext;

/*
* The Scsi middle level does not make its scsi_done() function (lower level command
* completion function) and its time out function public. We must get them from a Scsi_Cmnd
* and save them for later use.
*/
void (*middle_level_scsi_done)(Scsi_Cmnd *) = NULL;
void (*middle_level_timeoutfunc)(Scsi_Cmnd *) = NULL;
/***  PROCEDURES  ***/
static void mppLnx_set_command_timeout(Scsi_Host_Template *pshtp);

/* 
* The virtual HBA driver initialization and exit. The Scsi middle level source file
* "scsi_module.c will take care of this lower level driver's initialization/registration
* and un-registration.
*/
#ifdef MODULE
   Scsi_Host_Template driver_template = MPPLNX_SCSI_HOST_TEMPLATE;
#include "scsi_module.c"

#else   /* new kernel scsi initialization scheme */
   static  Scsi_Host_Template driver_template = MPPLNX_SCSI_HOST_TEMPLATE;
#include "scsi_module.c"
#endif


/*******************************************************************************
* PROCEDURE
*
* NAME:      mppLnx_proc_info 
* SUMMARY:  Returns the virtual HBA's state information to the user space via
 /proc file system.
* SCOPE:    public
*
*
* DESCRIPTION:This function is an interface to the mpp virtual bus driver's proc 
   file /proc/scsi/mpp/n, where "n" is this virtual HBA driver's host number. The 
   host number is assigned by the Scsi middle level and is dependent on Scsi host 
   registration order.
   This function will be assigned to Scsi_Host_Template data struct's "proc_info" 
   function pointer.
   This function is invoked when an end user types "cat /proc/scsi/mpp/n".
   The virtual HBA driver will not take input from the /proc file. The function 
   reports the following information to the caller.
      Mpp Driver version information. This includes mpp driver version and driver
      build time
      Virtual HBA attributes
      Physical HBA driver name of each attached physical HBA
      Physical HBA host attributes such as supported maximum target id, supported
      maximum LUNs
      Each managed storage array attributes/states
*
*      @buffer: anchor point to output to (0==writeto1_read0) or fetch from
*               (1==writeto1_read0).
*      @start: where "interesting" data is written to. Ignored when
*              1==writeto1_read0.
*      @offset: offset within buffer 0==writeto1_read0 is actually
*               interested in. Ignored when 1==writeto1_read0 .
*      @length: maximum (or actual) extent of buffer
*      @host_no: host number of interest (struct Scsi_Host::host_no)
*      @writeto1_read0: 1 -> data coming from user space towards driver
*                            (e.g. "echo some_string > /proc/scsi/mpp/2")
*                       0 -> user what data from this driver
*                            (e.g. "cat /proc/scsi/mpp/2")
* RESTRICTIONS:
*        The virtual HBA driver will not take input from the /proc file.
* 
* RETURNS:number of chars output to buffer past offset.
*
* ERRNO:
*
* NOTES:Driven from drivers/scsi/scsi_proc.c which interfaces to proc_fs
* offset is 0 or greater than 0. But it should be less than length
*    |<------------             lenght     ------------------->|
*    ----------------------------------------------------------
*    |                            |                            |
*    ----------------------------------------------------------
*                               offset
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_proc_info
(
   char * buffer,
   char ** start,
   off_t offset,
   int length,
   int hostno,
   int writeto1_read0
)
{
   static char                   mppLnx_proc_info_buff[MPPLNX_PROC_INFO_BUFFSIZE];
   int                           len=0, begin=0, pos=0;
   int				               controllerCount, pathCount, lunCount;
	unsigned char			         nullWWN[WWN_LENGTH];
	RdacDeviceInformation_t 	   *rdacInfo;
	RdacControllerInformation_t	*controllerInfo;
	RdacControllerPath_t		      *controllerPath;
	RdacLunInformation_t		      *lunInfo;
	LunPathObjects_t		         *lunPath;
	LunPathInfo_t			         *lunPathInfo;
	char				               strBuf[16], strBuf2[16];
   char                          asciiWWN[34];
   int                           vTarget = 0;
   int                           isTruncated = 0;
   struct Scsi_Host              *this_SHp = NULL;

   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_proc_info() offset %d length %d \n",
      offset,length));
   memset(mppLnx_proc_info_buff, 0, MPPLNX_PROC_INFO_BUFFSIZE);
   /**
   * we do not support write to the proc
   */
	if (writeto1_read0)
   {
		
		return -EINVAL;
	}
  
	begin = 0;
	len = sprintf(&mppLnx_proc_info_buff[0], 
      "\n\nLinux MPP driver. Version:%s Build:%s\n",
	    MPPLNX_VERSION, MPPLNX_BUILDTIMESTAMP);
   /*
   * the virtula HBA Scsi_Host information
   */
   for(this_SHp = scsi_hostlist; this_SHp != NULL; this_SHp = this_SHp->next)
   {
      if(this_SHp->host_no == hostno)
      {
         break;
      }
   }
   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
      "the virtual HBA host information\n");

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       host number %d\n",this_SHp->host_no);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       base = %ld\n",    this_SHp->base);

   VHBA_PROC_INFO_POS;
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       can_queue = %d\n",    this_SHp->can_queue);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       cmd_per_lun = %d\n",    this_SHp->cmd_per_lun);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       dma_channel = %d\n",    this_SHp->dma_channel);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       extra_bytes = %d\n",    this_SHp->extra_bytes);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       host_blocked = %d\n",    this_SHp->host_blocked);
   
   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       host_self_blocked = %d\n",    this_SHp->host_self_blocked);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       host_failed = %d\n",    this_SHp->host_failed);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       in_recovery = %d\n",    this_SHp->in_recovery);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       io_port = %ld\n",    this_SHp->io_port);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       last_reset = %ld\n",    this_SHp->last_reset);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       loaded_as_module = %d\n",    this_SHp->loaded_as_module);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       max_channel = %d\n",    this_SHp->max_channel);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       max_channel = %d\n",    this_SHp->max_channel);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       max_cmd_len = %d\n",    this_SHp->max_cmd_len);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       max_lun = %d\n",    this_SHp->max_lun);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       max_id = %d\n",    this_SHp->max_id);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       max_sectors = %d\n",    this_SHp->max_sectors);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       sg_tablesize = %d\n",    this_SHp->sg_tablesize);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       this_id = %d\n",    this_SHp->this_id);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       unchecked_isa_dma = %d\n",    this_SHp->unchecked_isa_dma);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       unique_id = %d\n",    this_SHp->unique_id);

   VHBA_PROC_INFO_POS
   len = sprintf(&mppLnx_proc_info_buff[pos],
         "       use_clustering = %d\n\n",    this_SHp->use_clustering);

   /*
   * go through all storage arrays and print rdac info
   */
   for(vTarget = 0; vTarget < mppLnx_Config_Parameters.mpp_MaxArrayModules; vTarget++)
   {
      if( (rdacInfo = mpp_ModuleArray[vTarget].RdacInfo) == NULL)
      {
         MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
            "no storage array at virtual target %d\n",vTarget));
         continue;
      }
      MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
         "virtula target %d array %s\n",
         rdacInfo->VirtualTargetId, (char *)rdacInfo->ModuleName));
     
      memset(nullWWN, 0, WWN_LENGTH);

	   /**
	   ** Global array information
	   **/
      VHBA_PROC_INFO_POS
	   len = sprintf(&mppLnx_proc_info_buff[pos], 
         "MPP Information:\n");
	   
      VHBA_PROC_INFO_POS
	   len = sprintf(&mppLnx_proc_info_buff[pos],
         "----------------\n");

	   VHBA_PROC_INFO_POS
	   len = sprintf(&mppLnx_proc_info_buff[pos],
         "      ModuleName: %-32.32s         SingleController: %c\n",
		   (char *)rdacInfo->ModuleName, YES_NO(rdacInfo->SingleController));
	   
      VHBA_PROC_INFO_POS
	   len = sprintf(&mppLnx_proc_info_buff[pos],
         " VirtualTargetID: 0x%03x                                       ScanTriggered: %c\n",
		   rdacInfo->VirtualTargetId, YES_NO(rdacInfo->ScanTriggered));
	   
      VHBA_PROC_INFO_POS
	   len = sprintf(&mppLnx_proc_info_buff[pos],
         "     ObjectCount: 0x%03x                                          AVTEnabled: %c\n",
		   rdacInfo->ObjectCount, YES_NO(rdacInfo->AVTEnabled));
	   
     
      memset(asciiWWN,0,34);
      mpp_ConvertWWNtoAscii(&asciiWWN[0],rdacInfo->WWN);
      VHBA_PROC_INFO_POS
	   len = sprintf(&mppLnx_proc_info_buff[pos],
         "             WWN:%s\n", asciiWWN);
	   
      
	   VHBA_PROC_INFO_POS
	   len = sprintf(&mppLnx_proc_info_buff[pos],
         "                                                              RestoreCfg: %c\n", 
         YES_NO(rdacInfo->RestoreCfg));
	   

	   /**
	    ** Controller-specific information
	    **/
      //len = 0;
	   for(controllerCount = 0; controllerCount < 2; controllerCount++)
	   {
		   controllerInfo = rdacInfo->ControllerInfo[controllerCount];
		   if(!controllerInfo->ControllerPresent)
		   {
            //len = 0;
			   continue;
		   }

		   VHBA_PROC_INFO_POS
	      len = sprintf(&mppLnx_proc_info_buff[pos],
            "\nController '%c' Status:\n", controllerCount + 'A');

		   VHBA_PROC_INFO_POS
	      len = sprintf(&mppLnx_proc_info_buff[pos],
            "-----------------------\n");


		   VHBA_PROC_INFO_POS
	      len = sprintf(&mppLnx_proc_info_buff[pos],
            "ControllerPresent: %c\n",
		      YES_NO(controllerInfo->ControllerPresent));

		   if(controllerInfo->UTMLunExists)
		   {
			   VHBA_PROC_INFO_POS
	         len = sprintf(&mppLnx_proc_info_buff[pos],
               "    UTMLunExists: %c (%-3.3d)                                            Failed: %c\n",
				   YES_NO(controllerInfo->UTMLunExists), controllerInfo->UTMLunNumber,
				   YES_NO(controllerInfo->Failed));
		   }
		   else
		   {
			   VHBA_PROC_INFO_POS
	         len = sprintf(&mppLnx_proc_info_buff[pos],
               "    UTMLunExists: %c                                                  Failed: %c\n",
				   YES_NO(controllerInfo->UTMLunExists), YES_NO(controllerInfo->Failed));
		   }

		   VHBA_PROC_INFO_POS
	         len = sprintf(&mppLnx_proc_info_buff[pos],
               "   NumberOfPaths: %1d                                          FailoverInProg: %c\n",
			      controllerInfo->NumberOfPaths, YES_NO(controllerInfo->FailoverInProgress));

       
		   for(pathCount = 0; pathCount < controllerInfo->NumberOfPaths; pathCount++)
		   {
			   controllerPath = &controllerInfo->ControllerPath[pathCount];

			   VHBA_PROC_INFO_POS
	         len = sprintf(&mppLnx_proc_info_buff[pos],
               "\n    Path #%1d\n", pathCount + 1);

			   VHBA_PROC_INFO_POS
	         len = sprintf(&mppLnx_proc_info_buff[pos],
               "    ---------\n");

			   memset(strBuf,0,16);
            sprintf(strBuf,"h%dc%dt%d",
               controllerPath->Address.HostId,
               controllerPath->Address.ChannelId,
               controllerPath->Address.TargetId);
			   VHBA_PROC_INFO_POS
	         len = sprintf(&mppLnx_proc_info_buff[pos],
               " DirectoryVertex: %-10.10s                                        Present: %c\n", 
				   strBuf, YES_NO(controllerPath->Present));

			   VHBA_PROC_INFO_POS
	         len = sprintf(&mppLnx_proc_info_buff[pos],
               "                                                                     Failed: %c\n",
				   YES_NO(controllerPath->Failed));
		   }
	   }

	   /**
	   ** Lun-specific information
	   **/

	   VHBA_PROC_INFO_POS
	   len = sprintf(&mppLnx_proc_info_buff[pos],
            "\n\nLun Information\n");

	   VHBA_PROC_INFO_POS
	   len = sprintf(&mppLnx_proc_info_buff[pos],
           "---------------\n");
 
	   for(lunCount = 0; lunCount < mppLnx_Config_Parameters.mpp_MaxLunsPerArray; lunCount++)
	   {
		   lunInfo = &rdacInfo->RdacLuns[lunCount];
		   if(!lunInfo->Present || lunInfo->PseudoLunObject == NULL)
         {
			   continue;
		   }

		   if(lunCount != 0)
		   {
			   VHBA_PROC_INFO_POS
	         len = sprintf(&mppLnx_proc_info_buff[pos],
               "\n");
		   }

		   if(lunInfo->WWN[0] != 0)
		   {
            memset(asciiWWN,0,34);
            mpp_ConvertWWNtoAscii(&asciiWWN[0],lunInfo->WWN);
			   VHBA_PROC_INFO_POS
	         len = sprintf(&mppLnx_proc_info_buff[pos],
               "    Lun #%d - WWN:%s\n ",lunCount,asciiWWN);
			
		   }
		   else
		   {
			   VHBA_PROC_INFO_POS
	         len = sprintf(&mppLnx_proc_info_buff[pos],
               "    Lun #%d - WWN:\n", lunCount);
		   }
         
         MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
            " Lun %d\n",lunCount));

		   VHBA_PROC_INFO_POS
	      len = sprintf(&mppLnx_proc_info_buff[pos],
               "    ----------------\n");
         
         memset(strBuf, 0, 16);

         
         if(NULL == lunInfo->PseudoLunObject)
         {
            sprintf(strBuf,"Null");
         }
         else
         {
            MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
               "PseudoLunObject 0x%lx\n",(unsigned long)lunInfo->PseudoLunObject));
            sprintf(strBuf,"h%dc%dt%dl%d",
               lunInfo->PseudoLunObject->host->host_no,
               lunInfo->PseudoLunObject->channel,
               lunInfo->PseudoLunObject->id,
               lunInfo->PseudoLunObject->lun);

         }

		   VHBA_PROC_INFO_POS
         len = sprintf(&mppLnx_proc_info_buff[pos],
            "       LunObject: %-10.10s                              CurrentOwningPath: %c\n",
            strBuf, A_OR_B(lunInfo->CurrentOwningPath));

		   VHBA_PROC_INFO_POS
         len = sprintf(&mppLnx_proc_info_buff[pos],
            "  RemoveEligible: %c                                          BootOwningPath: %c\n",
			   YES_NO(lunInfo->PseudoRemoveEligible), 
            A_OR_B(lunInfo->BootOwningPath));

		   VHBA_PROC_INFO_POS
         len = sprintf(&mppLnx_proc_info_buff[pos],
            "   NotConfigured: %c                                           PreferredPath: %c\n",
			   YES_NO(lunInfo->NotConfigured),
            A_OR_B(lunInfo->PreferredPath));

         VHBA_PROC_INFO_POS
         len = sprintf(&mppLnx_proc_info_buff[pos],
            "        DevState: %d                     ReportedPresent: %c\n",
			   lunInfo->PseudoDevState, YES_NO(lunInfo->ReportedPresent));


		   VHBA_PROC_INFO_POS
         len = sprintf(&mppLnx_proc_info_buff[pos],
            "    PrevDevState: %d                     ReportedMissing: %c\n\n",
			   lunInfo->PseudoPreviousDevState, YES_NO(lunInfo->ReportedMissing));

         MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
            " Lun info is inserted %d\n",lunCount));

		   //len = 0;
         for(controllerCount = 0; controllerCount < 2; controllerCount++)
		   {
			   if(!lunInfo->LunControllers[controllerCount])
			   {
				   continue;
			   }
			   lunPath = lunInfo->LunControllers[controllerCount];
			   
            VHBA_PROC_INFO_POS
            len = sprintf(&mppLnx_proc_info_buff[pos],
               "    Controller '%c' Path\n", controllerCount + 'A');

			   VHBA_PROC_INFO_POS
            len = sprintf(&mppLnx_proc_info_buff[pos],
               "    --------------------\n");

			   VHBA_PROC_INFO_POS
            len = sprintf(&mppLnx_proc_info_buff[pos],
               "   NumLunObjects: %1d                                         RoundRobinIndex: %1d\n",
				   lunPath->NumberOfLunObjects, lunPath->RoundRobinPathIndex);

            MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
               "   NumLunObjects: %1d                                         RoundRobinIndex: %1d\n",
				   lunPath->NumberOfLunObjects, lunPath->RoundRobinPathIndex));

            for(pathCount = 0; pathCount < lunPath->NumberOfLunObjects; pathCount++)
			   {
				   lunPathInfo = &lunPath->LunPaths[pathCount];
				   if(!lunPathInfo->LunPathDevice)
				   {
                  MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
                     "controller %d path %d is not available\n",
                     controllerCount,pathCount));

					   continue;
				   }

               MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
                  "h%dc%dt%dl%d",
                  lunPathInfo->LunPathDevice->host->host_no,
                  lunPathInfo->LunPathDevice->channel,
                  lunPathInfo->LunPathDevice->id,
                  lunPathInfo->LunPathDevice->lun));

               memset(strBuf, 0 , 16);
               sprintf(strBuf,"h%dc%dt%dl%d",
                  lunPathInfo->LunPathDevice->host->host_no,
                  lunPathInfo->LunPathDevice->channel,
                  lunPathInfo->LunPathDevice->id,
                  lunPathInfo->LunPathDevice->lun);
			      VHBA_PROC_INFO_POS
               len = sprintf(&mppLnx_proc_info_buff[pos],
                  "   PathRemoveState: 0x%x    PathStartState: 0x%x    PathPowerState: 0x%x\n", 
                  lunPathInfo->PathRemoveState, 
                  lunPathInfo->PathStartState, 
                  lunPathInfo->PathPowerState);

				   
               memset(strBuf2, 0 , 16);
               if(NULL == lunPathInfo->UtmLunDevice)
               {
                  sprintf(strBuf2,"Un-available");
               }
               else
               {
                  sprintf(strBuf2,"h%dc%dt%dl%d",
                        lunPathInfo->UtmLunDevice->host->host_no,
                        lunPathInfo->UtmLunDevice->channel,
                        lunPathInfo->UtmLunDevice->id,
                        lunPathInfo->UtmLunDevice->lun);
               }
                     
					VHBA_PROC_INFO_POS
               len = sprintf(&mppLnx_proc_info_buff[pos],
                     "         Path #%1d: LunPathDevice: %-12.12s           UTMLunDevice: %-10.10s\n",
						   pathCount + 1, strBuf, strBuf2);
				   

				   VHBA_PROC_INFO_POS
               len = sprintf(&mppLnx_proc_info_buff[pos],
                     "                        IoCount: %d\n", lunPathInfo->IoCount);
            } /* for path Count */

			   VHBA_PROC_INFO_POS
            len = sprintf(&mppLnx_proc_info_buff[pos],"\n");

         } /*for controllerCount */

	   }/*for lunCount*/
   }/*for vTarget*/
   /*
   * In case the output length is greater than our pre-allocated buffer size,
   * terminate the processing.
   */
proc_info_output_stop:

   if(isTruncated)
   {
      pos = pos +len;
      len = sprintf(&mppLnx_proc_info_buff[pos],
         "\nOutput is too long. Truncated......\n");
      pos = pos + len;
   }else
   {
      pos = pos +len;
      len = sprintf(&mppLnx_proc_info_buff[pos],
         "\n---------End of the message-------------\n");
      pos = pos + len;
   }

   /*
   * the pos is the valid data length of our working buffer.
   */
	if (pos < offset) {
		len = 0;
		begin = pos;
      *start = NULL;
	}
   else
   {
      begin = offset;
      len = pos - offset;
      if(len >length)
      {
         len = length;
      }
      *start = buffer;
   }
  
   /*
   * copy our data from our working buffer to the buffer
   */
   if(len >0 )
   {
      memcpy(
         buffer,
         &mppLnx_proc_info_buff[begin],
         len);
   }
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_proc_info() buffer 0xlx offset %d begin %d pos %d len %d\n",
      (unsigned long)buffer,offset,begin,pos,len));
	return (len);

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_detect
* SUMMARY:  Detect virtual HBA host and initialize the virtual HBA driver's 
            internal
* SCOPE:    public 
*
* DESCRIPTION:
   This function is assigned to Scsi_Host_Template data struct's "int (*detect)"
   function pointer. It is the first function of the virtual HBA driver being called
   by the middle level. The mppLnx_detect() function has the following responsibilities:
      Initialize un-initialized attributes. At the virtual HBA driver initialization time, 
         certain attributes in the Scsi_Host_Template data struct cannot be initialized 
         because those attributes are dependent on attributes of a physical HBA driver.
         This function will initialize all un-initialized virtual HBA attributes by 
         assigning corresponding attributes from a physical HBA.
      Register Scsi_Host to the middle level. The virtual HBA has only one virtual host.
      Initialize the virtual HBA internal data model
      Initialize the virtual HBA tasklet
      Initialize and create device topology tree in /proc file system
      Create failback rescan kernel thread.
      Create SyncIO worker thread.
  This function always return "1" to indicate that there is one virtual HBA associated 
  with the virtual HBA driver.
*
*  @shtp: host template for this driver.
* RESTRICTIONS:
* 
* RETURNS: returns 1 to indicate that there is one virtual HBA instance. 0 if the 
*          Linux MPP upper level driver is not loaded yet.
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_detect(Scsi_Host_Template *shtp)
{
   struct Scsi_Host * vshp;
   Scsi_Host_Template *pshtp = NULL;
   mppLnx_PhostEntry_t *phostEntry;
   struct Scsi_Host * pshp;
   
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_1,"Entering mppLnx_detect()\n"));
  
   /*
   *check if the upper level driver is loaded. If not, don't try
   *to load this driver
   */
   if(mpp_ModuleArray == NULL)
   {
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL,400, 
         "mpp:vhba:mppLnx_detect: the upper level driver is not loaded\n")); 
      printk("The mpp virtual HBA driver requires a mpp Upper lelel driver to be presented\n");
      return 0;
   }
   /*
   * initialize internal data struct mppLnx_VHBAContext and create
   * proc node
   */
   mppLnx_initVHBAContext();
   
   /*
   * initialize certain attributes of this driver's template. They are dependent on the
   * physical HBA driver's attributes.
   */
    
   /*
   * check whether or not there are LSI devices. 
   * 
   */
   if(NULL == mppLnxVHBAContext.phostList)
   {
      /*
      * We get here beause
      * 1. The host may not have a HBA adapter
      * 2. The HBA driver may not be loaded yet
      * 3. The HBAs may not connect to LSI storage arrays
      * 4. The HBAs don't see any LSI LUNs
      * we can not initialize ourself correctly if there is not physical HBA(s) that 
      * has discovered our devices. Sorry I have to leave. Bye.
      */
      pshtp = NULL;
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL,400, 
         "mpp:vhba:mppLnx_detect: no LSI device is discovered. Leaving ...\n")); 
      printk("The mpp virtual HBA driver didn't discover LSI devices. Leaving ...\n");
      goto detect_errorout;

   }
   phostEntry = list_entry(mppLnxVHBAContext.phostList->list.next,
                  mppLnx_PhostEntry_t, list);
  
   pshtp = phostEntry->host->hostt;
   pshp = phostEntry->host;
   shtp->can_queue = pshtp->can_queue;
   shtp->sg_tablesize = pshtp->sg_tablesize;
   shtp->cmd_per_lun = pshtp->cmd_per_lun;
   shtp->use_clustering = pshtp->use_clustering;
   shtp->max_sectors = pshtp->max_sectors;
   shtp->this_id = 255;
   //if set highmem_io, we must take care of pci_dev
   //shtp->highmem_io = pshtp->highmem_io;
   //shtp->can_do_varyio = pshtp->can_do_varyio;
   
   /*
   * save and replace physical HBA driver's error handling functions
   * don't forget to restore them back at our module unload time
   */
   mppLnxSavedPHBAEhHandlers.eh_device_reset_handler = pshtp->eh_device_reset_handler;
   mppLnxSavedPHBAEhHandlers.eh_bus_reset_handler = pshtp->eh_bus_reset_handler;
   mppLnxSavedPHBAEhHandlers.eh_host_reset_handler = pshtp->eh_host_reset_handler;
   /*
   pshtp->eh_device_reset_handler = mppLnx_PHBA_eh_device_reset_handler;
   pshtp->eh_bus_reset_handler = mppLnx_PHBA_eh_bus_reset_handler;
   pshtp->eh_host_reset_handler = mppLnx_PHBA_eh_host_reset_handler;
   */


   /*
   * setting proxy command time out value
   */
   mppLnx_set_command_timeout(pshtp);
   /*
   * register the virtual host with the middle level
   */
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Detecting registering host\n"));
   vshp = scsi_register(shtp,0);
   if(vshp == NULL)
   {
      /*
      *The middle level cannot create the virtual host. It may be out of memory
      */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 404,
         "Failed to register the virtual hba host with the SCSI middle level. Exiting ...\n"));
      goto detect_errorout;
   }
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Complete registring host\n"));
#if USE_THREAD_FOR_DPC
   /*
   * initialize dpc queue
   */
   mppLnxDpcQueueContext.dpc_sem = &mppLnx_dpc_sem;
   mppLnxDpcQueueContext.detect_sem = &mppLnx_detect_sem;
   mppLnxDpcQueueContext.cmndQueueLock = &taskQueue_lock;
   atomic_set(&mppLnxDpcQueueContext.needSchedule, 1);
   /*
   * allocate an empty list head
   */
   mppLnxDpcQueueContext.cmndEntry = kmalloc(sizeof(mppLnx_CmndEntry_t), GFP_ATOMIC);
   if(NULL == mppLnxDpcQueueContext.cmndEntry)
   {
      /* memory error*/
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 402,
         "Cannot get kernel memory for mppLnx_CmndEntry_t\n"));
      goto detect_errorout;
   }
   memset(mppLnxDpcQueueContext.cmndEntry, 0, sizeof(mppLnx_CmndEntry_t));
   /*
   * initialize the list header
   */
   INIT_LIST_HEAD(&(mppLnxDpcQueueContext.cmndEntry->list));
   
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_1,
      "Starting dpc thread.\n"));
	kernel_thread((int (*)(void *))mppLnx_dpc_handler, (void *) vshp, 0); 
	/*
	 * Now wait for the kernel dpc thread to initialize
	 * and go to sleep.
	 */   
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_1,
      "Detecting sleeping. Waiting for dpc thread finishing init.\n"));
	down(&mppLnx_detect_sem);
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_1,
      "Detecting wakes up.\n"));
#else
       /*
   * initialize tasklet queue
   */
   mppLnxTaskletQueue.mppTasklet = &(mppLnxTasklet); 
   mppLnxTaskletQueue.cmndQueueLock = &taskQueue_lock;
   atomic_set(&mppLnxTaskletQueue.needSchedule, 1);
   /*
   * allocate an empty list head
   */
   mppLnxTaskletQueue.cmndEntry = kmalloc(sizeof(mppLnx_CmndEntry_t), GFP_ATOMIC);
   if(NULL == mppLnxTaskletQueue.cmndEntry)
   {
      /* memory error*/
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 402,
         "Cannot get kernel memory for mppLnx_CmndEntry_t\n"));
      goto detect_errorout;
   }
   memset(mppLnxTaskletQueue.cmndEntry, 0, sizeof(mppLnx_CmndEntry_t));
   /*
   * initialize the list header
   */
   INIT_LIST_HEAD(&(mppLnxTaskletQueue.cmndEntry->list));
#endif

   /*
   *syncIO worker thread queue initialization
   */
   mppLnxWorkerQueueContextContext.cmndQueueLock = &syncIOQueue_lock;
   /*
   * allocate an empty list head
   */
   mppLnxWorkerQueueContextContext.cmndEntry = kmalloc(sizeof(mppLnx_CmndEntry_t), GFP_ATOMIC);
   if(NULL == mppLnxWorkerQueueContextContext.cmndEntry)
   {
      /* memory error*/
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 403,
         "Cannot get kernel memory for mppLnx_CmndEntry_t\n"));
      goto detect_errorout;
   }
   memset(mppLnxWorkerQueueContextContext.cmndEntry, 0, sizeof(mppLnx_CmndEntry_t));
   /*
   * initialize the list header
   */
   INIT_LIST_HEAD(&(mppLnxWorkerQueueContextContext.cmndEntry->list));

   /*
   * The following attributes in the Scsi_Host struct need to be set
   */
   vshp->max_channel = 0;
   vshp->max_id =  mppLnx_Config_Parameters.mpp_MaxArrayModules;
   vshp->max_lun = mppLnx_Config_Parameters.mpp_MaxLunsPerArray;
   vshp->can_queue = pshp->can_queue;
   /*
   * inter-module communication registration
   */
   mppLnx_regSymbols();
   
   /*
   * start the failback scan thread
   */
   
   mppLnxFailbackScanContext.failback_sem = &mppLnx_failback_sem;
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_1,
      "Starting failback scan thread.\n"));
   mppLnxFailbackScanContext.detect_sem = &mppLnx_detect_sem;
	kernel_thread((int (*)(void *))mppLnx_failback_handler, (void *) vshp, 0);  
	/*
	 * Now wait for the kernel dpc thread to initialize
	 * and go to sleep.
	 */
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_1,
      "Detecting sleeping. Waiting for failback thread finishing init.\n"));
	down(&mppLnx_detect_sem);
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_1,
      "Detecting wakes up.\n"));
   
   /*
   * start the syncIO worker thread
   */
   mppLnxWorkerQueueContextContext.workerQueue_sem = &mppLnx_syncIOWorker_sem;
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_1,
      "Starting syncIO worker thread.\n"));
   mppLnxWorkerQueueContextContext.detect_sem = &mppLnx_detect_sem;
	kernel_thread((int (*)(void *))mppLnx_worker_handler, (void *) vshp, 0); 
	/*
	 * Now wait for the kernel dpc thread to initialize
	 * and go to sleep.
	 */   
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_1,
      "Detecting sleeping. Waiting for syncIO thread finishing init.\n"));
	down(&mppLnx_detect_sem);
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_1,
      "Detecting wakes up.\n"));

   /*
   * dump Scsi_Host and Scsi_Host_Template info if debug
   */
   MPPLNX_DUMP_SCSI_HOST(vshp);
   MPPLNX_DUMP_SCSI_HOST(phostEntry->host);
   MPPLNX_DUMP_SCSI_HOST_TEMPLATE(shtp);
   MPPLNX_DUMP_SCSI_HOST_TEMPLATE(pshtp);

   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_detect()\n"));

   return 1;

   /*error out
   * we need to clean up all resources
   */
detect_errorout:
   
   /* restore physical HBA error handdler
   */
   if(pshtp != NULL)
   {
      pshtp->eh_device_reset_handler = mppLnxSavedPHBAEhHandlers.eh_device_reset_handler;
      pshtp->eh_bus_reset_handler = mppLnxSavedPHBAEhHandlers.eh_bus_reset_handler;
      pshtp->eh_host_reset_handler = mppLnxSavedPHBAEhHandlers.eh_host_reset_handler;
   }

#if USE_THREAD_FOR_DPC
   if(NULL != mppLnxDpcQueueContext.cmndEntry)
   {
      kfree(mppLnxDpcQueueContext.cmndEntry);
   }
#else
   if(NULL != mppLnxTaskletQueue.cmndEntry)
   {
      kfree(mppLnxTaskletQueue.cmndEntry);
   }
#endif

   if(NULL != mppLnxWorkerQueueContextContext.cmndEntry)
   {
      kfree(mppLnxWorkerQueueContextContext.cmndEntry);
   }

   /*
   * clean up the internal data model
   */
   mppLnx_cleanVHBAContext();

   return 0;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_revoke
* SUMMARY:  deletes internal resouce of the specified Scsi_Device object.
* SCOPE:    public
*
*
* DESCRIPTION:
   This function is assigned to Scsi_Host_Template data struct's "int (*revoke)" 
   function pointer. This function is called by the middle level when a Scsi device is
   removed. It gives HBA driver a chance  to clean up its internal data structure. 
   The Linux MPP virtual HBA driver uses this function to clean up its /proc file 
   system when a virtual device is removed.

* @sdp The Sics_Device will be deleted.
*
* RESTRICTIONS:
* 
* RETURNS: Returns 1
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_revoke(Scsi_Device * sdp)
{
   /*this function is the virtual hba's revoke function. The passed-in
   * scsi device must be a virtual LUN
   */
   /*
   * remove from its virtual device linked list, delete its proc file and
   * free its memory
   */
   mppLnx_VDeviceEntry_t *vde;
   vde = mppLnx_get_VdeviceEntry(sdp);
   if(vde == NULL)
   {
      MPPLNX_VHBA_BUG();
   }
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_revoke()\n"));
   mppLnx_delete_VdeviceEntry(vde);
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_revoke()\n"));
   return 1;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_release
* SUMMARY:  release all resources associated with given host
* SCOPE:    public 
*
*
* DESCRIPTION:
   This function is assigned to Scsi_Host_Template data struct's "int(*release)"
   function pointer. It is invoked from the middle level's scsi_unregister_host()
   function at the time of unloading the driver as a module.
   The mppLnx_release() function has the following responsibilities:
      Un-register the virtual HBA host
      Terminate failback kernel thread
      Free up dynamically allocated memory and other system resources.
*  Locks: lock_kernel() active on entry and expected to be active
*  on return.
*
*  Notes: Invoked from mid level's scsi_unregister_host().
*  This function should call scsi_unregister(shp) [found in drivers/scsi/hosts.c]
*  prior to returning.
*  @shp The Scsi_Host will be released.
* RESTRICTIONS:
* 
* RETURNS: returned value ignored by the middle level
*
* ERRNO:
*
* NOTES:Invoked from mid level's scsi_unregister_host().
*      This function should call scsi_unregister(shp) [found in driver/scsi/hosts.c]
 *     prior to returning.
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_release(struct Scsi_Host * shp)
{
   Scsi_Host_Template      * pshp;
   mppLnx_PhostEntry_t     * phostEntry = NULL;
#if USE_THREAD_FOR_DPC != 1
   unsigned long           iflages;
#endif

   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_release() release hostNo %d\n",
      shp->host_no));
   /*
   * restore physical HBA Scsi_Host_Template's error handlers
   */  
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "restore physical HBA driver's error handlers\n"));

   phostEntry = list_entry(mppLnxVHBAContext.phostList->list.next,
      mppLnx_PhostEntry_t, list);

   if(NULL == phostEntry)
   {
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 405,
         "Failed to access physical HBA Scsi_Host_Template object for un-registering its error handlers.\n"));
   }
   else
   {
      pshp =phostEntry->host->hostt;
      pshp->eh_device_reset_handler = mppLnxSavedPHBAEhHandlers.eh_device_reset_handler;
      pshp->eh_bus_reset_handler = mppLnxSavedPHBAEhHandlers.eh_bus_reset_handler;
      pshp->eh_host_reset_handler = mppLnxSavedPHBAEhHandlers.eh_host_reset_handler;
      MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
         "done with restoring physical HBA driver's error handlers\n"));
   }
   
   /*
   *stop the failback rescan thread if it is still running.
   *(it may be killed by someone)
   */ 
   if(mppLnxFailbackScanContext.failback_process != NULL)
   {
	   send_sig(SIGHUP, mppLnxFailbackScanContext.failback_process, 1);
   }
   if(mppLnxFailbackScanContext.detect_sem != NULL)
   {
      down(mppLnxFailbackScanContext.detect_sem);
   }
      
	
   mppLnxFailbackScanContext.detect_sem = NULL;
   /*
   *stop the sync IO worker thread
   */ 
   if(mppLnxWorkerQueueContextContext.workerQueue_process != NULL)
   {
      send_sig(SIGHUP, mppLnxWorkerQueueContextContext.workerQueue_process, 1);
   }
   if(mppLnxWorkerQueueContextContext.detect_sem != NULL)
   {
	   down(mppLnxWorkerQueueContextContext.detect_sem);
   }
	mppLnxWorkerQueueContextContext.detect_sem = NULL;
   

   /*
   * take care of the tasklet queue.
   * If some commands are still in the task queue, we need to complete them all
   * 1. lock the queue list
   * 2. complete all command (if any) in the queue
   * 3. disable the tasklet
   * 4. release lock
   */
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "de-queue all commands in the taskqueue\n"));

#if USE_THREAD_FOR_DPC
   /*
   *stop the dpc thread
   */ 
   if(mppLnxDpcQueueContext.dpcQueue_process != NULL)
   {
      send_sig(SIGHUP, mppLnxDpcQueueContext.dpcQueue_process, 1);
   }
   if(mppLnxDpcQueueContext.detect_sem != NULL)
   {
	   down(mppLnxDpcQueueContext.detect_sem);
   }
	mppLnxDpcQueueContext.detect_sem = NULL;
   
#else
   mppLnx_tasklet_func((unsigned long) &mppLnxTaskletQueue);
   spin_lock_irqsave(mppLnxTaskletQueue.cmndQueueLock, iflages);
   tasklet_disable(mppLnxTaskletQueue.mppTasklet);
   mppLnxTaskletQueue.mppTasklet = NULL;
   kfree(mppLnxTaskletQueue.cmndEntry);
   spin_unlock_irqrestore(mppLnxTaskletQueue.cmndQueueLock, iflages);
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "Done with de-queue all commands in the taskqueue\n"));

   /*clean up syncIO worker thread's queue
   */
   spin_lock_irqsave(mppLnxWorkerQueueContextContext.cmndQueueLock, iflages);
   kfree(mppLnxWorkerQueueContextContext.cmndEntry);
   spin_unlock_irqrestore(mppLnxWorkerQueueContextContext.cmndQueueLock, iflages);
#endif
   /*
   * unregister this host
   */
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "Unregister the host with the middle level \n"));
   scsi_unregister(shp);
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "Done with Unregister the host with the middle level \n"));

   /*
   * clean up internal data model
   */
   mppLnx_cleanVHBAContext();
   /*
   * inter-module communication unregistration
   */
   mppLnx_unregSymbols();
   
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_release()\n"));
   return 1;

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_info
* SUMMARY:  supply information about given host: driver name plus data
 *          to distinguish given host
* SCOPE:    public 
*
* DESCRIPTION:
   This function is assigned to Scsi_Host_Template data struct's "char (*info)" 
   function pointer. The information returned from this function will be used by
   the middle level in a number places. It will go to system boot message file.
   The returned string must be only one line.
   This function will return the following string:
	"mpp virtual host bus adapter: version: xx.xx.xx.xx, timestamp:xxxxxxxxx"

*
* RESTRICTIONS:
* 
* RETURNS: the info char string of this driver. 
*
* ERRNO:
*
* NOTES:It is assumed the returned information fits on one line 
*      (i.e. does not included embedded newlines).
*      The SCSI_IOCTL_PROBE_HOST ioctl yields the string returned by this
*      function.
*      In a similar manner, scsi_register_host() outputs to the console
*      each host's "info" (or name) for the driver it is registering.
*     .
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
const char * mppLnx_info(struct Scsi_Host * shp)
{
   static char infoBuf[1024];
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_info()\n"));
   memset(infoBuf,0,1024);
   sprintf(infoBuf,"%s:version:%s,%s:%s",
         MPPVHBASTRING,MPPLNX_VERSION,TIME_STAMP,MPPLNX_BUILDTIMESTAMP);
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_info()\n"));
   return infoBuf;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_queuecommand
* SUMMARY:  queue scsi command to the virtual HBA driver, invoke 'done' on completion
* SCOPE:    public 
*
* DESCRIPTION:
   This function is assigned to Scsi_Host_Template data struct's "int (*queuecommand)" 
   function pointer. This function receives SCSI command from the Linux SCSI middle 
   level for virtual devices. The "void (*done)" function pointer is the command 
   completion function. This function must be returned before the "done" is called.

*  @vSCp: pointer to a scsi command object
*  @done: function pointer to be invoked on completion
*
*
* RESTRICTIONS:
* 
* RETURNS:1 if the virtual HBA adapter (host) is busy, else returns 0. 
*		One reason for an adapter to be busy is that the number
*		of outstanding queued commands is already equal to
*		struct Scsi_Host::can_queue. If the virtual HBA driver can not
*       allocate Scsi_Cmnd from all its available physical devices, returns 1.
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_queuecommand(Scsi_Cmnd * vSCp, void (*done)(Scsi_Cmnd *))
{
   int                  vchannel = vSCp->channel;
   int                  targetId = vSCp->target;
   int                  lun      = vSCp->lun;
   unsigned char        *cmd = (unsigned char *) vSCp->cmnd;
   void                 (*timeoutfunc) (Scsi_Cmnd *);
   SenseData_t          *senseData;

   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_queuecommand()\n"));
   
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_queuecommand received cmnd %lx serial_number %ld\n",
      (unsigned long)vSCp, vSCp->serial_number));
   
   vSCp->scsi_done = done;
   /*
   * validate the command
   */
   if( targetId >= mppLnx_Config_Parameters.mpp_MaxArrayModules ||
      (vchannel != 0) ||    /* no such channel */
      (mpp_ModuleArray[targetId].RdacInfo == NULL) || /*no such tid */
      /* no such LUN */
      (lun >= mppLnx_Config_Parameters.mpp_MaxLunsPerArray) ||
      (!mpp_ModuleArray[targetId].RdacInfo->RdacLuns[lun].Present)||
      (mpp_ModuleArray[targetId].RdacInfo->RdacLuns[lun].UTMLun)
     )
   {
      /*schedule the tasklet to perform the response and return */
      /* the result in the vSCp will be DID_NO_CONNECT*/
      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
         "mppLnx_queuecommand: No such device vtarget %d vlun %d\n",targetId, lun));
      if(targetId >= mppLnx_Config_Parameters.mpp_MaxArrayModules)
      {
         MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
            "mppLnx_queuecommand() return DID_NO_CONNECT because of id is out of range %d\n",
            targetId));
      }
      else if(mpp_ModuleArray[targetId].RdacInfo == NULL)
      {
         MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
            "mppLnx_queuecommand() return DID_NO_CONNECT because of rdacinfo == NULL %d\n",
            targetId));
      }
      else if(lun >= mppLnx_Config_Parameters.mpp_MaxLunsPerArray)
      {
         MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
            "mppLnx_queuecommand() return DID_NO_CONNECT because Lun is out off rang id%d lun%d\n",
            targetId,lun));
      }
      else if(!mpp_ModuleArray[targetId].RdacInfo->RdacLuns[lun].Present)
      {
         MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
            "mppLnx_queuecommand() return DID_NO_CONNECT because Lun is not present id%d lun%d\n",
            targetId,lun));
      }
      else if(mpp_ModuleArray[targetId].RdacInfo->RdacLuns[lun].UTMLun)
      {
         MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
            "mppLnx_queuecommand() return DID_NO_CONNECT because the Lun is UTM id%d lun%d\n",
            targetId,lun));
      }



      vSCp->result = DID_NO_CONNECT<<16;
      mppLnx_schedule_resp(vSCp);
      return 0;
   }
   
   if(done == NULL)
   {
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 406,
         "mppLnx_queuecommand NULL done function target %d lun %d\n",
         targetId,lun));
      return 0;
   }
   else if(middle_level_scsi_done == NULL)
   {
      /**
      * save the middle level scsi done function
      */
      middle_level_scsi_done = done;
   }
   /*
   *handling special IO 
   */
   switch (*cmd) 
   {
      case RESERVE:
      case RELEASE:
      case RESERVE_10:
      case RELEASE_10:
         /* 
         * what do we do if the command is targeted to a pre-Sonoran 4 array. FIX-ME
         * we need a decision.
         */
         if(!mpp_ModuleArray[targetId].RdacInfo->FWSonoran4OrLater)
         {
            /*
            * The array is a pre-Sonoran 4 array. It does not support SCSI-3
            * persistent reservation. We return an error back
            * status - check condition
            * sense key - illegal request
            * ASC/ASCQ 20/00 
            */
            MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 406,
               "Scsi-2 reserve/release command to a pre-Sonoran4 array %s\n",
               (char *)mpp_ModuleArray[targetId].ModuleName));
            senseData = (SenseData_t *) vSCp->sense_buffer;
            senseData->Error_Code = 0x70; /* 
                                           *Error type: current, 
                                           *sense data format:fixed 
                                           */
            senseData->Sense_Key = SENSE_KEY_ILLEGAL_REQUEST;
            senseData->Additional_Len = 0x06; /*no addtional sense bytes*/
            senseData->ASC =0x20;    /* INVALID COMMAND OPERATION CODE*/
            senseData->ASCQ =0x00;
            vSCp->result = (COMMAND_COMPLETE << 8) |
				    (CHECK_CONDITION << 1);
            mppLnx_schedule_resp(vSCp);
            return 0;
            
         }
         mppLnx_Scsi2toScsi3_handler(vSCp);
	      return 0;
         break;
         /*
         *The following commands are supported by the LSI storage array. Because the MPP 
         *driver has a scsi2-to-scsi3 emulation layer. The SCSI3 persistent reservation
         *with Exclusive-Access Registrants-only type allows those commands pass through.
         *But the Scsi2 spec says the target should reject those command with RESERVATION
         *CONFLICT status if the initiator is not the reservation owner. We must handle 
         *those command specially.
         */
      case  SCSI_TEST_UNIT_READY:
      case  SCSI_SAFE_PASS_THROUGH:
      case  SCSI_LOG_SENSE:
      case  SCSI_READ_CAPACITY:
      case  SCSI_READ_CAPACITY_16:
      case  SCSI_REPORT_LUNS:
      case  SCSI_START_STOP_UNIT:
      case  SCSI_LSI_QUIESCE:

         if(!mpp_ModuleArray[targetId].RdacInfo->FWSonoran4OrLater)
         {
           /* there is no scsi-2 emulation on pre-sonoran 4 array*/
            break;
         }
         else
         {
            mppLnx_Scsi2toScsi3_handler(vSCp);
	         return 0;
            break;
         }
         break;
      default:
	      break;
   }
   /*
   * delete the virtual command's timer
   */
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
      "VHBA Rcmd: Sn %ld, [%02x] vlun T%dL%d ",
      vSCp->serial_number,
      vSCp->cmnd[0],
      vSCp->target,vSCp->lun));

   timeoutfunc = (void (*)(Scsi_Cmnd *))vSCp->eh_timeout.function;
   /*
   * remember the middle level's timer out function.
   * when we restore the timer in out virtual command completion function,
   * we need the middle level's timeout function
   */
   if(middle_level_timeoutfunc == NULL)
   {
      middle_level_timeoutfunc = timeoutfunc;
   }
   scsi_delete_timer(vSCp);
   
   
   mppLnx_schedule_queuecmd(vSCp);
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_queuecommand()\n"));
   return 0;

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_eh_device_reset_handler(
* SUMMARY: issue SCSI bus reset
* SCOPE:    public 
*
* DESCRIPTION:
   This function is assigned to Scsi_Host_Template data struct's 
   "int (*eh_device_reset_handler)" function pointer. It can be invoked by
   middle level SCSI error handler thread and/or a user space application
   that sends ioctl "SG_SCSI_RESET_DEVICE" command to a scsi pass-through
   device node.
   A "bus device reset" message causes a target device performing reset and
   clear all SCSI-2 reservations. The mpp virtual HBA driver must provide 
   this function and perform the following tasks:
      call common API to perform SCSI-2-to-SCSI-3 translation layer's 
      device_reset function
      select a physical path and invoke the physical HBA's eh_device_reset_handler.
*  @scp: SCSI target that contains this device should be reset
*
* RESTRICTIONS:
* The scsi middle level holds a spin-lock while invoking the eh_device_reset_handler.
* IO can not be issued in this function. This function returns immediately and schedules
* the scsi2-to-scsi3 bus device reset call to be run in the worker kernel thread.
* 
* RETURNS:SUCCESS if  else FAILED
*
* ERRNO:
*
* NOTES: Invoked from scsi_eh thread. No other commands will be
*      queued on current host during error handling.
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_eh_device_reset_handler(Scsi_Cmnd * scp)
{
   Scsi_Cmnd *copyVscp = NULL;
   
   MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_eh_device_reset_handler Lun %d\n",scp->lun));
   
   /*
   * the command in the middle level is from stack. We must make a copy and give it
   * to the handler. The handler must free it after completing the operations
   */
   copyVscp = (Scsi_Cmnd *) kmalloc(sizeof(Scsi_Cmnd), GFP_ATOMIC);
   if(copyVscp == NULL)
   {
      /* error log*/
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 402,
         "Cannot get kernel memory for Scsi_Cmnd\n"));

      return FAILED;
   }
   else
   {
       copyVscp->cmnd[0]=MPPLNX_BDR_CMND_OPCODE;
       memset(&copyVscp->eh_timeout, 0, sizeof(scp->eh_timeout));
       copyVscp->host                     = scp->host;
       copyVscp->device                   = scp->device;
       copyVscp->target                   = scp->target;
       copyVscp->lun                      = scp->lun;
       copyVscp->channel                  = scp->channel;
       copyVscp->request.rq_status        = scp->request.rq_status;
       copyVscp->request.waiting          = NULL;
       copyVscp->use_sg                   = 0;
       copyVscp->old_use_sg               = 0;
       copyVscp->old_cmd_len              = 0;
       copyVscp->underflow                = 0;
       copyVscp->transfersize             = 0;
       copyVscp->resid                    = 0;
       copyVscp->serial_number            = 0;
       copyVscp->serial_number_at_timeout = 0;
       copyVscp->host_scribble            = NULL;
       copyVscp->next                     = NULL;
       copyVscp->state                    = scp->state;
       copyVscp->owner                    = scp->owner;
       copyVscp->scsi_done                = scp->scsi_done;
       copyVscp->done                     = scp->done;
       copyVscp->reset_chain              = scp->reset_chain;
        
       copyVscp->buffer                   = NULL;
       copyVscp->bufflen                  = 0;
       copyVscp->request_buffer           = NULL;
       copyVscp->request_bufflen          = 0;

       copyVscp->internal_timeout         = scp->internal_timeout;
       copyVscp->abort_reason             = scp->abort_reason;


       copyVscp->sc_data_direction        = scp->sc_data_direction;
       copyVscp->sc_request               = NULL;
       copyVscp->sc_magic                 = scp->sc_magic;

       /* sometimes the command can get back into the timer chain, so
        * use the pid as an identifier */
       copyVscp->pid                      = 0;
   }

   mppLnx_Scsi2toScsi3_handler(copyVscp);
   
   MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_eh_device_reset_handler- Lun %d\n",scp->lun));
   return SUCCESS;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_eh_bus_reset_handler
* SUMMARY:  Reset the virtual bus
* SCOPE:    public 
* DESCRIPTION:
   This function is assigned to Scsi_Host_Template data struct's 
   "int (*eh_bus_reset_handler)" function pointer. This function is invoked by 
   the error handler thread as well as sg's ioctl operation. This function is 
   called to reset the SCSI bus.
   The implication of this function will have an empty function body and 
   return "SUCCESS" status. The consideration is that the virtual bus should not 
   reset physical buses since middle level error recovery processes already 
   performed necessary error recovery procedures for a proxy command.

*
*  @scp: SCSI bus that contains this device should be reset
* RESTRICTIONS:
* 
* RETURNS: SUCCESS
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_eh_bus_reset_handler(Scsi_Cmnd * scp)
{
   return SUCCESS;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_eh_host_reset_handler
* SUMMARY:  Resets the virtual HBA adaptor
* SCOPE:    public 
*
* DESCRIPTION:
   This function is assigned to Scsi_Host_Template data struct's 
   "int (*eh_host_reset_handler)" function pointer. This function is invoked 
   by the error handler thread as well as sg's ioctl operation. The virtual HBA driver
   will not provide this function. This function asks a HBA driver to re-initialize the 
   HBA card.
   The implication of this function will have an empty function body and return "SUCCESS" status.
   The consideration is that the virtual bus should not reset physical buses since middle 
   level error recovery processes already performed necessary error recovery procedures 
   for a proxy command.

* @scp: SCSI host that contains this device should be reset
* RESTRICTIONS:
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_eh_host_reset_handler(Scsi_Cmnd * scp)
{
   return SUCCESS;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_bios_param
* SUMMARY:  fetch head, sector, cylinder info for a disk
* SCOPE:    public 
*
* DESCRIPTION:
   This function is assigned to Scsi_Host_Template data struct's "int (*bios_param)" 
   function pointer. This function is invoked by a upper level driver to report head, 
   sector, cylinder info for a disk device.
   The implementation of this function will be a proxy function of a physical 
   HBA's same function. The algorithm will be:
      1.	call mpp_SelectPath to get a physical device for this virtual device
      2.	get the physical device's Scsi_Host and then Scsi_Host_Template.
      3.	call the bios_param function pointer of the Scsi_Host_Template data struct
      4.	return to the caller.
*      @sdev: pointer to scsi device context (defined in sd.h)
*      @bdev: pointer to block device context (defined in fs.h)
*	   @capacity:  device size (in 512 byte sectors)
*      @params: three element array to place output:
*              params[0] number of heads (max 255)
*              params[1] number of sectors (max 63)
*              params[2] number of cylinders 
*
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_bios_param
(
   Disk * sdev,
   kdev_t dev,
   int *params
)
{
   Scsi_Host_Template   *phtp;
   Scsi_Device          * vSDp;
   Scsi_Device          * pSDp;

   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_bios_param()\n"));
   /** save the Disk's virtual Scsi_Device */
   vSDp = sdev->device;
   /*
   * get a physical scsi device for the virtual
   * In this case, we are not selecting a path for IO,
   * we just want to get any path of this virtual scsi device
   **/
   pSDp = mppLnx_getAnyPathByVDev(vSDp);
   /** get the physical device's Scsi_Host_Template **/
   phtp = pSDp->host->hostt;
    
   if(NULL != phtp->bios_param)
   {
      /*replace the Disk's device with the physical Scsi_Device to make the
       * physical HBA happy */
      MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
         "mppLnx_bios_param() using hba's bios_param func\n"));
      sdev->device = pSDp;  
      phtp->bios_param(sdev,dev,params);
        /*set the virtual Scsi_Device back to the disk */
      sdev->device = vSDp;
   }
   else
   {
      /* make up something meaningful
      */
      MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
         "mppLnx_bios_param() - Make up params\n"));
      scsicam_bios_param(
            sdev,	/* SCSI disk */
		      dev,	/* Device major, minor */
		      params);/* Heads, sectors, cylinders in that order */
   }
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_bios_param heads %d sectors %d cylinders %d \n",
         params[0] ,params[1] ,params[2] ));
   
   return 0;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_select_queue_depths
* SUMMARY:  Determines virtual Scsi_Device's queue depths
* SCOPE:    public 
*
* DESCRIPTION:
   This function is assigned to Scsi_Host_Template data struct's "void (*select_queue_depths)"
   function pointer. It is used to calculate allowable number of scsi commands that 
   can be queued on each device (disk) on the given host. This function should examine 
   all devices on the given host. The next device can be fetched with 
   sdp->next (NULL when finished). Queue depths should be placed 
   in Scsi_Device::queue_depth.
   A virtual device's queue_depth is dependent on its physical device's queue_depth. 
   For each virtual device in the virtual host , this function will perform the following:
      1.	call mpp_SelectPath() to get a physical device
      2.	assign the queue_depth of the physical device to the virtual 
            device's queue_depth field.
*
*  @shp the Scsi_Host object where the Scsi_Device is connected
*  @sdp the first Scsi_Device object in the Scsi_Device list
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_select_queue_depths(struct Scsi_Host * shp, Scsi_Device * sdp)
{
   Scsi_Device * pSDp; /* a physical device of a virtual device*/
   Scsi_Device * vSDp; /* a virtal device of the host */

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_select_queue_depths()\n"));
   /*
   * If the Lun 0 is a un-configured Lun, the shp->host_queue, which is passed to this 
   * function as the sdp argument, will be NULL. Check this condition.
   */

   if(NULL == sdp)
   {
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "Leaving mppLnx_select_queue_depths()- a null Scsi_Device pointer is passed in\n"));
      return;
   }

   for(vSDp = sdp; vSDp != NULL; vSDp = vSDp->next)
   {

      if(vSDp->lun >= mppLnx_Config_Parameters.mpp_MaxLunsPerArray )
      {
         MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
            "mppLnx_select_queue_depths()- virtual device LUN %d is out off range\n",
            vSDp->lun));

         continue;
      }
      
      pSDp = mppLnx_getAnyPathByVDev(vSDp);
      /* 
      * In the case of lun dynamic adding (middle level "scsi add-single-device" proc 
      * interface, a buggy Scsi_Device is hooked up to the host_queue by the scan_scsi, 
      * we need to check this condition here
      */


      if(pSDp == NULL)
      {
         MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
            "mppLnx_select_queue_depths()- cannot find a physical device for virtual device Lun %d\n",
            vSDp->lun));

         continue;
      }
      /*
      * In a single path per controller environment, the physical device's queue 
      * may be full before the virtual device's queue is full. This is because the 
      * failback scan thread performs synchronous io directly to the physical device.
      * There is not a good way to handle the physical device queue full if the mpp
      * driver does not queue-up command itself. The other case is that the failover
      * command will be issued asynchronously to a physical device directly. Therefore, 
      * to be safe, we will reduce the virtual device's queue_depth by 2.
      */
      if(pSDp->queue_depth >2)
      {
         vSDp->queue_depth = pSDp->queue_depth -2;
      }else
      {
         vSDp->queue_depth = pSDp->queue_depth;
      }
   }
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_select_queue_depths()\n"));
   return;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_newDeviceAdded
* SUMMARY:  Receives a notification of a new added Scsi_Device and create /proc 
*           node for the new device.
* SCOPE:    public 
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   The function is used for new-device-add event notification. During device 
   hot-add time, when a new device is added to the LSI mpp driver common data model, 
   the Linux MPP upper level driver calls this function to notify the virtual 
   HBA driver of a newly added Scsi device.
   This function insert this new scsi device to its internal data model and creates 
   the device's /proc leaf node
*  @sdp a Scsi_Device object that is newly added to the MPP driver
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_newDeviceAdded(Scsi_Device * sdp)
{
   int                        isVirtualDevice = 0;
   mppLnx_PDeviceEntry_t      *pde;
   mppLnx_VDeviceEntry_t      *vde;

   int                        vTarget =0;
   int                        vLun = 0;
   int                        cntl = 0;
   int                        path = 0;
   int                        foundpSDp = 0;
   RdacDeviceInformation_t    *rdacInfo;
   RdacLunInformation_t       *rdaclun;
   LunPathObjects_t           * lunPathObject;
   Scsi_Device                *pSDp;

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_2,
      "Entering mppLnx_newDeviceAdded() devies H%dC%dT%dL%d\n",
      sdp->host->host_no,
      sdp->channel,
      sdp->id,
      sdp->lun));
   /**
   * check if this device is a virtual or physical device
   */
   if(strcmp(sdp->host->hostt->proc_name,MPPLNX_PROC_NAME) == 0)
   {
      isVirtualDevice = 1;
   }
   else
   {
      isVirtualDevice = 0;
   }
   /*
   *check if the device is already in the internal data model
   */
   if(isVirtualDevice)
   {
      vde = mppLnx_get_VdeviceEntry(sdp);
      if(vde != NULL)
      {
         return;
      }
   }
   else
   {
      pde = mppLnx_get_PdeviceEntry(sdp);
      if(pde != NULL)
      {
         return;
      }
   }
   
   if(isVirtualDevice)
   {
      rdacInfo = mpp_ModuleArray[sdp->id].RdacInfo;
      mppLnx_insert_VdeviceEntry(sdp->lun, rdacInfo);
   }else
   {
      for(vTarget = 0; vTarget < mppLnx_Config_Parameters.mpp_MaxArrayModules; vTarget++)
      {
         if( (rdacInfo = mpp_ModuleArray[vTarget].RdacInfo) == NULL)
         {
            continue;
         }
      
         for(vLun = 0; vLun < mppLnx_Config_Parameters.mpp_MaxLunsPerArray; vLun++)
         {
            rdaclun = &(rdacInfo->RdacLuns[vLun]);
            if(!rdaclun->Present)
            {
               continue;
            }

            for(cntl = 0; cntl < 2; cntl++)
            {
               lunPathObject = rdaclun->LunControllers[cntl];
               for(path = 0; path < mppLnx_Config_Parameters.mpp_MaxPathsPerController; path++)
               {
                  pSDp = lunPathObject->LunPaths[path].LunPathDevice;
                  if(sdp == pSDp)
                  {
                     /*
                     *We found the device from the RdacInfo.
                     *1  insert it to the internal data model
                     *2. if the physical device is the first physical path
                     *   of a virtual device, we need to create its mppLnx_VDeviceEntry_t,
                     *   insert it to virtual device list, and create its proc file node.
                     *   We rely on mppLnx_insert_VdeviceEntry(). This function doesn't
                     *   create a new entry if the virtual device is already in the list
                     */
                     mppLnx_insert_PdeviceEntry(sdp, rdacInfo, cntl, path);
                     mppLnx_insert_VdeviceEntry(sdp->lun, rdacInfo);
                     foundpSDp = 1;
                     MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_2,
                           "Exiting mppLnx_newDeviceAdded()\n"));
                     return;
                  }
               }/* for path */


            }/* for cntl */

         } /* for vLun */

      } /* for vTarget */
   } /* else */

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_2,
      "Exiting mppLnx_newDeviceAdded()\n"));
   return;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_PHBA_eh_device_reset_handler
* SUMMARY:  Physical HBA driver's bus device reset error handler function
* SCOPE:    public
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   This is the interceptor error handler function of a physical HBA driver for sending 
   a bus device reset message to the target device.
   The interceptor function will check the incoming command. If the command is
   a proxy command, it will bypass the physical HBA's error handler. Otherwise, 
   it will pass the command to the physical HBA's error handler.

*
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_PHBA_eh_device_reset_handler(Scsi_Cmnd * scp)
{
   int isProxyCmnd = 0;
   int rt = SUCCESS;

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
      "Entering mppLnx_PHBA_eh_device_reset_handler() h%dc%dt%dl%d\n",
      scp->device->host->host_no, scp->channel, scp->target, scp->lun));
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_2,
      "The command is serial_number %ld command 0x%02x\n",
      scp->serial_number, scp->cmnd[0]));
   if(scp->done == mppLnx_scsi_done)
   {
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
         "mppLnx_PHBA_eh_device_reset_handler. Bypassed the HBA's error handler\n"));
      isProxyCmnd = 1;
   }
   else
   {
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
         "mppLnx_PHBA_eh_device_reset_handler. Invoke the HBA's error handler\n"));
      isProxyCmnd = 0;
   }

   if( !isProxyCmnd )
   {
      rt = mppLnxSavedPHBAEhHandlers.eh_device_reset_handler(scp);
   }

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
      "Exiting mppLnx_PHBA_eh_device_reset_handler()\n"));
   return rt;
   
}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_PHBA_eh_bus_reset_handler
* SUMMARY:  The interceptor function of a physical HBA driver's bus reset handler
* SCOPE:    public 
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   This is the interceptor error handler function for resetting the bus.
   The interceptor function will check the incoming command. If the command is 
   a proxy command, it will bypass the physical HBA's error handler. Otherwise, 
   it will pass the command to the physical HBA's error handler.

*
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_PHBA_eh_bus_reset_handler(Scsi_Cmnd * scp)
{
   int isProxyCmnd = 0;
   int rt = SUCCESS;


   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
      "Entering mppLnx_PHBA_eh_bus_reset_handler() h%dc%dt%dl%d\n",
      scp->device->host->host_no, scp->channel, scp->target, scp->lun));
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
      "The command is serial_number %ld command 0x%02x\n",
      scp->serial_number, scp->cmnd[0]));
   if(scp->done == mppLnx_scsi_done)
   {
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
         "mppLnx_PHBA_eh_bus_reset_handler. Bypassed the HBA's error handler\n"));
      isProxyCmnd = 1;
   }
   else
   {
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
         "mppLnx_PHBA_eh_bus_reset_handler. Invoke the HBA's error handler\n"));
      isProxyCmnd = 0;
   }

   if( !isProxyCmnd )
   {
      rt = mppLnxSavedPHBAEhHandlers.eh_bus_reset_handler(scp);
   }

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
      "Exiting mppLnx_PHBA_eh_bus_reset_handler()\n"));
   return rt;
}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_PHBA_eh_host_reset_handler
* SUMMARY:  The interceptor function of a physcial HBA driver/s host reset handler
* SCOPE:    public 
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   This is the interceptor error handler function for resetting the HBA card.
   The interceptor function will check the incoming command. If the command is 
   a proxy command, it will bypass the physical HBA's error handler. Otherwise, 
   it will pass the command to the physical HBA's error handler.

*
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_PHBA_eh_host_reset_handler(Scsi_Cmnd * scp)
{
   int isProxyCmnd = 0;
   int rt = SUCCESS;

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
      "Entering mppLnx_PHBA_eh_host_reset_handler()\n h%dc%dt%dl%",
      scp->device->host->host_no, scp->channel, scp->target, scp->lun));
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
      "The command is serial_number %ld command 0x%x\n",
      scp->serial_number, scp->cmnd[0]));
   if(scp->done == mppLnx_scsi_done)
   {
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
         "mppLnx_PHBA_eh_host_reset_handler. Bypassed the HBA's error handler\n"));
      isProxyCmnd = 1;
   }
   else
   {
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
         "mppLnx_PHBA_eh_host_reset_handler. Invoke the HBA's error handler\n"));
      isProxyCmnd = 0;
   }

   if( !isProxyCmnd )
   {
      rt = mppLnxSavedPHBAEhHandlers.eh_host_reset_handler(scp);
   }

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_3,
      "Exiting mppLnx_PHBA_eh_host_reset_handler()\n"));
   return rt;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_set_command_timeout
* SUMMARY:  Tune-up proxy command timeout value based on physical HBA
* SCOPE:    public 
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   Qlogic HBA driver resets allowed command retries if the allowed retries is less
   than its default value. 
   The failover method in the common code would like to failover to alternative controller
   in the time window of mpp_ControllerIoWaitTime. We need to tune-up proxy command timeout
   value to meet the failover time requirement
*
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
static void mppLnx_set_command_timeout(Scsi_Host_Template *pshtp)
{
   /*
   * check if it is qla HBA
   */
   if(memcmp(QLA_DRIVER_PROC_NAME_PREFIX,pshtp->proc_name,3) == 0)
   {
      /*
      * The QLA driver's default retry count is 20. With the following timeout per command
      * value, after the command comes back from the middle level, we still have a chance to
      * try the other controller.
      */
      mpLnx_proxy_cmnd_timeout = 10 + (mppLnx_Config_Parameters.mpp_ControllerIoWaitTime *HZ)/3;
   }
   else
   {
      mpLnx_proxy_cmnd_timeout = 60 * HZ;
	
   }
}

MODULE_AUTHOR("IBM");
MODULE_DESCRIPTION("MPP Virtual HBA Driver");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
