/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************

AME            mppLnx_vhbaio.c
SUMMARY         %description%
VERSION         %version: 32 %
UPDATE DATE     %date_modified: Fri Feb 20 10:11:18 2004 %
PROGRAMMER      %created_by:    yqi %

        Copyright 2003, LSI Logic Storage Systems, Inc.

DESCRIPTION:This source file contains Linux MPP virtual HBA driver functions 
   that are related to IO routing. The function mppLnx_queuecommand() in 
   mppLnx_vhba.c receives Scsi command for all virtual Scsi devices from 
   the Linux SCSI middle level. The functions in this file are responsible 
   for routing those virtual Scsi commands to physical HBA drivers, taking 
   care of path/controller failover, and completing those virtual commands. 
   The functions in this file can be considered as "the Heart of the Linux 
   MPP virtual HBA driver".

INCLUDE FILES:

NOTES:

RESTRICTIONS:

SEE ALSO:

REFERENCE:

  349-1033500 Linux MPP Driver Feature Architecture Specification
  349-1034610 Linux MPP Virtual HBA Driver Feature Design Document
  349-1034750 Linux MPP Upper Level Driver Feature Design Document
  349-1034730 Linux MPP Utility Feature Design Document
  349-1034620 Linux MPP Installation Feature Design Document

IMPLEMENTATION:

MODIFICATION HISTORY:

*******************************************************************************/

#define __SRCmppLnx_vhbaio_c

/***  INCLUDES  ***/  
/***  Linux kernel INCLUDES  ***/  
#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/genhd.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/interrupt.h>

#include <asm/io.h>

#include <linux/blk.h>
#include "scsi.h"
#include "hosts.h"
#include "sd.h"

#include <linux/stat.h>

/***  MPP commomn API INCLUDES  ***/  
#include "MPP_Common.h"
#include "MPP_RdacInfo.h"
#include "MPP_ProtoTypes.h"
#include "MPP_Sysdep.h"
#include "mppLnx_version.h"
/***  Linux MPP Virtual HBA INCLUDES  ***/  
#include "mppLnx_vhba.h"
#include "mppLnx_shared.h"

/***  CONSTANT DEFINITIONS  ***/


/***  MACRO DEFINITIONS  ***/

/***  TYPE DEFINITIONS  ***/

extern void (*middle_level_scsi_done)(Scsi_Cmnd *);
extern void (*middle_level_timeoutfunc)(Scsi_Cmnd *);
extern int mpLnx_proxy_cmnd_timeout;

/***  LOCALS  ***/

static void mppLnx_scsi_init_proxy_cmnd(Scsi_Cmnd *Scpnt);
static void mppLnx_scsicmnd_virtual_to_proxy(Scsi_Cmnd *vscp, Scsi_Cmnd *pscp);
static void mppLnx_scsicmnd_proxy_to_virtual(Scsi_Cmnd *vscp, Scsi_Cmnd *pscp);

#if USE_THREAD_FOR_DPC
extern mppLnx_dpcQueueContext_t   mppLnxDpcQueueContext;
#endif

/*FIX ME. It should be declared somewhere else
*/
extern LWORD mppLnx_Translate_Linux_Status(int status);
extern BYTE mpp_IsFailoverPossible(RdacDeviceInformation_t *RdacInfo, 
                                   BYTE Controller, BYTE Path, BYTE Lun, BYTE level);
/***  PROCEDURES  ***/

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_do_queuecommand
* SUMMARY:  dispatch a virtual Scsi_Cmnd to physical HBA driver
* SCOPE:     public
*
*
* DESCRIPTION:
   This function receives a virtual Scsi_Cmnd and dispatches a proxy Scsi_Cmnd to
   a physical HBA driver.

*  @vscp The virtual Scsi_Cmnd needs to be routed.
*  @done The virutal Scsi_Cmnd's command completion function
*  @timeoutfunc   The middle level time out function
*  @ml_done       The middle level command completion function
*  @previous_controller If the virtual IO is re-sent to a new path, we need to know
                  which controller of the virtual IO was tried. For a virtual IO, which
                  is first time queued to the virtual HBA driver, the previous_controller
                  will be UNKNOWN_PREVIOUS_CONTROLLER.
*  @previous_path If the virtual IO is re-sent to a new path, we need to know
                  which path of the virtual IO was tried. For a virtual IO, which
                  is first time queued to the virtual HBA driver, the previous_controller
                  will be UNKNOWN_PREVIOUS_PATH.
*  @CommandTimeoutRetryCount  The time out retry count. If this function is invoked from
                              the mppLnx_queuecommand(), the value should be the 
                              value from the field configurable parameter. If this is a
                              command retry, the value should be the value returned by the
                              common code
*  @RetryCount                Command retry counter. If this function is invoked from
                              the mppLnx_queuecommand(), the value should be the 
                              value from the field configurable parameter. If this is a
                              command retry, the value should be the value returned by the
                              common code
*  @UaRetryCount              Unit Attention retry count.
*  @vcmndStartTime            The start up time of a virtual command being dispatched
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_do_queuecommand(Scsi_Cmnd  * vSCp, 
                                  void (*timeoutfunc)(Scsi_Cmnd *),
                                  void (*ml_done)(Scsi_Cmnd *),
                                  BYTE  previous_controller,
                                  BYTE  previous_path,
                                  BYTE CommandTimeoutRetryCount,
                                  BYTE RetryCount,
                                  BYTE SelectionTimeoutRetryCount,
                                  BYTE UaRetryCount,
                                  MPP_TIME vcmndStartTime,
                                  MPP_TIME ControllerIoWaitTime,
                                  MPP_TIME ArrayIoWaitTime)
{
   int                           initialSelect = 1;
   Scsi_Device                   * pSDp;
   Scsi_Device                   * firstSDp = NULL;
   Scsi_Cmnd                     * proxyCmnd;
   RdacDeviceInformation_t       *rdacInfo;
   BYTE                          lun;
   BYTE                          controller;
   BYTE                          path;
   int                           vtarget;
   mppLnx_ProxyRequestEntry_t    *proxyRequest;
   
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_do_queuecommand()\n"));
   
   if(NULL == vSCp)
   {
      MPPLNX_VHBA_BUG();
   }

   if(NULL == ml_done)
   {
      MPPLNX_VHBA_BUG();           
   }

   if(NULL == timeoutfunc)
   {
      MPPLNX_VHBA_BUG();
   }

   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
      "cmnd SN 0x%lx cmndtimeoutfunc 0x%ld ml_done 0x%lx \n",
      vSCp->serial_number,
      (unsigned long)timeoutfunc,
      (unsigned long) ml_done));

   vtarget = vSCp->target;
   lun = vSCp->lun;
   rdacInfo = mpp_ModuleArray[vtarget].RdacInfo;

   
   while(1)
   {
      pSDp = mpp_SelectPath(rdacInfo, lun, &controller, &path);
      if(initialSelect)
      {
         firstSDp = pSDp;
      }

      /*
      * if there is not path available, (all of them failed)
      * we need let the middle level know
      */
      if( pSDp == NULL)
      {
         /* check if failover possible */
         if (mpp_IsFailoverPossible(rdacInfo, rdacInfo->RdacLuns[lun].CurrentOwningPath, 0, lun, 1))
         {
            MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 502,
               "mppLnx_do_queuecommand: no valid paths found, failing over to controller %d for array %s, lun %d\n", 
               rdacInfo->RdacLuns[lun].CurrentOwningPath^1,
               mpp_ModuleArray[rdacInfo->VirtualTargetId].ModuleName,
               lun));
            mpp_FailController(rdacInfo, rdacInfo->RdacLuns[lun].CurrentOwningPath, lun);
            rdacInfo->RdacLuns[lun].CurrentOwningPath ^= 1;
            /*
            *I did controller failover and re-set the current owing path to the alternative
            *controller. I would like go back to the while loop to re-select a new path
            *from the alternative controller.
            */
            continue;
         }
         else
         {
            /*return DID_NO_CONNECT. The IO will be failed*/
            MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 480,
               "selected path retuns NULL. Array %s Lun %d SN %ld\n",
               (char *)rdacInfo->ModuleName, lun,vSCp->serial_number));

            vSCp->result = DID_NO_CONNECT<<16;
            scsi_add_timer(vSCp,5*HZ,timeoutfunc);
            vSCp->scsi_done(vSCp);
            return 0;
         }
      }
      /* 
      * OK, we get a path. Find a command slot from the device
      * scsi_allocate_device(Scsi_Device *, int wait, int interruptable)
      * wait 1-> if we should wait in the event that none device are available
      * interruptable 1 if we should unblock and return NULL.
      * if we cannot get a free slot, try another path
      */
      proxyCmnd = scsi_allocate_device(pSDp,FALSE, FALSE);
      if(proxyCmnd != NULL)
      {
         break;
      }
      else
      {
         MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 481,
            "mppLnx_do_queuecommand: can not find free slot from h%dc%dt%dl%d\n",
            pSDp->host->host_no, pSDp->channel, pSDp->id, pSDp->lun));
      }

      if(!initialSelect && pSDp == firstSDp )
      { 
         break;
      }
      initialSelect = 0;
   }
   
   if(NULL == proxyCmnd)
   {
      /*
      * why cannot we get a command slot? 
      * We should not get here in a normal case. The possible reasons are:
      * 1, physical command didn't get released properly (bug)
      * 2. A virtual device has the same queue depth as a physical device.
      *    In the case where we have only one path per controller and the device
      *    is performing IO stress testing, a sync IO may make the physical command
      *    queue full so that a virtual IO can not get a free slot.
      * Log an error and return buss busy. The SCSI middle level should perform 
      * a re-try.
      */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 482,
         "BIG ERROR Could not find a free command slot from all paths SN %ld\n",
         vSCp->serial_number));
      
      /*
      * If we can not get a free command slot from a physical device, set the virtual 
      * command's allowed-retry-count to 20 and return DID_BUS_BUSY scsi status to 
      * the middle level. The middle SCSI middle level should do retry on this command. 
      * On the other hands, we reduce the virtual device's queue depth by 2 so that
      * we should not get this condition.
      */
      vSCp->allowed = 20;
      vSCp->result = DID_BUS_BUSY<<16;
      scsi_add_timer(vSCp,5*HZ,timeoutfunc);
      vSCp->scsi_done(vSCp);
      return 0;
   }

   /**
   * create a mppLnx_ProxyRequestEntry_t object and insert it
   */
   proxyRequest = mppLnx_createProxyRequest(vSCp,proxyCmnd);

   if(NULL == proxyRequest)
   {
      MPPLNX_VHBA_BUG();
   }

   mppLnx_insert_proxyRequest(proxyRequest);

   /*
   * assign the startTime (jiffies)
   */
   MPP_GETTIME(proxyRequest->startTime);
   proxyRequest->vcmndStartTime =vcmndStartTime;
   proxyRequest->CommandTimeoutRetryCount = CommandTimeoutRetryCount;
   proxyRequest->SelectionTimeoutRetryCount = SelectionTimeoutRetryCount;
   proxyRequest->UaRetryCount = UaRetryCount;
   proxyRequest->ControllerIoWaitTime = ControllerIoWaitTime;
   proxyRequest->ArrayIoWaitTime = ArrayIoWaitTime;
   
   if( ( (previous_path != UNKNOWN_PREVIOUS_PATH) && 
         (previous_controller != UNKNOWN_PREVIOUS_CONTROLLER) ) &&
      ( (previous_controller != controller) || (previous_path != path) ) )
   {
      proxyRequest->RetryCount = mppLnx_Config_Parameters.mpp_SyncRetryCount;
   }
   else
   {
      proxyRequest->RetryCount = RetryCount;
   }
   /*
   * Check if we need to reset wait time states and timers
   */
   if( (previous_controller != UNKNOWN_PREVIOUS_CONTROLLER) && (previous_controller != controller) )
   {
      	// IO being resent on a new controller, reset counters and timers.
      MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_2, 
				"mppLnx_do_queuecommand: IO retried on new controller, %s:%d:%d:%d\n", 
				rdacInfo->ModuleName, controller, path,	lun));


      proxyRequest->UaRetryCount = 0;
      proxyRequest->CommandTimeoutRetryCount = 0;
      proxyRequest->SelectionTimeoutRetryCount = 0;
      rdacInfo->RdacLuns[lun].NotReady = FALSE;
      rdacInfo->RdacLuns[lun].Busy = FALSE;
      rdacInfo->RdacLuns[lun].Quiescent = FALSE;
      rdacInfo->RdacLuns[lun].QuiescienceStart = 0;
      rdacInfo->RdacLuns[lun].NotReadyStart = 0;
      rdacInfo->RdacLuns[lun].BusyStart = 0;		
      MPP_GETTIME(proxyRequest->ControllerIoWaitTime);
   }
   /*
   * copy/set necessary fields from the virtual command to its proxy command.
   */
   mppLnx_scsicmnd_virtual_to_proxy(vSCp, proxyCmnd);

#ifdef MPP_DEBUG
   if(vSCp->cmnd[0] == SCSI_INQUIRY && 
         (vSCp->cmnd[1] & 0x01) == 0 )
   {
      printk("Dispatch Inquiry cmnd for virtual device %d:%d\n",
         vSCp->target,vSCp->lun);
         
   }
   printk("v cmnd %02x %02x \n",vSCp->cmnd[0],vSCp->cmnd[1]);
#endif

   /*
   * Queue the proxy SCSI command to the SCSI middle level
   *
   * Arguments:   proxyCmnd     - command descriptor.
   *              cmnd      - actual SCSI command to be performed.
   *              buffer    - data buffer.
   *              bufflen   - size of data buffer.
   *              done      - completion function to be run.
   *              timeout   - how long to let it run before timeout.
   *              retries   - number of retries we allow.
   *
   * Lock status: With the new queueing code, this is SMP-safe, and no locks
   *              need be held upon entry.   The old queueing code the lock was
   *              assumed to be held upon entry.
   *
   * Returns:     Nothing
   */
   scsi_do_cmd(
      proxyCmnd, 
      (const void *)vSCp->cmnd,
      (void *)vSCp->buffer,
      vSCp->bufflen,
      mppLnx_scsi_done,
      mpLnx_proxy_cmnd_timeout,  /* the value of a proxy command is tuned up to fit
                                 *  a physical HBA driver's implementation
                                   */
      0);
 
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_do_queuecommand()\n"));
   return 0;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_RetrySamePath
* SUMMARY:  dispatch a virtual Scsi_Cmnd to a given Scsi_Device (path)
* SCOPE:     private
*
*
* DESCRIPTION:
   This function receives a virtual Scsi_Cmnd and dispatches a proxy Scsi_Cmnd to
   the specified Scsi_Device

*  @pSDp The physical Scsi_Device object to which the virtual Scsi_Cmnd will issue
*         a proxy command to
*  @vscp The virtual Scsi_Cmnd needs to be routed.
*  @done The virutal Scsi_Cmnd's command completion function
*  @timeoutfunc   The middle level time out function
*  @ml_done       The middle level command completion function
*  @CommandTimeoutRetryCount  The time out retry count. If this function is invoked from
                              the mppLnx_queuecommand(), the value should be the 
                              value from the field configurable parameter. If this is a
                              command retry, the value should be the value returned by the
                              common code
*  @RetryCount                Command retry counter. If this function is invoked from
                              the mppLnx_queuecommand(), the value should be the 
                              value from the field configurable parameter. If this is a
                              command retry, the value should be the value returned by the
                              common code
*  @UaRetryCount              Unit Attention retry count.
*  @vcmndStartTime            The start up time of a virtual command being dispatched
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_RetrySamePath(
                           Scsi_Device *pSDp,
                           Scsi_Cmnd  * vSCp, 
                           void (*timeoutfunc)(Scsi_Cmnd *),
                           void (*ml_done)(Scsi_Cmnd *),
                           BYTE CommandTimeoutRetryCount,
                           BYTE RetryCount,
                           BYTE SelectionTimeoutRetryCount,
                           BYTE UaRetryCount,
                           MPP_TIME vcmndStartTime,
                           MPP_TIME ControllerIoWaitTime,
                           MPP_TIME ArrayIoWaitTime)
{
   Scsi_Cmnd                     * proxyCmnd;
   mppLnx_ProxyRequestEntry_t    *proxyRequest;
   
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_RetrySamePath()\n"));
   
   if(NULL == vSCp)
   {
      MPPLNX_VHBA_BUG();
   }

   if(NULL == ml_done)
   {
      MPPLNX_VHBA_BUG();           
   }

   if(NULL == timeoutfunc)
   {
      MPPLNX_VHBA_BUG();
   }

   if(NULL == pSDp)
   {
      MPPLNX_VHBA_BUG();
   }

   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
      "cmnd SN 0x%lx cmndtimeoutfunc 0x%ld ml_done 0x%lx \n",
      vSCp->serial_number,
      (unsigned long)timeoutfunc,
      (unsigned long) ml_done));


      /* 
      * OK, we use the same device and get another command slot to re-issue
      * the IO.
      */
      proxyCmnd = scsi_allocate_device(pSDp,FALSE, FALSE);
     
   
   if(NULL == proxyCmnd)
   {
      /*
      * why cannot we get a command slot? 
      * We should not get here in a normal case. The possible reasons are:
      * 1, physical command didn't get released properly (bug)
      * 2. A virtual device has the same queue depth as a physical device.
      *    In the case where we have only one path per controller and the device
      *    is performing IO stress testing, a sync IO may make the physical command
      *    queue full so that a virtual IO can not get a free slot.
      * Log an error and return buss busy. The SCSI middle level should perform 
      * a re-try.
      */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 503,
         "BIG ERROR Could not find a free command slot from all paths SN %ld\n",
         vSCp->serial_number));
      
      /*
      * If we can not get a free command slot from a physical device, set the virtual 
      * command's allowed-retry-count to 20 and return DID_BUS_BUSY scsi status to 
      * the middle level. The middle SCSI middle level should do retry on this command. 
      * On the other hands, we reduce the virtual device's queue depth by 2 so that
      * we should not get this condition.
      */
      vSCp->allowed = 20;
      vSCp->result = DID_BUS_BUSY<<16;
      scsi_add_timer(vSCp,5*HZ,timeoutfunc);
      vSCp->scsi_done(vSCp);
      return 0;
   }

   /**
   * create a mppLnx_ProxyRequestEntry_t object and insert it
   */
   proxyRequest = mppLnx_createProxyRequest(vSCp,proxyCmnd);

   if(NULL == proxyRequest)
   {
      MPPLNX_VHBA_BUG();
   }

   mppLnx_insert_proxyRequest(proxyRequest);

   /*
   * assign the startTime (jiffies)
   */
   MPP_GETTIME(proxyRequest->startTime);
   proxyRequest->vcmndStartTime =vcmndStartTime;
   proxyRequest->CommandTimeoutRetryCount = CommandTimeoutRetryCount;
   proxyRequest->SelectionTimeoutRetryCount = SelectionTimeoutRetryCount;
   proxyRequest->UaRetryCount = UaRetryCount;
   proxyRequest->RetryCount = RetryCount;
   proxyRequest->ControllerIoWaitTime = ControllerIoWaitTime;
   proxyRequest->ArrayIoWaitTime = ArrayIoWaitTime;
   /*
   * copy/set necessary fields from the virtual command to its proxy command.
   */
   mppLnx_scsicmnd_virtual_to_proxy(vSCp, proxyCmnd);

   scsi_do_cmd(
      proxyCmnd, 
      (const void *)vSCp->cmnd,
      (void *)vSCp->buffer,
      vSCp->bufflen,
      mppLnx_scsi_done,
      mpLnx_proxy_cmnd_timeout,  /* the value of a proxy command is tuned up to fit
                                 *  a physical HBA driver's implementation
                                   */
      0);
 
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_RetrySamePath()\n"));
   return 0;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_IssueFailover
* SUMMARY:  Issues an async Mode Select page-2c command to a selected physical device
* SCOPE:    public 
*
* DESCRIPTION:;
   This function is an implementation of the system dependent function mppSys_IssueFailover(),
   which is specified by the MPP driver command API as a system dependent function. 
   It issues a Model Select Page2c command to the alternative controller.
   This function performs the following operations:
      Asks the Linux Scsi middle level to allocate a Scsi_Cmnd object from 
         the "FailovorHandle" Scsi_Device object
      Assigns the input data and CDB to the Scsi_Cmnd
      Assigns a special command completion function to this command. The special 
         "(*done)" function will be mppLnx_failoverCmnd_done().
      Issues the command to the physical HBA driver
      The completion of this command will come to mppLnx_failoverCmnd_done() function.
*  @rdacInfo         the RdacDeviceInformation_t object which represents a storage array
*  @controller       Controller slot 
*  @Path             The path index
*  @FailoverHandler  The physical Scsi_Device object to which the failover command will send to
*  @Page2C           The buffer which contains the Mode Select Page 2C 
                     (Redundant Controller Page)
*  @cdb              The Command Descriptor Block of the Mode Select command
* RESTRICTIONS:
* 
* RETURNS: Nothing
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_IssueFailover(RdacDeviceInformation_t *rdacInfo, BYTE controller, 
		BYTE Path, MPP_HANDLE FailoverHandle, BYTE *Page2C, BYTE *cdb)
{
   Scsi_Device                   * pSDp;
   Scsi_Cmnd                     * pSCp;
   Scsi_Cmnd                     * fakeVscp;
   mppLnx_ProxyRequestEntry_t    * fakeRequestp;
   LINT                          LockLevel = 0;
   int                           PageSize;
   
   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_IssueFailover()\n"));
   pSDp = FailoverHandle;

   if(NULL == pSDp)
   {
      MPPLNX_VHBA_BUG();
   }

   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_IssueFailover() S1\n"));
   /*
   * get a scsi cmnd slot from the physical device for issuing the command
   */
   pSCp = scsi_allocate_device(pSDp, FALSE, FALSE);
   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_IssueFailover() S2\n"));
   /*
   * check if we get a physical command
   */
   if(NULL == pSCp)
   {
      /*
      * BIG error
      */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 486,
         "BIG ERROR Could not find a free command slot from all possible failover paths\n"));
      /*
      * we didn't issue a command to the controller. What do I do.
      * the only thing I can do is to mark the failover-in-progress to false
      */
      MPP_LOCK(rdacInfo->Lock, LockLevel);
      rdacInfo->ControllerInfo[controller]->FailoverInProgress = FALSE;
      MPP_UNLOCK(rdacInfo->Lock, LockLevel);
      return;
   }

   /*
   * this is a manufacture IO command by the mpp driver. It does not have a
   * virtual command from the upper level driver associated with it.
   * But the commmand asynchly, we need to keep tracking it by using a proxy 
   * request object
   */
   fakeVscp = kmalloc(sizeof(Scsi_Cmnd), GFP_ATOMIC);
   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_IssueFailover() S4\n"));
   
   if(fakeVscp == NULL)
   {
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 487,
         "Error in allacating memory for fail-over command\n"));
      /*
      * we didn't issue a command to the controller. What do I do.
      * the only thing I can do is to mark the failover-in-progress to false
      */
      MPP_LOCK(rdacInfo->Lock, LockLevel);
      rdacInfo->ControllerInfo[controller]->FailoverInProgress = FALSE;
	   MPP_UNLOCK(rdacInfo->Lock, LockLevel);
      
      return;
   }
   memset(fakeVscp, 0, sizeof(Scsi_Cmnd));
   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_IssueFailover() S5\n"));
   
   fakeVscp->lun = pSDp->lun;
   fakeVscp->target = rdacInfo->VirtualTargetId;
   fakeVscp->channel = 0;
   fakeVscp->device = rdacInfo->RdacLuns[fakeVscp->lun].PseudoLunObject;
   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_IssueFailover() S6\n"));

   fakeRequestp = mppLnx_createProxyRequest(fakeVscp,pSCp);
   mppLnx_insert_proxyRequest(fakeRequestp);

   fakeRequestp->CommandTimeoutRetryCount = 0;
   fakeRequestp->RetryCount = mppLnx_Config_Parameters.mpp_SyncRetryCount;
   fakeRequestp->UaRetryCount = 0;
   fakeRequestp->SelectionTimeoutRetryCount = 0;
   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_IssueFailover() S7\n"));

   MPP_GETTIME(fakeRequestp->startTime);
   MPP_GETTIME(fakeRequestp->vcmndStartTime);

   /**
   * preparing the out going command
   */
   mppLnx_scsi_init_proxy_cmnd(pSCp);
   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_IssueFailover() S8\n"));
   /**
   * the command completion function mppLnx_failoverCmd_done() should remember 
   * to free the virtual command memory
   */
   PageSize = rdacInfo->Page2CSubPage1 ? sizeof(ModePage2CSubPage_t) :
									   sizeof(ModePage2C_t);
   pSCp->request_buffer = (void *)Page2C;
   pSCp->request_bufflen = PageSize;
   pSCp->done = mppLnx_failoverCmd_done;
   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_IssueFailover() S10\n"));

   /*
   * send the failover command to the SCSI middle level
   */
   
   scsi_do_cmd(
      pSCp,
      (const void *)cdb,
      (void *)Page2C,
      PageSize,
      mppLnx_failoverCmd_done,
      mppLnx_Config_Parameters.mpp_FailoverTimeout*HZ,  /*the value is 120 seconds */
      0);

   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_IssueFailover dump command after sending to middle level\n"));

   MPPLNX_DUMPCMND(pSCp);

   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exitng mppLnx_IssueFailover() for T%dL%dCtrl%d\n",
      rdacInfo->VirtualTargetId,
      pSCp->lun,
      controller));
}




/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_failoverCmd_done
* SUMMARY:  Receives failover command's completion status
* SCOPE:    public 
*
* DESCRIPTION:
   This is a special command completion function for a failover command. 
   Unlike the normal command completion function, because the failover command 
   is not requested by a host applications, this function will not cause a virtual I
   /O to be completed.

   This function will call mppCmn_AnalyseFailoverError() to determine whether or 
   not the failover command needs to be retried.
   If the command is successful, it will make RdacControllerInformation_t object's 
   FaiolverInProgress to false.

*  @scp The failover command which is completed by a physical HBA driver
* RESTRICTIONS:
* 
* RETURNS: Nothing
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_failoverCmd_done(Scsi_Cmnd * pSCp)
{
   long                          mpp_status = 0;
   long                          mppAction;
   RdacDeviceInformation_t       *RdacInfo;
   BYTE                          Controller;
   BYTE                          Path;
   int                           Lun;
   long                          StartTime;
   SenseData_t                   * SenseData;
   int                           SenseLength;
   long                          DataLengthXfered;
   mppLnx_PDeviceEntry_t         *pde;
   mppLnx_ProxyRequestEntry_t    *proxyRequest;
   Scsi_Device                   *newPath;
   LINT                          LockLevel = 0;
   int                           FailoverInProgress = 0;
   int                           PageSize;

   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_3,
      "Entering mppLnx_failoverCmd_done()\n"));

   if(pSCp == NULL)
   {
      MPPLNX_VHBA_BUG();
   }
   
   proxyRequest = mppLnx_get_virtualRequest(pSCp);
   pde = mppLnx_get_PdeviceEntry(pSCp->device);
   RdacInfo = pde->rdacInfo;
   Controller = pde->controllerSlot;
   Path = pde->pathId;
   Lun = pSCp->lun;
   SenseData = (SenseData_t *)pSCp->sense_buffer;
   SenseLength = sizeof(pSCp->sense_buffer);

   PageSize = RdacInfo->Page2CSubPage1 ? sizeof(ModePage2CSubPage_t) :
									   sizeof(ModePage2C_t);

   if(pSCp->sense_buffer[0] != 0)
   {
      MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_3,
         "Sense Data Dump: Response Code 0x%02x Sense-Key/ASC/ASCQ 0x%02x/0x%02x/0x%02x \n",
         SenseData->Error_Code,
         SenseData->Sense_Key,
         SenseData->ASC,
         SenseData->ASCQ));
   }
   StartTime = proxyRequest->startTime;
   DataLengthXfered = pSCp->transfersize;
   mpp_status = mppLnx_Translate_Linux_Status(pSCp->result);

   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_3,
      "mppLnx_failoverCmd_done() mppStatus %d\n",mpp_status));
   if(mpp_status)
   {
      mppAction = mpp_AnalyseFailoverError(
            RdacInfo, 
            Controller,
            Path, 
            Lun, 
            mpp_status, 
			   SenseData, 
            SenseLength, 
            DataLengthXfered, 
            StartTime);

      MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_3,
            "mppLnx_failoverCmd_done() mppAction %d\n",mpp_status));
      switch(mppAction)
      { 
         case MPP_NO_FAILOVER_ERROR:
            MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_3,
               "mppLnx_failoverCmd_done. No error\n"));
            FailoverInProgress = 0;
            break;
         case MPP_RETURN_FAILOVER_ERROR:
            MPP_ERRORLOGPRINT((MPP_ERR_FATAL,488,
               "mppLnx_failoverCmd_done: Fatal Error, array %s, ctrl %d, path %d, lun %d mpp_status %d\n",
               RdacInfo->ModuleName,
               Controller,
               Path,
               Lun,
               mpp_status));
            FailoverInProgress = 0;
            break;
         case MPP_RETRY_FAILOVER:
            MPP_ERRORLOGPRINT((MPP_ERR_FATAL,489,
               "mppLnx_failoverCmd_done retry failover commanmd\n"));
            newPath = mpp_SelectFailoverPath(
                     RdacInfo,
                     Lun,
                     &Controller,
                     &Path);
            if(NULL != newPath)
            {
               MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_3,
                  "mppLnx_failoverCmd_done. Re-issue failover\n"));
               mppLnx_IssueFailover(
                  RdacInfo,
                  Controller,
                  Path,
                  newPath,
                  pSCp->buffer,
                  pSCp->cmnd);
               FailoverInProgress = 1;
               break;
            }else
            {
               MPP_ERRORLOGPRINT((MPP_ERR_FATAL,490,
                  "mppLnx_failoverCmd_done. No path to re-issue failover\n"));
               FailoverInProgress = 0;
               break;
            }
         default:
            MPP_ERRORLOGPRINT((MPP_ERR_FATAL,491,
               "mppLnx_failoverCmd_done. Unknown Action \n"));
            FailoverInProgress = 0;
            /* log an erro */
            break;
      }

   }
   /*
   * clean up
   * 1. release the proxy command
   * 2. free proxy request's memory
   */

   kfree(proxyRequest->vcmnd);
   scsi_release_command(pSCp);
   kfree(proxyRequest);

   if(FailoverInProgress == 0)
   {
      MPP_FREE(pSCp->buffer,PageSize);
      MPP_LOCK(RdacInfo->Lock, LockLevel);
      RdacInfo->ControllerInfo[Controller]->FailoverInProgress = FALSE;
      MPP_UNLOCK(RdacInfo->Lock, LockLevel);
   }

   MPP_DEBUGPRINT((MPP_FAILOVER_IO_DEBUG+MPP_DEBUG_LEVEL_3,
      "Existing mppLnx_failoverCmd_done()\n"));
}

#if USE_THREAD_FOR_RETRY
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_scsi_done
* SUMMARY:  The Scsi_Cmnd completion function of a proxy Scsi_Cmnd.
* SCOPE:    public 
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   This function is the proxy command completion function. When a proxy command 
   This function is the proxy command completion function. When a proxy command 
   is completed by the HBA driver and further processed by the Scsi middle level, 
   this function is called by the Scsi middle level to notify the proxy 
   command's completion. This function is assigned to each proxy command's "(*done)" function.
   When USE_THREAD_FOR_RETRY is defined and sets to 1, this function will allocate a mppLnx_CmndEntry_t 
   object and insert the completed proxy command to the command queue for completion or retry by mpp_DCR 
   kernel thread.
*  @pscp The proxy Scsi_Cmnd that is completed by a physical HBA driver
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_scsi_done(Scsi_Cmnd * pSCp)
{
   mppLnx_CmndEntry_t      * completedCmndEntry;
   int                     size = 0;
   unsigned long           iflages;

   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "mppLnx_scsi_done received cmnd %lx serial_number %ld\n",
      (unsigned long)pSCp, pSCp->serial_number));
  
   size = sizeof(mppLnx_CmndEntry_t);
   completedCmndEntry = kmalloc(size, GFP_ATOMIC);
   if( NULL == completedCmndEntry )
   {
      /* error. Cannot get memory */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL,503,
         "cannot get memory for mppLnx_CmndEntry_t\n"));
      /*
      * complete the command directly by the SCSI bottom half if we get here
      */
      __mppLnx_scsi_done(pSCp);
      return;
   }
   else
   {
      spin_lock_irqsave(mppLnxDpcQueueContext.cmndQueueLock, iflages);
      memset(completedCmndEntry, 0, size);
      completedCmndEntry->vcmnd = pSCp;
      completedCmndEntry->taksType = MPPLNX_COMPLETED_CMND_TYPE;
      list_add_tail(&(completedCmndEntry->list), 
			&(mppLnxDpcQueueContext.cmndEntry->list));
      spin_unlock_irqrestore(mppLnxDpcQueueContext.cmndQueueLock, iflages);

      if(atomic_read(&mppLnxDpcQueueContext.needSchedule))
      {
         up(mppLnxDpcQueueContext.dpc_sem);
      }
   }
   
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Exiting mppLnx_scsi_done\n"));
}
#endif

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_scsi_done or __mppLnx_scsi_done(Scsi_Cmnd * pSCp)
* SUMMARY:  The Scsi_Cmnd completion function of a proxy Scsi_Cmnd.
* SCOPE:    public 
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   This function is the proxy command completion function. When a proxy command 
   is completed by the HBA driver and further processed by the Scsi middle level, 
   this function is called by the Scsi middle level to notify the proxy 
   command's completion. This function is assigned to each proxy command's "(*done)" function.
   This function is a core function of the Linux MPP virtual HBA driver.

  If the USE_THREAD_FOR_RETRY is set to 1, this function is __mppLnx_scsi_done(). 
  If the the USE_THREAD_FOR_RETRY is set to 0, this function is mppLnx_scsi_done(). 
  The __mppLnx_scsi_done() will be executed from mpp_DCR kernel thread while 
  mppLnx_scsi_done() is executed from the Linux SCSI middle level's bottom half.

*  @pscp The proxy Scsi_Cmnd that is completed by a physical HBA driver
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
#if USE_THREAD_FOR_RETRY
void __mppLnx_scsi_done(Scsi_Cmnd * pSCp)
#else
void mppLnx_scsi_done(Scsi_Cmnd * pSCp)
#endif
{
   long                          mppAction;
   unsigned long                 roundTrip;
   RdacDeviceInformation_t       *RdacInfo;
   BYTE                          Controller;
   BYTE                          Path;
   BYTE                          Lun;
   LWORD                         mpp_status;
   SenseData_t                   *SenseData;
   SINT                          SenseLength;
   LWORD                         DataLengthXfered;
   MPP_TIME                      StartTime;
   mppLnx_ProxyRequestEntry_t    * prequest;
   Scsi_Cmnd                     *vSCp;
   mppLnx_PDeviceEntry_t         * pde = NULL;
   mppLnx_VDeviceEntry_t         * vde = NULL;
   int                           ioErrorLevel;
   int                           isCompleteIo = 0;
   MPP_TIME                      currentTime;
   LINT                          LockLevel;


   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Entering  mppLnx_scsi_done()\n"));
  
   prequest = mppLnx_get_virtualRequest(pSCp);
   
   vSCp = prequest->vcmnd;

   RdacInfo = mpp_ModuleArray[vSCp->target].RdacInfo;

   if(NULL == RdacInfo)
   {
      MPPLNX_VHBA_BUG();
   }
   
   if(NULL == vSCp)
   {
      MPPLNX_VHBA_BUG();
   }
   
   if(NULL == pSCp->device)
   {
      MPPLNX_VHBA_BUG();
   }
   if(NULL == vSCp->device)
   {
      MPPLNX_VHBA_BUG();
   }


   
   pde = mppLnx_get_PdeviceEntry(pSCp->device);
   vde = mppLnx_get_VdeviceEntry(vSCp->device);
   
   if(NULL == pde)
   {
      MPPLNX_VHBA_BUG();
   }
   if(NULL == vde)
   {
      MPPLNX_VHBA_BUG();
   }
   

   RdacInfo = pde->rdacInfo;
   Controller = pde->controllerSlot;
   Path = pde->pathId;
   Lun = pSCp->lun;
   SenseData = (SenseData_t *)pSCp->sense_buffer;

   SenseLength = sizeof(pSCp->sense_buffer);
   DataLengthXfered = pSCp->transfersize; /*FIX ME */
   StartTime = prequest->startTime;

   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "mppLnx_scsi_done() status %x v SN %ld pcmnd 0x%lx on ctr%d:path%d:Lun%d\n",
      pSCp->result,prequest->vcmnd->serial_number,
      (unsigned long) pSCp->serial_number,Controller,Path,Lun));

   /*
   * get the current time. It is used later
   */
   MPP_GETTIME(currentTime);
   mpp_status = mppLnx_Translate_Linux_Status(pSCp->result);
   
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "%s-%d:%d:%d mppStatus %s\n",(char *)RdacInfo->ModuleName,
         Controller,Path,Lun,mppStatusStrings[mpp_status]));
 
#ifdef MPP_DEBUG
   if(vSCp->cmnd[0] == SCSI_INQUIRY && 
         (vSCp->cmnd[1] & 0x01) == 0 )
   {
      printk("the Inquiry cmnd for virtual device %d:%d come back from physical device %d:%d:%d:%d\n",
         vSCp->target,vSCp->lun,
         pSCp->host->host_no,
         pSCp->channel,
         pSCp->target,
         pSCp->lun);
      

      printk("physical command status 0x%04x\n",
         pSCp->result);     
   }
   if(SenseData->Error_Code != 0)
   {
      printk(
         "Sense Data Dump: Response Code 0x%02x Sense-Key/ASC/ASCQ 0x%02x/0x%02x/0x%02x \n",
         SenseData->Error_Code,
         SenseData->Sense_Key,
         SenseData->ASC,
         SenseData->ASCQ);
   }
#endif

   if(mpp_status)
   {
      mppAction = mpp_AnalyseIoError(
         RdacInfo, 
         Controller,
         Path, 
         Lun, 
         mpp_status, 
			SenseData, 
         SenseLength, 
         DataLengthXfered,
         prequest->startTime,
			&prequest->SelectionTimeoutRetryCount, 
         &prequest->CommandTimeoutRetryCount,
         &prequest->UaRetryCount, 
         &prequest->RetryCount,
         prequest->ControllerIoWaitTime,
         prequest->ArrayIoWaitTime);

      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
         "%s-%d:%d:%d mppAction %s\n",(char *)RdacInfo->ModuleName,
         Controller,Path,Lun, mppActionStrings[mppAction]));
     
      switch( mppAction)
      {
         case MPP_NO_ERROR:
            /* no error. Complete the IO*/
            isCompleteIo = 1;
            roundTrip = currentTime - prequest->startTime;
            mppLnx_update_PdeviceStat(pde, MPPLNX_IO_SUCCESS, roundTrip);
            roundTrip = currentTime - prequest->vcmndStartTime;
            mppLnx_update_VdeviceStat(vde, MPPLNX_IO_SUCCESS, roundTrip);
           
            /*
            * free the memory of proxy request
            */
            kfree(prequest);
            MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
               "mppLnx_scsi_done() no error case\n"));
            break;
         case MPP_RETURN_ERROR:
            /* 
            * complete the IO with error. The middle level will receive an error
            * of the proxy command's
            * Shall we change the return status to DID_NO_CONNECT (selection timeout)?
            */
            MPP_ERRORLOGPRINT((MPP_ERR_FATAL,492,
               "IO FAILURE array %s ctrl %d path %d lun %d mpp_status %d\n",
               mpp_ModuleArray[RdacInfo->VirtualTargetId].ModuleName,
               Controller, Path,Lun, mpp_status));
            
            
            isCompleteIo = 1;
            roundTrip = currentTime - prequest->startTime;
            mppLnx_update_PdeviceStat(pde, MPPLNX_IO_FAILED, roundTrip);
            roundTrip = currentTime - prequest->vcmndStartTime;
            mppLnx_update_VdeviceStat(vde, MPPLNX_IO_FAILED, roundTrip);
            
            /*
            * free the memory of proxy request
            */
            kfree(prequest);
            break;
         case MPP_RETRY_DEC_COUNT:
         case MPP_RETRY_NO_DEC:
            /* we need to perform a retry for this condition
            * MPP_RETRY_DEC_COUNT will perform a re-try and decrease the retry counter
            * MPP_RETRY_NO_DEC will perform a retry without changing the retry counter
            */
            /* don't complete the IO */
            isCompleteIo = 0;
            /* 
            * don't release the physical command. We are going to re-send the
            * proxy command. As long as we don't release the proxy command, we
            * are still holding the command slot.
            */
            ioErrorLevel = MPPLNX_IO_FAILED;
            /* retry the command on the same path*/
            /*1. add the request back to list
             *2. copy data from virtual command to proxy command again
             *3. reset the proxy request's startTime
             *3. resend the command
             */

            MPP_GETTIME(prequest->startTime);

            roundTrip = currentTime - prequest->startTime;
            mppLnx_update_PdeviceStat(pde, MPPLNX_IO_FAILED, roundTrip);


            MPP_ERRORLOGPRINT((MPP_ERR_FATAL,493,
               "Cmnd failed-retry %d:%d:%d the same path vSCp SN %ld T%dL%d pcmnd SN H%dC%dT%dL%d\n",
                Controller,
                 Path, 
               Lun,
               prequest->vcmnd->serial_number,
               prequest->vcmnd->target,
               prequest->vcmnd->lun,
               prequest->pcmnd->host->host_no,
               prequest->pcmnd->channel,
               prequest->pcmnd->target,
               prequest->pcmnd->lun));
            /*
            * re-issue the command to the same physical device
            */
            mppLnx_RetrySamePath(
                 prequest->pcmnd->device,
                 prequest->vcmnd,
                 middle_level_timeoutfunc,
                 middle_level_scsi_done,
                 prequest->CommandTimeoutRetryCount,
                 prequest->RetryCount,
                 prequest->SelectionTimeoutRetryCount,
                 prequest->UaRetryCount,
                 prequest->vcmndStartTime,
                 prequest->ControllerIoWaitTime,
                 prequest->ArrayIoWaitTime);
            MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
               "mppLnx_scsi_done() retry on the \n"));
            kfree(prequest);
            break;
            
         case MPP_RETRY_ALTERNATE_CONTROLLER:
            /*
             * retry on alternative controller. Will select a new path and re-issue the
             * virtual IO
             * 1.don't complete the virtual IO
             * 2.release the physical command
             * 3.free proxy request's memory
             * 4.update physical device entry's statistics
             * 5.mark the ownership chanle of the volume
             * 6.issue the command to a new path
             */
              isCompleteIo = 0;

              roundTrip = currentTime - prequest->startTime;
              mppLnx_update_PdeviceStat(pde, MPPLNX_IO_FAILED, roundTrip);
              
              /*
              * Mark the change of the volume ownership so that the mpp_SelectPath
              * will pick up a path from the other controller
              */
              MPP_LOCK(RdacInfo->Lock, LockLevel);
              RdacInfo->RdacLuns[Lun].CurrentOwningPath =	Controller^1;
              MPP_UNLOCK(RdacInfo->Lock, LockLevel);

              printk("mppLnx_scsi_done() Y13\n");

              MPP_ERRORLOGPRINT((MPP_ERR_FATAL,494,
                 "Cmnd-failed try alt ctrl %d vcmnd SN %ld T%dL%d pcmnd SN H%dC%dT%dL%d\n",
                  Controller^1,
                  prequest->vcmnd->serial_number,
                  prequest->vcmnd->target,
                  prequest->vcmnd->lun,
                  prequest->pcmnd->host->host_no,
                  prequest->pcmnd->channel,
                  prequest->pcmnd->target,
                  prequest->pcmnd->lun));
              /* 
              *try on a new controller. We need to reset the ControllerIoWaitTime
              */
              prequest->ControllerIoWaitTime = currentTime;

              mppLnx_do_queuecommand(
                 prequest->vcmnd,
                 middle_level_timeoutfunc,
                 middle_level_scsi_done,
                 Controller,
                 Path,
                 prequest->CommandTimeoutRetryCount,
                 prequest->RetryCount,
                 prequest->SelectionTimeoutRetryCount,
                 prequest->UaRetryCount,
                 prequest->vcmndStartTime,
                 prequest->ControllerIoWaitTime,
                 prequest->ArrayIoWaitTime);
             
              kfree(prequest);
              MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
                 "mppLnx_scsi_done() dispatched to alternative controller\n"));
              break;


         case MPP_FAILOVER:
            printk("mppLnx_scsi_done() Y15\n");
            /*
            * fail the path
            */
            if( mpp_FailPath(RdacInfo, Controller,Path) )
            {
               /*
               *mpp_Failpath returns a value greater than 0 to indicate
               *that other paths to the controller are still OK
               *if returns 0, this case will fall to the controller fail-over
               *case.
               */
               /*
               * 1. don't complete the virtual IO
               * 2.release the physical command
               * 3.free proxy request's memory
               * 4.update physical device entry's statistics
               * 5.issue the command to a new path
               */
               isCompleteIo = 0;

               roundTrip = currentTime - prequest->startTime;
               mppLnx_update_PdeviceStat(pde, MPPLNX_IO_PATHFAILOVER, roundTrip);
               mppLnx_update_VdeviceStat(vde, MPPLNX_IO_PATHFAILOVER, roundTrip);
               printk("mppLnx_scsi_done() Y17\n");

               MPP_ERRORLOGPRINT((MPP_ERR_FATAL,495,
                  "Try a new path vcmnd SN %ld T%dL%d pcmnd SN H%dC%dT%dL%d\n",
                  prequest->vcmnd->serial_number,
                  prequest->vcmnd->target,
                  prequest->vcmnd->lun,
                  prequest->pcmnd->host->host_no,
                  prequest->pcmnd->channel,
                  prequest->pcmnd->target,
                  prequest->pcmnd->lun));

               mppLnx_do_queuecommand(
                 prequest->vcmnd,
                 middle_level_timeoutfunc,
                 middle_level_scsi_done,
                 Controller,
                 Path,
                 prequest->CommandTimeoutRetryCount,
                 prequest->RetryCount,
                 prequest->SelectionTimeoutRetryCount,
                 prequest->UaRetryCount,
                 prequest->vcmndStartTime,
                 prequest->ControllerIoWaitTime,
                 prequest->ArrayIoWaitTime);

               kfree(prequest);
               MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
                  "mppLnx_scsi_done() sent IO to the new path\n"));
               break;
            }/*intentionally fall to the next case */
            else
            {
                MPP_ERRORLOGPRINT((MPP_ERR_FATAL,496,
                   "No new path: fall to failover controller case vcmnd SN %ld T%dL%d pcmnd SN H%dC%dT%dL%d\n",
                  prequest->vcmnd->serial_number,
                  prequest->vcmnd->target,
                  prequest->vcmnd->lun,
                  prequest->pcmnd->host->host_no,
                  prequest->pcmnd->channel,
                  prequest->pcmnd->target,
                  prequest->pcmnd->lun));
            }

         case MPP_FAILOVER_CONTROLLER:
            /*
            * I believe we never get this case directly.
            */
            if(RdacInfo->ControllerInfo[Controller^1]->ControllerPresent && /* we have other cntl*/
               !RdacInfo->ControllerInfo[Controller^1]->Failed &&
               mpp_FailController(RdacInfo, Controller, Lun))
            {
               /* the mpp_FailController() function will issue failover command
               */
               /*
               * 1. don't complete the virtual IO
               * 2.release the physical command
               * 3.free proxy request's memory
               * 4.update physical device entry's statistics
               * 5.issue the command to a new path
               */
               isCompleteIo = 0;

               roundTrip = currentTime - prequest->startTime;
               mppLnx_update_PdeviceStat(pde, MPPLNX_IO_PATHFAILOVER, roundTrip);
               mppLnx_update_VdeviceStat(vde, MPPLNX_IO_CTRLFAILOVER, roundTrip);


               MPP_ERRORLOGPRINT((MPP_ERR_FATAL,497,
                  "Failed controller to %d. retry  vcmnd SN %ld T%dL%d pcmnd SN H%dC%dT%dL%d\n",
                  Controller^1,
                  prequest->vcmnd->serial_number,
                  prequest->vcmnd->target,
                  prequest->vcmnd->lun,
                  prequest->pcmnd->host->host_no,
                  prequest->pcmnd->channel,
                  prequest->pcmnd->target,
                  prequest->pcmnd->lun));
               
               /*
               * Mark the other controller as the owner of this lun
               */
               MPP_LOCK(RdacInfo->Lock, LockLevel);
               RdacInfo->RdacLuns[Lun].CurrentOwningPath =	Controller^1;
               MPP_UNLOCK(RdacInfo->Lock, LockLevel);
               /* 
               *try on a new controller. We need to reset the ControllerIoWaitTime
               */
               prequest->ControllerIoWaitTime = currentTime;

               mppLnx_do_queuecommand(
                 prequest->vcmnd,
                 middle_level_timeoutfunc,
                 middle_level_scsi_done,
                 Controller,
                 Path,
                 prequest->CommandTimeoutRetryCount,
                 prequest->RetryCount,
                 prequest->SelectionTimeoutRetryCount,
                 prequest->UaRetryCount,
                 prequest->vcmndStartTime,
                 prequest->ControllerIoWaitTime,
                 prequest->ArrayIoWaitTime);
               
               kfree(prequest);
               MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
                  "mppLnx_scsi_done() cmd sent to the other controller after ctl failover\n"));
            }
            else
            {
               /*
               * can not failover controller. We have to fail the virtual IO
               */
               /*
               * 1. complete the IO (failed)
               * 2. update statistics (both physical device entry and virtual
               *    device entry
               * 3. release the physical command
               * 4. free proxy request memory
               */
               isCompleteIo = 1;

               roundTrip = currentTime - prequest->startTime;
               mppLnx_update_PdeviceStat(pde, MPPLNX_IO_FAILED, roundTrip);
               roundTrip = currentTime - prequest->vcmndStartTime;
               mppLnx_update_VdeviceStat(vde, MPPLNX_IO_FAILED, roundTrip);
               printk("mppLnx_scsi_done() Y28\n");

               MPP_ERRORLOGPRINT((MPP_ERR_FATAL,498,
                  "Can not failctrl. IO failed. vcmnd SN %ld T%dL%d pcmnd SN H%dC%dT%dL%d\n",
                  prequest->vcmnd->serial_number,
                  prequest->vcmnd->target,
                  prequest->vcmnd->lun,
                  prequest->pcmnd->host->host_no,
                  prequest->pcmnd->channel,
                  prequest->pcmnd->target,
                  prequest->pcmnd->lun));
               
               /*
               * free the memory of proxy request
               */
               kfree(prequest);
            }
            break;

         case MPP_FAILOVER_TO_CURRENT:
            /*
            * The command we sent came back with sense data that indicate the
            * controller does not own the LUN. But we think this controller should
            * own the LUN. We need to tell the other controller to failover to this
            * controller
            * The alternative controller must be maked as un-failed. Let the
            * mpp_FailController() to change its state
            */
            MPP_LOCK(RdacInfo->Lock, LockLevel);
				RdacInfo->ControllerInfo[Controller^1]->Failed = FALSE;
				MPP_UNLOCK(RdacInfo->Lock, LockLevel);
            
            if(mpp_FailController(RdacInfo, Controller^1, Lun))
            {
               
               /* the mpp_FailController() function will issue failover command
               */
               /*
               * 1. don't complete the virtual IO
               * 2.release the physical command
               * 3.free proxy request's memory
               * 4.update physical device entry's statistics
               * 5.issue the command to a new path
               */
               isCompleteIo = 0;
             

               roundTrip = currentTime - prequest->startTime;
               mppLnx_update_PdeviceStat(pde, MPPLNX_IO_FAILED, roundTrip);

               MPP_ERRORLOGPRINT((MPP_ERR_FATAL,500,
                  "Fail to current ctrl %d.vcmnd SN %ld T%dL%d pcmnd SN H%dC%dT%dL%d\n",
                  Controller,
                  prequest->vcmnd->serial_number,
                  prequest->vcmnd->target,
                  prequest->vcmnd->lun,
                  prequest->pcmnd->host->host_no,
                  prequest->pcmnd->channel,
                  prequest->pcmnd->target,
                  prequest->pcmnd->lun));

               mppLnx_do_queuecommand(
                 prequest->vcmnd,
                 middle_level_timeoutfunc,
                 middle_level_scsi_done,
                 Controller,
                 Path,
                 prequest->CommandTimeoutRetryCount,
                 prequest->RetryCount,
                 prequest->SelectionTimeoutRetryCount,
                 prequest->UaRetryCount,
                 prequest->vcmndStartTime,
                 prequest->ControllerIoWaitTime,
                 prequest->ArrayIoWaitTime);

               kfree(prequest);
               MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
                  "mppLnx_scsi_done() failover-to-current. sent the command\n"));
            }
            else
            {
               /*
               * can not failover controller. We have to fail the virtual IO
               * Why do we get here. We shouldn't get here at all.
               */
               /*
               * 1. complete the IO (failed)
               * 2. update statistics (both physical device entry and virtual
               *    device entry
               * 3. release the physical command
               * 4. free proxy request memory
               */
               isCompleteIo = 1;

               roundTrip = currentTime - prequest->startTime;
               mppLnx_update_PdeviceStat(pde, MPPLNX_IO_FAILED, roundTrip);
               roundTrip = currentTime - prequest->vcmndStartTime;
               mppLnx_update_VdeviceStat(vde, MPPLNX_IO_FAILED, roundTrip);

               MPP_ERRORLOGPRINT((MPP_ERR_FATAL,501,
                  "Cannot fail to current ctrl vcmnd SN %ld T%dL%d pcmnd SN H%dC%dT%dL%d\n",
                  prequest->vcmnd->serial_number,
                  prequest->vcmnd->target,
                  prequest->vcmnd->lun,
                  prequest->pcmnd->host->host_no,
                  prequest->pcmnd->channel,
                  prequest->pcmnd->target,
                  prequest->pcmnd->lun));

               /*
               * free the memory of proxy request
               */
               kfree(prequest);
            }
      }/* switch */

   }/*if mpp_status */
   else
   {
      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
         "mppLnx_scsi_done() no-error no-CC\n"));
      isCompleteIo = 1;
      RdacInfo->RdacLuns[Lun].QuiescienceStart = 
			RdacInfo->RdacLuns[Lun].NotReadyStart	=	 
			RdacInfo->RdacLuns[Lun].BusyStart	= 0;
		RdacInfo->RdacLuns[Lun].Busy = FALSE;
		RdacInfo->RdacLuns[Lun].NotReady = FALSE;
		RdacInfo->RdacLuns[Lun].Quiescent = FALSE;

      
      /*
      * free the memory of proxy request
      */
      roundTrip = currentTime - prequest->startTime;
      mppLnx_update_PdeviceStat(pde, MPPLNX_IO_SUCCESS, roundTrip);
      roundTrip = currentTime - prequest->vcmndStartTime;
      mppLnx_update_VdeviceStat(vde, MPPLNX_IO_SUCCESS, roundTrip);

      kfree(prequest);
           
   }
  
   if(isCompleteIo)
   {  
      /*
      * copy the data back to the virtual command
      */
      mppLnx_scsicmnd_proxy_to_virtual(vSCp,pSCp);
      /*
      * check the command to see whether or not it is a standard inquiry
      * if so, we need to change its vendor ID and product ID
      */
      if(vSCp->cmnd[0] == SCSI_INQUIRY && 
         (vSCp->cmnd[1] & 0x01) == 0 )
      {
         MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
            "mppLnx_scsi_done- replace vendorId ProductId \n"));
         mpp_VirtualdiskInquiryData(vSCp->buffer);
      }
   
      scsi_add_timer(vSCp, vSCp->timeout_per_command, middle_level_timeoutfunc);

      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
         "VHBA complete IO Sn = %ld vl T%dL%d\n",
         vSCp->serial_number,
         vSCp->target,
         vSCp->lun));
      
      vSCp->scsi_done(vSCp);

   }
   /*
   * release the physical command so that it is ready for the next IO
   */
   scsi_release_command(pSCp);
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Exiting mppLnx_scsi_done()\n"));
}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_scsi_init_proxy_cmnd
* SUMMARY:  Sets a proxy command to its preper initialization states
* SCOPE:     local
*
*
* DESCRIPTION:
   After the Linux Scsi middle level allocates a proxy Scsi_Cmnd from 
   a Scsi_Device's command pool, This function initializes a proxy command 
   to its proper states. A proxy Scsi command's proper states before dispatching
   to a physical HBA driver are specified in the document 349-1034610

*  @Scpnt a proxy Scsi_Cmnd that needs to be initialized.
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
static void mppLnx_scsi_init_proxy_cmnd(Scsi_Cmnd *SCpnt)
{
   
	SCpnt->owner = SCSI_OWNER_LOWLEVEL;
	SCpnt->reset_chain = NULL;
   
	SCpnt->serial_number_at_timeout = 0;
	SCpnt->flags = 0;
	SCpnt->retries = 0;

	SCpnt->abort_reason = 0;

	memset((void *) SCpnt->sense_buffer, 0, sizeof SCpnt->sense_buffer);
	/*
	 * We need saved copies of a number of fields - this is because
	 * error handling may need to overwrite these with different values
	 * to run different commands, and once error handling is complete,
	 * we will need to restore these values prior to running the actual
	 * command.
	 */
	SCpnt->old_use_sg = SCpnt->use_sg;
	SCpnt->old_cmd_len = SCpnt->cmd_len;
	SCpnt->sc_old_data_direction = SCpnt->sc_data_direction;
	SCpnt->old_underflow = SCpnt->underflow;
	memcpy((void *) SCpnt->data_cmnd,
	       (const void *) SCpnt->cmnd, sizeof(SCpnt->cmnd));
	SCpnt->buffer = SCpnt->request_buffer;
	SCpnt->bufflen = SCpnt->request_bufflen;

	SCpnt->reset_chain = NULL;

	SCpnt->internal_timeout = NORMAL_TIMEOUT;
	SCpnt->abort_reason = 0;

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_scsicmnd_virtual_to_proxy>
* SUMMARY:  Copy data from a virtual Scsi_Cmnd to a proxy Scsi_Cmnd
* SCOPE:    local
*
*
* DESCRIPTION:
   Before a proxy command is issued to a physical HBA driver, this function 
   copies certain fields from the virtual Scsi_Cmnd object,  "*vcmnd", to 
   the proxy Scsi_Cmnd object, "*pcmnd". Which fields in the virtual command 
   need to be copied to the proxy command are specified in 349-1034610


*  @vscp the virtual Scsi_Cmnd
*  @pscp the Proxy Scsi_Cmnd
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
static void mppLnx_scsicmnd_virtual_to_proxy(Scsi_Cmnd *vscp, Scsi_Cmnd *pscp)
{
   pscp->serial_number_at_timeout = 0;
   pscp->retries = 0;
   pscp->allowed = 1;
   pscp->timeout_per_command = vscp->timeout_per_command;
   pscp->timeout = vscp->timeout;
   pscp->internal_timeout = vscp->internal_timeout;
   pscp->cmd_len = vscp->cmd_len;
   pscp->old_cmd_len = vscp->old_cmd_len;
   pscp->request_bufflen = vscp->request_bufflen;
   pscp->request_buffer = vscp->request_buffer;
   pscp->bounce_buffers = vscp->bounce_buffers;
   memcpy(pscp->data_cmnd, vscp->data_cmnd, MAX_COMMAND_SIZE);
   pscp->old_use_sg = vscp->old_use_sg;
   pscp->use_sg = vscp->use_sg;
   pscp->sglist_len = vscp->sglist_len;
   pscp->abort_reason = vscp->abort_reason;
   pscp->bufflen = vscp->bufflen;
   pscp->buffer = vscp->buffer;
   pscp->underflow = vscp->underflow;
   pscp->old_underflow = vscp->old_underflow;
   pscp->transfersize = vscp->transfersize;
   pscp->resid = vscp->resid;
   memset(pscp->sense_buffer, 0 , SCSI_SENSE_BUFFERSIZE);
   pscp->flags = vscp->flags;
   pscp->done_late = vscp->done_late;
   pscp->result = vscp->result;
   pscp->sc_data_direction = vscp->sc_data_direction;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_scsicmnd_proxy_to_virtual
* SUMMARY:  Copy data from a proxy Scsi_Cmnd to a virtual Scsi_Cmnd
* SCOPE:    local
*
* DESCRIPTION:
   After a proxy command is completed, this function copies certain fields from 
   the proxy Scsi_Cmnd object,  "*pcmnd", to the virtual Scsi_Cmnd object, "*vcmnd". 
   Which fields in the proxy command need to be copied to the virtual command are 
   specified in 349-1034610.

*  @vscp the virtual Scsi_Cmnd (destination)
*  @pscp the proxt Scsi_Cmnd (source)
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
static void mppLnx_scsicmnd_proxy_to_virtual(Scsi_Cmnd *vscp, Scsi_Cmnd *pscp)
{
   vscp->state = SCSI_OWNER_LOWLEVEL;
   /*
   * set the retries to the max allowed. A virtual command does
   * not need to be retried
   */
   vscp->result = pscp->result;
   vscp->retries = vscp->allowed;
   vscp->request_buffer = pscp->request_buffer;
   vscp->bounce_buffers = pscp->bounce_buffers;
   vscp->buffer = pscp->buffer;
   vscp->underflow = pscp->underflow;
   vscp->old_underflow = pscp->old_underflow;
   vscp->resid = pscp->resid;
   memcpy(vscp->sense_buffer, pscp->sense_buffer, SCSI_SENSE_BUFFERSIZE);

   return;
}
