/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************

NAME            mppLnx_vhbamisc.c
SUMMARY         %description%
VERSION         %version: 6 %
UPDATE DATE     %date_modified: Tue Nov 18 17:33:57 2003 %
PROGRAMMER      %created_by:    yqi %

        Copyright 2003, LSI Logic Storage Systems, Inc.

DESCRIPTION:This file contains miscellaneous functions that are used in the 
Linux MPP virtual HBA driver.

INCLUDE FILES:

NOTES:

RESTRICTIONS:

SEE ALSO:

REFERENCE:
  349-1033500 Linux MPP Driver Feature Architecture Specification
  349-1034610 Linux MPP Virtual HBA Driver Feature Design Document
  349-1034750 Linux MPP Upper Level Driver Feature Design Document
  349-1034730 Linux MPP Utility Feature Design Document
  349-1034620 Linux MPP Installation Feature Design Document


IMPLEMENTATION:

MODIFICATION HISTORY:

*******************************************************************************/

#define __SRCmppLnx_vhbamisc_c

/***  INCLUDES  ***/  
/***  Linux kernel INCLUDES  ***/  
#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/genhd.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/interrupt.h>

#include <asm/io.h>

#include <linux/blk.h>
#include "scsi.h"
#include "hosts.h"
#include "sd.h"

#include <linux/stat.h>

/***  MPP commomn API INCLUDES  ***/  
#include "MPP_Common.h"
#include "MPP_RdacInfo.h"
#include "MPP_ProtoTypes.h"
#include "MPP_Sysdep.h"
#include "mppLnx_version.h"
/***  Linux MPP Virtual HBA INCLUDES  ***/  
#include "mppLnx_vhba.h"
#include "mppLnx_shared.h"


/***  CONSTANT DEFINITIONS  ***/


/***  MACRO DEFINITIONS  ***/


/***  TYPE DEFINITIONS  ***/


/***  LOCALS  ***/


/***  PROCEDURES  ***/

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_regSymbols
* SUMMARY:  Makes inter-module communication registration to register functions
            with the kernel.
* SCOPE:   local
*
*
* DESCRIPTION:
      Makes the virtual HBA driver's functions available to the Linux MPP Upper
      level driver. This function is invoked by mppLnx_detect() function.
* 
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*/
void mppLnx_regSymbols()
{
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_regSymbols()\n"));
   inter_module_register(MPPLNX_NEWDENOTIFICATION_KEY,THIS_MODULE,mppLnx_newDeviceAdded);
   inter_module_register(MPPLNX_ISSUEFAILOVER_KEY,THIS_MODULE,mppLnx_IssueFailover);
   inter_module_register(MPPLNX_FAILBACKRESCAN_KEY,THIS_MODULE,mppLnx_failbackRescan);
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exitering mppLnx_regSymbols()\n"));
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_unregSymbols
* SUMMARY:  Makes inter-module communication unregistration.
* SCOPE:   local
*
*
* DESCRIPTION:
      Unregister the virtual HBA functions with the kernel. This function is
      invoked by the mppLnx_release() function.
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*/
void mppLnx_unregSymbols()
{
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_unregSymbols()\n"));
   inter_module_unregister(MPPLNX_NEWDENOTIFICATION_KEY);
   inter_module_unregister(MPPLNX_ISSUEFAILOVER_KEY);
   inter_module_unregister(MPPLNX_FAILBACKRESCAN_KEY);
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exitering mppLnx_unregSymbols()\n"));
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_dumpCmnd
* SUMMARY:  Dumps the content of a Scsi_Cmnd object.
* SCOPE:   local
*
*
* DESCRIPTION:
   Dumps the content of a Scsi_Cmnd object. This function is used only for
   debugging

* @SCnp The Scsi_Cmnd object to be dumped.
* RESTRICTIONS:
*   This function is only complied to the driver module when MPP_DEBUG is on.
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
#ifdef MPP_DEBUG

void mppLnx_dumpCmnd(Scsi_Cmnd * SCnp)
{
   char tmpBuffer[4096];
   int len;
   printk("Scsi_Cmnd dump:\n");
   printk("\tcommand id. device h%dc%dt%dl%d\n",
      SCnp->host->host_no, SCnp->channel, SCnp->target, SCnp->lun);
   printk("\tserial number %ld\n", SCnp->serial_number);
   printk("\t abort_reason %d\n",SCnp->abort_reason);
   printk("\t allowed %d\n",SCnp->allowed);

   printk("\t next address 0x%lx\n", (unsigned long)SCnp->next);
   printk("\t bh_next address 0x%lx\n", (unsigned long)SCnp->bh_next);

   /* the bounce_buffers are an array of pointers
   */
   printk("\t bounce_buffers address 0x%lx\n",(unsigned long)SCnp->bounce_buffers);

   printk("\tbufflen %d\n",SCnp->bufflen);
   if(SCnp->bufflen >2048)
   {
      len = 2048;
   }else
   {
      len = SCnp->bufflen;
   }
   memset(tmpBuffer,0,4096);
   mppLnx_BytestoAscii(tmpBuffer,SCnp->buffer, len);
   printk("\t buffer %s\n", tmpBuffer);

   printk("\t cmd_len %d\n",SCnp->cmd_len);

   memset(tmpBuffer,0, 4096);
   mppLnx_BytestoAscii(tmpBuffer, SCnp->cmnd, SCnp->cmd_len);
   printk("\t cmnd %s\n",tmpBuffer);
   
   memset(tmpBuffer, 0, 4096);
   mppLnx_BytestoAscii(tmpBuffer, SCnp->data_cmnd, sizeof(SCnp->data_cmnd));
   printk("\t data_cmnd %s",tmpBuffer);

   printk("\t eh_state %d\n",SCnp->eh_state);
   printk("\t flags %d\t", SCnp->flags);
   
   printk("\t internal_timeout %d\n",  SCnp->internal_timeout);
   printk("\t old_cmdlen %d\n",SCnp->old_cmd_len);
   printk("\t old_underflow %d\n",SCnp->old_underflow);
   printk("\t old_use_sg %d\n", SCnp->old_use_sg);
   printk("\t ower %x\n", SCnp->owner);
   printk("\t pid %ld\n", SCnp->pid);

   
   printk("\t request struct address 0x%lx\n",(unsigned long)&SCnp->request);

   printk("\t request_bufflen %d\n", SCnp->request_bufflen);

   if(SCnp->request_bufflen >2048)
   {
      len = 2048;
   }
   else
   {
      len = SCnp->request_bufflen;
   }
   memset(tmpBuffer, 0, 4096);
   mppLnx_BytestoAscii(tmpBuffer, SCnp->request_buffer, len);
   printk("\t request_buffer %s\n", tmpBuffer);
   
   printk("\t reset_chaind address 0x%ld\n", (unsigned long)SCnp->reset_chain);
   printk("\tn resid %d\n",SCnp->resid);
   printk("\t result 0x%d\n",SCnp->result);
   printk("\t retires %d\n",SCnp->retries);
   printk("\t sc_data_direction 0x%x\n",SCnp->sc_data_direction);

   
   printk("\t sc_magic 0x%x\n", SCnp->sc_magic);

   printk("\t sc_old_data_direction 0x%x\n",SCnp->sc_old_data_direction);
   
   printk("\t sc_request struct address 0x%lx\n",(unsigned long)SCnp->sc_request);
   printk("\t SCp struct address 0x%lx\n",(unsigned long)&SCnp->SCp);

   memset(tmpBuffer, 0 , 4096);
   len = sizeof(SCnp->sense_buffer);
   mppLnx_BytestoAscii(tmpBuffer, SCnp->sense_buffer, len);
   printk("\t sense_buffer %s\n",tmpBuffer);
   
   printk("\t scsi_done function address 0x%lx\n",(unsigned long)SCnp->scsi_done);
   printk("\t serial_number_at_timeout %ld \n",SCnp->serial_number_at_timeout);
   printk("\t sglist_len %d\n", SCnp->sglist_len);
   printk("\t state 0x%x\n",SCnp->state);
   printk("\t tag %d\n",SCnp->tag);
   printk("\t timeout %d seconds\n",SCnp->timeout/HZ);
   printk("\t timeout_per_command %d seconds \n",SCnp->timeout_per_command/HZ);
   printk("\t timeout_total %d \n", SCnp->timeout_total);
   printk("\t transfersize %d\n",SCnp->transfersize);
   printk("\t underflow %d\n", SCnp->underflow);
   printk("\t use_sg %d\n", SCnp->use_sg);

}

#endif
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_BytestoAscii
* SUMMARY:  Converts a sequence of bytes to its ASCII representation
* SCOPE:   local
    local
*
*
* DESCRIPTION:
   This function takes a char buffer as input and converts the buffer to 
   ASCII string.

* @ascii The ascii string of the input buff. The length of ascii must be equale
         or greater than 2*bufflen. Output
  @buff  a char buffer whose content will be converted to ascii string. Input
  @bufflen the length of the buff. Input
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_BytestoAscii(char *ascii, char *buff, int bufflen)
{
	int	count;

	for(count = 0; count < bufflen; count++)
		sprintf(&ascii[count*2], "%02x", buff[count]);
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_dump_mppLnx_PDeviceEntry
* SUMMARY:  Dump the content of a mppLnx_PDeviceEntry_t object
* SCOPE:   local
*
*
* DESCRIPTION:
   This function dumps the content of a mppLnx_PDeviceEntry_t object.
   It is used for debugging only

* @pde a pointer to an mppLnx_PDeviceEntry_t object whose the content
   will be dumped. 
     
* RESTRICTIONS:
This function is only complied to the driver module when MPP_DEBUG is on.
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
#ifdef MPP_DEBUG

void mppLnx_dump_PDeviceEntry(mppLnx_PDeviceEntry_t * pde)
{
   printk("mppLnx_PDeviceEntry dump:\n");

   printk("\tcontrollerSlot %d\n",pde->controllerSlot);

   mppLnx_dump_proxyRequestList(pde->currentCmndList);

   printk("\tioCount %d\n",pde->ioCount);
   printk("\tioFails %d\n",pde->ioFails);
   printk("\tlongestTrip %d\n",pde->longestTrip);
   printk("\tpathId %d\n",pde->pathId);
   printk("\tporcNode Name %s\n",(char *)pde->procNode.nodeName);
   printk("\t Array name %s\n",(char *)pde->rdacInfo->ModuleName);
   printk("\tshortestTrip %d\n",pde->shortestTrip);
   printk("\tphysical Device h%dc%dt%dl%d\n",
      pde->pdevice->host->host_no, pde->pdevice->channel,
      pde->pdevice->id, pde->pdevice->lun);
}
#endif

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_dump_proxyRequestList
* SUMMARY:  Dump the content of a list of mppLnx_ProxyRequest_t objects
* SCOPE:   local
*
*
* DESCRIPTION:
   This function dumps the context of a list of mppLnx_ProxyRequest_t objects.
   This function is only for debugging

* @requestList a pointer to the first entry of the mppLnx_ProxyRequest_t object
* RESTRICTIONS:
* This function is only complied to the driver module when MPP_DEBUG is on.
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/

#ifdef MPP_DEBUG

void mppLnx_dump_proxyRequestList(mppLnx_ProxyRequestEntry_t * requestList)
{
   struct list_head * listPt = NULL;
   mppLnx_ProxyRequestEntry_t * proxyRequest = NULL;

   if(requestList->list.next == &(requestList->list))
   {
      printk("Proxy Request List Dump: Empty list\n");
   }
   else
   {
      printk("Proxy Request List Dump:\n");
      for(listPt = requestList->list.next;
          listPt != &(requestList->list);
          listPt = listPt->next)
      {
         proxyRequest = list_entry(listPt, mppLnx_ProxyRequestEntry_t, list);
         printk("\tProxy Request Entry:\n");
         printk("\t\tphysical command %d h%dc%dt%dl%d SN %ld\n", 
            proxyRequest->pcmnd->cmnd[0],
            proxyRequest->pcmnd->host->host_no,
            proxyRequest->pcmnd->channel,
            proxyRequest->pcmnd->target,
            proxyRequest->pcmnd->lun,
            proxyRequest->pcmnd->serial_number);
         printk("\t\tvirtual command %d h%dc%dt%dl%d SN %ld\n", 
            proxyRequest->vcmnd->cmnd[0],
            proxyRequest->vcmnd->host->host_no,
            proxyRequest->vcmnd->channel,
            proxyRequest->vcmnd->target,
            proxyRequest->vcmnd->lun,
            proxyRequest->vcmnd->serial_number);

            printk("\t\tStartTime %ld\n", proxyRequest->startTime);
      }
   }
}

#endif
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_dump_proxyRequestList
* SUMMARY:  Dump the content of a list of mppLnx_ProxyRequest_t objects
* SCOPE:   local
*
*
* DESCRIPTION:
   This function dumps the context of a list of mppLnx_ProxyRequest_t objects.
   This function is only for debugging

* @requestList a pointer to the first entry of the mppLnx_ProxyRequest_t object
* RESTRICTIONS:
This function is only complied to the driver module when MPP_DEBUG is on.
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/

#ifdef MPP_DEBUG

void mppLnx_dump_Scsi_Host_Template(Scsi_Host_Template *sht)
{
   printk("======== Scsi_Host_Template %s ========\n",sht->name);
   
   printk("       sht->can_queue = %d\n",    sht->can_queue);
   printk("       sht->cmd_per_lun = %d\n",    sht->cmd_per_lun);
   printk("       sht->emulated = %d\n",    sht->emulated);
  
   printk("       sht->max_sectors = %d\n",    sht->max_sectors);
   printk("       sht->name = %s\n",    sht->name);
   printk("       sht->present = %d\n",    sht->present);
   printk("       sht->sg_tablesize = %d\n",    sht->sg_tablesize);
   printk("       sht->this_id = %d\n",    sht->this_id);
   printk("       sht->unchecked_isa_dma = %d\n",    sht->unchecked_isa_dma);
   printk("       sht->use_clustering = %d\n",    sht->use_clustering);
   printk("       sht->use_new_eh_code = %d\n",    sht->use_new_eh_code);

}

#endif

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_dump_proxyRequestList
* SUMMARY:  Dump the content of a list of mppLnx_ProxyRequest_t objects
* SCOPE:   local
*
*
* DESCRIPTION:
   This function dumps the context of a list of mppLnx_ProxyRequest_t objects.
   This function is only for debugging

* @requestList a pointer to the first entry of the mppLnx_ProxyRequest_t object
* RESTRICTIONS:
This function is only complied to the driver module when MPP_DEBUG is on.
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/

#ifdef MPP_DEBUG

void mppLnx_dump_Scsi_Host(struct Scsi_Host *shp)
{
   printk("=== Scsi_Host %d ======\n",shp->host_no);
   printk("       shp->base = %ld\n",    shp->base);
   printk("       shp->can_queue = %d\n",    shp->can_queue);
   printk("       shp->cmd_per_lun = %d\n",    shp->cmd_per_lun);
   printk("       shp->dma_channel = %d\n",    shp->dma_channel);
   printk("       shp->extra_bytes = %d\n",    shp->extra_bytes);
   printk("       shp->host_no = %d\n",    shp->host_no);
   printk("       shp->host_blocked = %d\n",    shp->host_blocked);
   printk("       shp->host_self_blocked = %d\n",    shp->host_self_blocked);
   printk("       shp->host_failed = %d\n",    shp->host_failed);
   printk("       shp->in_recovery = %d\n",    shp->in_recovery);
   printk("       shp->io_port = %ld\n",    shp->io_port);
   printk("       shp->last_reset = %ld\n",    shp->last_reset);
   printk("       shp->loaded_as_module = %d\n",    shp->loaded_as_module);
   printk("       shp->max_channel = %d\n",    shp->max_channel);
   printk("       shp->max_channel = %d\n",    shp->max_channel);
   printk("       shp->max_cmd_len = %d\n",    shp->max_cmd_len);
   printk("       shp->max_lun = %d\n",    shp->max_lun);
   printk("       shp->max_id = %d\n",    shp->max_id);
   printk("       shp->max_sectors = %d\n",    shp->max_sectors);
   printk("       shp->sg_tablesize = %d\n",    shp->sg_tablesize);
   printk("       shp->this_id = %d\n",    shp->this_id);
   printk("       shp->unchecked_isa_dma = %d\n",    shp->unchecked_isa_dma);
   printk("       shp->unique_id = %d\n",    shp->unique_id);
   printk("       shp->use_clustering = %d\n",    shp->use_clustering);

}
#endif
