/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************

NAME            mppLnx_prototypes.h
SUMMARY         %description%
VERSION         %version: 25.1.2 %
UPDATE DATE     %date_modified: Tue Feb 03 15:59:02 2004 %
PROGRAMMER      %created_by:    yqi %
        Copyright 2003, LSI Logic Storage Systems, Inc.

DESCRIPTION:
This header file contains all definitions that are used only 
in the Linux Upper Level driver. Also, the definitions, which are 
defined in this header file, are not used by the MPP common APIs.

INCLUDE FILES:

NOTES:

RESTRICTIONS:

SEE ALSO:

REFERENCE:

IMPLEMENTATION:

MODIFICATION HISTORY:

*******************************************************************************/


#ifndef __INCmppLnx_prototypes
#define __INCmppLnx_prototypes


/***  INCLUDES  ***/  
#ifndef __KERNEL__
#define __KERNEL__
#endif

#include "MPP_ProtoTypes.h"
#include "mppCmn_s2tos3.h"
#include <linux/list.h>


/***  CONSTANT DEFINITIONS  ***/

#define MPP_UTM_LUN_DEVICE   1001

/*** macro definition used for function pointer replacement ***/
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CONFIG_FILE_NAME "/etc/mpp.conf"
#define BINDING_FILE_NAME "/var/mpp/devicemapping"

/***  TYPE DEFINITIONS  ***/
typedef int (*sdtpntFuncPtr) (Scsi_Device*);

/*
* A data structure that handles unconfigured device list
*/
typedef struct mppLnx_UnconfiguredDeviceEntry
{
   struct list_head           list;
   Scsi_Device                *physical_device;
   unsigned int               isUnconfigured:1,
                              isAttached:1,
                              isDetected:1,
                              reserved:29;
} mppLnx_UnconfiguredDeviceEntry_t;

typedef struct mppLnx_UnconfiguredDeviceList
{
   spinlock_t                          *unconfiguredDeviceLock;
   mppLnx_UnconfiguredDeviceEntry_t    *deviceEntryList;
}  mppLnx_UnconfiguredDeviceList_t;



/***  FUNCTIONS DECLARATIONS ***/
extern BYTE mpp_IsFailoverPossible(RdacDeviceInformation_t *RdacInfo, BYTE Controller, BYTE Path, BYTE Lun, BYTE level);

extern BYTE mpp_FailController(RdacDeviceInformation_t *RdacInfo, BYTE Controller, BYTE Lun);

extern BYTE mpp_IsAnotherPathAvailableForFailover(RdacDeviceInformation_t *RdacInfo, BYTE Controller, BYTE Path, BYTE Lun); 


/* MPP upper level driver initialization and cleanup routines */
static int mppLnx_init(void);
static void mppLnx_exit(void);

/* MPP upper level driver initialization helper functions */
static int mppLnx_readConfigData(void);
static int mppLnx_readBindingData(void);
static void mppLnx_modifyConfigParameter(char *, char *);
static void mppLnx_FillInPersistentArray(char*, char* );
static char* mppLnx_strstrip(char *, char *);
static void mppLnx_parseLine(char **, char *, char **, char **);
static BOOL mppLnx_simple_strtoul(const char*, char**, unsigned int, unsigned long *);

/* The following functions are interceptor functions */
static int mppLnx_replaceFuncPtr (char*, sdtpntFuncPtr, sdtpntFuncPtr);
static int mppLnx_sd_detect (Scsi_Device*);
static int mppLnx_sd_attach (Scsi_Device*);
static int mppLnx_sg_detect (Scsi_Device*);
static int mppLnx_sg_attach (Scsi_Device*);

/* Function declarations for Scsi_Device_Template */
static int  mppLnx_target_init(void);
static void mppLnx_target_finish(void);
static int  mppLnx_target_attach(Scsi_Device *);
static int  mppLnx_target_detect(Scsi_Device *);
static void mppLnx_target_detach(Scsi_Device *);
static int  mppLnx_target_init_command(Scsi_Cmnd *);

/* Function declarations for file operations */
static int mppLnx_open (struct inode *inode, struct file *file);
static int mppLnx_release (struct inode *inode, struct file *file);
static int mppLnx_ioctl( struct inode *inode, struct file *file, unsigned int cmd_in, unsigned long arg );

/* Function to identify the disk type for a Scsi_Device */
static LWORD mppLnx_DeviceType(Scsi_Device*);

/*Functions for Synchronous I/O */
static LWORD mppLnx_Translate_Linux_Status (int result);

/* inter-module communication functions */
static void mppLnx_newDeviceAdded (Scsi_Device *);

/* functions for managing un-configured devices */
static mppLnx_UnconfiguredDeviceEntry_t * mppLnx_getUnconfiguredDevice(Scsi_Device * device);
static void mppLnx_deleteUnconfiugredDevice(Scsi_Device * device);
static void mppLnx_addUnconfiguredDevice(Scsi_Device * device);
static LWORD 
mppLnx_BuildRdacinfo(MPP_HANDLE DeviceVertex, ControllerAddress_t *DeviceAddress, 
                     LWORD Lun);
static LWORD 
mppLnx_IsArrayDevice(MPP_HANDLE DeviceVertex, LWORD Lun);
static void mppLnx_RescanUnconfiguredDevices(void);


/* All the functions provided in the Upper Level driver module that need to be used by the Virtual HBA driver have to be exported */
EXPORT_SYMBOL(mpp_ModuleArray) ;
EXPORT_SYMBOL(mppLnx_Config_Parameters);
EXPORT_SYMBOL(mpp_FailBackScan) ;
EXPORT_SYMBOL(mpp_CheckAVTState );
EXPORT_SYMBOL(mpp_CheckPreferredAVTOwner );
EXPORT_SYMBOL(mpp_VirtualdiskInquiryData );
EXPORT_SYMBOL(mpp_FailPath );
EXPORT_SYMBOL(mpp_IsFailoverPossible); 
EXPORT_SYMBOL(mpp_FailController); 
EXPORT_SYMBOL(mpp_IsAnotherPathAvailableForFailover) ;
EXPORT_SYMBOL(mpp_AnalyseFailoverError );
EXPORT_SYMBOL(mpp_AnalyseIoError);
EXPORT_SYMBOL(mpp_ConvertWWNtoAscii);
EXPORT_SYMBOL(mpp_ErrorLog);
EXPORT_SYMBOL(mpp_GetTime);
EXPORT_SYMBOL(mpp_SelectFailoverPath);
EXPORT_SYMBOL(mpp_SelectPath);
EXPORT_SYMBOL(mppLnx_Translate_Linux_Status);
EXPORT_SYMBOL(mpp_DebugPrint);
EXPORT_SYMBOL(mppLnx_zfree);

/*
* SCSI2 to SCSI3 Reservation translation symbols
*/
EXPORT_SYMBOL(mppCmn_MapBusDeviceReset);
EXPORT_SYMBOL(mppCmn_GenReservationKey);
EXPORT_SYMBOL(mppCmn_MapScsi2Reserve);
EXPORT_SYMBOL(mppCmn_MapScsi2Release);
EXPORT_SYMBOL(mppCmn_S2toS3SynchCmnd);
EXPORT_SYMBOL(mppCmn_CheckForReservationHolder);

#endif        /* End of __INCmppLnx_prototypes */
