/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************

NAME            mppLnx_vhba.h
SUMMARY         %description%
VERSION         %version: 29 %
UPDATE DATE     %date_modified: Sun Feb 15 10:15:28 2004 %
PROGRAMMER      %created_by:    yqi %

        Copyright 2003, LSI Logic Storage Systems, Inc.

DESCRIPTION: This header file defines data structs and definitions for the Linux
               MPP Virtual HBA driver.

INCLUDE FILES:

NOTES:
   This header file must be included after all Linux kernel include files.

RESTRICTIONS:

SEE ALSO:

REFERENCE:

  349-1033500 Linux MPP Driver Feature Architecture Specification
  349-1034610 Linux MPP Virtual HBA Driver Feature Design Document
  349-1034750 Linux MPP Upper Level Driver Feature Design Document
  349-1034730 Linux MPP Utility Feature Design Document
  349-1034620 Linux MPP Installation Feature Design Document

IMPLEMENTATION:

MODIFICATION HISTORY:

*******************************************************************************/
#ifndef __INCmppLnx_vhba_h
#define __INCmppLnx_vhba_h

/***  INCLUDES  ***/  
/*
* To use a kernel thread to dispatch commands,set USE_THREAD_FOR_DPC to 1. 
* If setting USE_THREAD_FOR_DPC to 0, the MPP driver will use tasklet to perform command
* dispatch
*/
#define USE_THREAD_FOR_DPC          1
/*
* To use the kernel thread to complete commands and retry commands, set USE_THREAD_FOR_RETRY to 1.
* If setting USE_THREAD_FOR_RETRY to 0, the MPP driver will use the context of the SCSI bottom half to perform 
* command completion and retries.
* USE_THREAD_FOR_RETRY can only be set to 1 when USE_THREAD_FOR_DPC is set to 1.
*/
#define USE_THREAD_FOR_RETRY        1

/***  CONSTANT DEFINITIONS  ***/
#define MPPVHBASTRING "mpp virtual bus adaptor "
#define TIME_STAMP    "timestamp"
#define         MPPLNX_PROC_NAME "mpp"

/** the size of a proc leaf node name **/
#define NODENAMESIZE    512
#define PROCDIRNAMESIZE 1024
/** proc leaf node mode_t */
#define PROC_LEAF_NODE_MODE S_IFREG|S_IRUGO|S_IWUSR

/** proc leaf node type */
#define MPPLNX_VDEVICE_NODE 0
#define MPPLNX_PDEVICE_NODE 1

/* tasklet queue command types */
#define MPPLNX_DELAY_RESPONSE_TYPE 		0
#define MPPLNX_SCSI2TOSCSI3_TRAN_TYPE 	1	
#define MPPLNX_QUEUED_CMND_TYPE        2

#if USE_THREAD_FOR_RETRY
#define MPPLNX_COMPLETED_CMND_TYPE     3
void __mppLnx_scsi_done(Scsi_Cmnd * scp);
#endif

/* IO completion status */
#define MPPLNX_IO_SUCCESS              0
#define MPPLNX_IO_PATHFAILOVER         1
#define MPPLNX_IO_CTRLFAILOVER         2
#define MPPLNX_IO_FAILED               3

/*
*We need assign a scsi command op code for the bus device reset message. This is no 
* such command in the scsi spec. We need to make-up an op code. The op code 0xEE 
* is in the vendor specific command group and it will not conflict with other command.
*/
#define  MPPLNX_BDR_CMND_OPCODE       0xEE
/***  MACRO DEFINITIONS  ***/
/*
* panic the kernel
* Used for reporting a program error.
*/
#define MPPLNX_VHBA_BUG() do { \
   printk("mpp vhba BUG at %s:%d!\n", __FILE__, __LINE__); \
	* ((char *) 0) = 0; \
} while (0)


/*
* physical device proc read buffer size
*/
#define MPPLNX_PROC_READ_BUFFSIZE 8192
/* define the /proc/scsi/mpp/host_no proc info buffer size
* as 8*1024
*/
#define MPPLNX_PROC_INFO_BUFFSIZE 4095*5
#define WWN_LENGTH                  16

/*
* A virtual IO which is first time dispatched to the
* virtual HBA driver. Its previous controller is not-defined
*/
#define UNKNOWN_PREVIOUS_CONTROLLER    200

#define UNKNOWN_PREVIOUS_PATH          200

//#ifdef MPP_DEBUG FIX ME
extern char * mppStatusStrings[];

extern char * mppActionStrings[];


#ifdef MPP_DEBUG
#define MPPLNX_DUMP_SCSI_HOST_TEMPLATE(x)  \
                mppLnx_dump_Scsi_Host_Template(x)
#define MPPLNX_DUMP_SCSI_HOST(x)  \
                mppLnx_dump_Scsi_Host(x)
#define MPPLNX_DUMPCMND(x)   \
                mppLnx_dumpCmnd(x)
#define MPPLNX_DUMP_PROXYREQUESTLIST(x) \
                mppLnx_dump_proxyRequestList(x)
#define MPPLNX_DUMP_PDEVICEENTRY(x) \
                mppLnx_dump_PDeviceEntry(x)
#else
#define MPPLNX_DUMP_SCSI_HOST_TEMPLATE(x)
#define MPPLNX_DUMP_SCSI_HOST(x) 
#define MPPLNX_DUMPCMND(x)  
#define MPPLNX_DUMP_PROXYREQUESTLIST(x)
#define MPPLNX_DUMP_PDEVICEENTRY(x)
#endif


/***  TYPE DEFINITIONS  ***/


/***  LOCALS  ***/


/***  PROCEDURES  ***/



typedef struct mppLnx_ProcLeafNode
{
   int                        nodeType;
   char                       nodeName[NODENAMESIZE];
   struct proc_dir_entry      *parent;
   struct proc_dir_entry      *thisNode;
   int                        isCreated;
   int                        needsClean;
   mode_t                     procMode;
}  mppLnx_ProcLeafNode_t;




typedef struct mppLnx_ProxyRequestEntry
{
   struct list_head           list;
   unsigned long              serial_number;
   Scsi_Cmnd                  *vcmnd;
   Scsi_Cmnd                  *pcmnd;
   unsigned long              startTime;
   unsigned long              vcmndStartTime;
   BYTE                       SelectionTimeoutRetryCount;
   BYTE                       CommandTimeoutRetryCount;
   BYTE                       UaRetryCount;
   BYTE                       RetryCount;
   MPP_TIME                   ControllerIoWaitTime;
   MPP_TIME                   ArrayIoWaitTime;
}  mppLnx_ProxyRequestEntry_t;




typedef struct mppLnx_PDeviceEntry
{
   mppLnx_ProcLeafNode_t			   procNode;
   Scsi_Device                      *pdevice;
   RdacDeviceInformation_t          *rdacInfo;
   int                              controllerSlot;
   int                              pathId;
   mppLnx_ProxyRequestEntry_t       *currentCmndList;
   int                              ioCount;
   int                              longestTrip;
   int                              shortestTrip;
   int                              ioFails;
   int                              numPathFailures;
}  mppLnx_PDeviceEntry_t;




typedef struct mppLnx_PtargetEntry
{
   struct list_head              list;
   int                           targetId;
   mppLnx_PDeviceEntry_t         ** pdevices;
} 	mppLnx_PtargetEntry_t;



typedef struct mppLnx_VDeviceEntry
{
   struct   list_head         list;
   mppLnx_ProcLeafNode_t      procNode;
   Scsi_Device                *vdevice;
   RdacDeviceInformation_t    *rdacInfo;
   int                        lun;
   int                        ioCount;
   int                        ioFails;
   int                        longestTrip;
   int                        shortestTrip;
   int                        numPathFailures;
   int                        numCntlFailures;
   unsigned long              cntlFailoverStartTime;
   unsigned long              pathFailoverStartTime;
   unsigned long              longestCntlFailoverTime;
   unsigned long              shortestCntlFailoverTime;
   unsigned long              longestPathFailoverTime;
   unsigned long              shortestPathFailoverTime;
}  mppLnx_VDeviceEntry_t;


typedef struct mppLnx_PhostEntry
{
   struct list_head           list;
   int                        hostNo;
   struct Scsi_Host           *host;
   mppLnx_PtargetEntry_t      *targetList;
} 	mppLnx_PhostEntry_t;



typedef struct mppLnx_VHBAContext
{
   spinlock_t                 *vhbaContextLock;
   mppLnx_PhostEntry_t        *phostList;
   mppLnx_VDeviceEntry_t      *virtualDeviceList;
   Scsi_Host_Template         *vhbaSHTP;
} 	mppLnx_VHBAContext_t;

/**
* The physical HBA driver's saved error handlers
*/
typedef struct mppLnx_saved_PHBA_eh_handler
{
   int (*eh_device_reset_handler)(Scsi_Cmnd *);
   int (*eh_bus_reset_handler)(Scsi_Cmnd *);
   int (*eh_host_reset_handler)(Scsi_Cmnd *);
} mppLnx_saved_PHBA_eh_handlers_t;

/*
*
*/
typedef struct mppLnx_CmndEntry
{
   struct         list_head list;
   int            taksType;
   void           (*vcmnd_timeoutfunc)(Scsi_Cmnd *);
   Scsi_Cmnd      *vcmnd;
} mppLnx_CmndEntry_t;

#if USE_THREAD_FOR_DPC
/*
* Use a kernel thread to dispatch proxy commands
*/
typedef struct mppLnx_dpcThreadQueue
{
   struct semaphore           *dpc_sem;
   struct semaphore           *detect_sem;
   struct task_struct         *dpcQueue_process;
   mppLnx_CmndEntry_t         *cmndEntry;
   spinlock_t                 *cmndQueueLock;
   atomic_t                   needSchedule; /* 1 yes 0 not*/
} mppLnx_dpcQueueContext_t;


#else

/**
* Use the tasklet to dispatch proxy commands
*/
typedef struct mppLnx_taskletQueue
{
   mppLnx_CmndEntry_t         *cmndEntry;
   spinlock_t                 *cmndQueueLock;
   atomic_t                   needSchedule; /* 1 yes 0 not*/
   struct tasklet_struct      *mppTasklet;
} mppLnx_taskletQueue_t;

#endif


/**
* data struct for failback rescan thread
*/
typedef struct mppLnx_failbackScanContext
{
   struct semaphore       *failback_sem;
   struct semaphore       *detect_sem;      /*for the thread start
                                             *and termination*/
   struct task_struct     *failback_process;
   atomic_t               scan_active;
   struct timer_list      fbs_timer;
} mppLnx_failbackScanContext_t;

/*
* data struct for sync IO queue
*/
typedef struct mppLnx_workerQueueContext
{
   struct semaphore       *workerQueue_sem;
   struct semaphore       *detect_sem;      /*for the thread start
                                             *and termination*/
   struct task_struct     *workerQueue_process;
   atomic_t               worker_active;
   spinlock_t             *cmndQueueLock;
   mppLnx_CmndEntry_t     *cmndEntry;
} mppLnx_workerQueueContext_t;



int mppLnx_proc_info(char * buffer,char ** start,off_t offset, int length,int hostno,int writeto1_read0);
int mppLnx_detect(Scsi_Host_Template *shtp);
int mppLnx_revoke(Scsi_Device * sdp);
int mppLnx_release(struct Scsi_Host * shp);
const char * mppLnx_info(struct Scsi_Host * shp);
int mppLnx_queuecommand(Scsi_Cmnd * scp, void (*done)(Scsi_Cmnd *));
int mppLnx_eh_device_reset_handler(Scsi_Cmnd * scp);
int mppLnx_eh_bus_reset_handler(Scsi_Cmnd * scp);
int mppLnx_eh_host_reset_handler(Scsi_Cmnd * scp);
int mppLnx_bios_param(Disk * sdev, kdev_t dev, int *params);
void mppLnx_select_queue_depths(struct Scsi_Host * shp, Scsi_Device * sdp);
int mppLnx_createVDeviceProcNode(mppLnx_VDeviceEntry_t * vde);
int mppLnx_initVHBAContext(void);
int mppLnx_createPDeviceProcNode(mppLnx_PDeviceEntry_t * pde);
void mppLnx_removeVDeviceProcNode(mppLnx_VDeviceEntry_t * vde);
void mppLnx_removePDeviceProcNode(mppLnx_PDeviceEntry_t * pde);
int mppLnx_PDevice_read_proc(char *, char ** , off_t , int , int *, void *);
void mppLnx_cleanProcFiles(mppLnx_VHBAContext_t * hbaContext);
void mppLnx_newDeviceAdded(Scsi_Device * sdp);
int mppLnx_do_queuecommand(Scsi_Cmnd  * vscp, 
                                  void (*timeoutfunc)(Scsi_Cmnd *),
                                  void (*ml_done)(Scsi_Cmnd *),
                                  BYTE  previous_controller,
                                  BYTE  previous_path,
                                  BYTE CommandTimeoutRetryCount,
                                  BYTE RetryCount,
                                  BYTE SelectionTimeoutRetryCount,
                                  BYTE UARetryCount,
                                  MPP_TIME vcmndStartTime,
                                  MPP_TIME ControllerIoWaitTime,
                                  MPP_TIME ArrayIoWaitTime);
void mppLnx_IssueFailover(RdacDeviceInformation_t *rdacInfo, BYTE controller, 
		BYTE Path, MPP_HANDLE FailoverHandle, BYTE *Page2C, BYTE *cdb);
void mppLnx_failoverCmd_done(Scsi_Cmnd * scp);
void mppLnx_scsi_done(Scsi_Cmnd * scp);
int mppLnx_PHBA_eh_device_reset_handler(Scsi_Cmnd * scp);
int mppLnx_PHBA_eh_bus_reset_handler(Scsi_Cmnd * scp);
int mppLnx_PHBA_eh_host_reset_handler(Scsi_Cmnd * scp);
int mppLnx_failback_handler(void * data);
void mppLnx_rescan_timer_handler(unsigned long data);
void mppLnx_failbackRescan(void);
mppLnx_VDeviceEntry_t * mppLnx_get_VdeviceEntry(Scsi_Device *vdevice);
mppLnx_VDeviceEntry_t * mppLnx_insert_VdeviceEntry(int, RdacDeviceInformation_t *);
mppLnx_VDeviceEntry_t * mppLnx_createVDeviceEntry(int, RdacDeviceInformation_t *);
mppLnx_PDeviceEntry_t * mppLnx_get_PdeviceEntry(Scsi_Device *pdevice);
void mppLnx_insert_PdeviceEntry(Scsi_Device * pdevice, 
		RdacDeviceInformation_t *rdacInfo, int , int );
mppLnx_PhostEntry_t * mppLnx_createPhostEntry(Scsi_Device * pdevice);
mppLnx_PtargetEntry_t * mppLnx_createPtargetEntry(Scsi_Device *);
void mppLnx_insert_proxyRequest(mppLnx_ProxyRequestEntry_t *request);
mppLnx_ProxyRequestEntry_t * mppLnx_get_virtualRequest(Scsi_Cmnd *proxyCmnd);
#if USE_THREAD_FOR_DPC
int mppLnx_dpc_handler(void * data);
#else
void mppLnx_tasklet_func(unsigned long taskletQueue);
#endif
void mppLnx_Scsi2toScsi3_handler( Scsi_Cmnd * scp);
int mppLnx_worker_handler(void * data);
void mppLnx_schedule_resp( Scsi_Cmnd * scp);
int  mppLnx_delete_VdeviceEntry(mppLnx_VDeviceEntry_t *);
int  mppLnx_delete_PdeviceEntry(Scsi_Device *);
int  mppLnx_cleanVHBAContext(void);
mppLnx_ProxyRequestEntry_t * mppLnx_createProxyRequest(
	  Scsi_Cmnd * vscp,
	  Scsi_Cmnd *proxyCmnd);
void mppLnx_update_VdeviceStat(
       mppLnx_VDeviceEntry_t * vde, 
       int ioStatus, 
       unsigned long roundTrip);
void mppLnx_update_PdeviceStat(
        mppLnx_PDeviceEntry_t *pde,
        int ioStatus, 
        unsigned long startTime);

Scsi_Device * mppLnx_getAnyPathByVDev(Scsi_Device *);

int mppLnx_VDevice_read_proc
(
      char *page, 
      char ** start, 
      off_t offset, 
      int count, 
      int *eof, 
      void *data);


void mppLnx_BytestoAscii(char *ascii, char *buff, int bufflen);

void mppLnx_schedule_queuecmd( Scsi_Cmnd * scp);
void mppLnx_regSymbols(void);
void mppLnx_unregSymbols(void);
void mppLnx_setCheckCondition(Scsi_Cmnd * vcmd, SenseData_t *sensedata);

extern void mpp_ConvertWWNtoAscii(BYTE *AsciiWWN, BYTE *ArrayWWN);
extern BYTE mpp_FailController(RdacDeviceInformation_t * rdacInfo,
                               BYTE controller,
                               BYTE lun);

#ifdef MPP_DEBUG
void mppLnx_dump_Scsi_Host_Template(Scsi_Host_Template *sht);
void mppLnx_dump_Scsi_Host(struct Scsi_Host *shp);
void mppLnx_dumpCmnd(Scsi_Cmnd * SCnp);
void mppLnx_dump_proxyRequestList(mppLnx_ProxyRequestEntry_t * requestList);
void mppLnx_dump_PDeviceEntry(mppLnx_PDeviceEntry_t * pde);
#endif

/**
* The virtual HBA driver template
*/
#define MPPLNX_SCSI_HOST_TEMPLATE {                             \
       next:   NULL,                                            \
       module: NULL,                                            \
       proc_dir: NULL,                                          \
       proc_info: mppLnx_proc_info,                             \
       name: MPPLNX_PROC_NAME,                                             \
       detect: mppLnx_detect,                                   \
       revoke: NULL,                                            \
       release: mppLnx_release,                                 \
       info: mppLnx_info,                                       \
       ioctl: NULL,                                             \
       command: NULL,                                           \
       queuecommand: mppLnx_queuecommand,                       \
       eh_strategy_handler: NULL,                               \
       eh_abort_handler: NULL,                                  \
       eh_device_reset_handler: mppLnx_eh_device_reset_handler, \
       eh_bus_reset_handler: mppLnx_eh_bus_reset_handler,       \
       eh_host_reset_handler: mppLnx_eh_host_reset_handler,     \
       abort: NULL,                                             \
       reset: NULL,                                             \
       slave_attach: NULL,                                      \
       bios_param: mppLnx_bios_param,                           \
       select_queue_depths: mppLnx_select_queue_depths,        \
       can_queue: -1,        /* max simultaneous cmds   */      \
       this_id: -1,          /* scsi id of host adapter   */    \
       sg_tablesize: -1,     /* max scatter-gather cmds */      \
       cmd_per_lun: -1,      /* cmds per lun (linked cmds) */   \
       present: 0,           /* number of 7xxx's present   */   \
       unchecked_isa_dma: 0, /* no memory DMA restrictions */   \
       use_clustering: -1,                                      \
       use_new_eh_code: 1,                                      \
	   /*\
       emulated: 0,                                            \
       highmem_io: 1,                                           \
       can_do_varyio: 1,                                        \
	   */ \
       proc_name: MPPLNX_PROC_NAME                              \
}

#endif
