/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************

NAME            mppUtilSysdep.c
SUMMARY         %description%
VERSION         %version: 10 %
UPDATE DATE     %date_modified: Thu Dec 18 13:59:11 2003 %
PROGRAMMER      %created_by:    yqi %

        Copyright 2003, LSI Logic Storage Systems, Inc.

DESCRIPTION:

INCLUDE FILES:

NOTES:

RESTRICTIONS:

SEE ALSO:

REFERENCE:     Linux MPP Driver utility OLBS - 349-1034730

IMPLEMENTATION:

MODIFICATION HISTORY:

*******************************************************************************/

/***  INCLUDES  ***/  
#include        "mppUtil_Sysdep.h"
#include        "MPP_Sysdep.h"
#include        "MPP_Common.h"
#include        "MPP_RdacInfo.h"
#include        "mppUtil_Common.h"
#include        "mppLnx_version.h"
#include        "mppLnx_rls_version.h"
#include        "mppLnx_bld_version.h"

/***  VARIABLE DEFINITIONS  ***/

extern  HANDLE   utilVirtualFd;
extern  LWORD    utilMaxPaths;
extern  LWORD    utilMaxLuns;
extern  char	 *progname;

/***  TYPE DEFINITIONS  ***/
#define MPP_BUS_NODE "/proc/mpp/mppVBusNode"


/***  LOCALS  ***/


/***  PROCEDURES  ***/
static void mppLnxUtil_reportSrcAttributes();

/*******************************************************************************
* PROCEDURE
*
* NAME:      utilDetermineBusFd
* SUMMARY:   open the MPP "control node" 
* SCOPE:     private 
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION: Open the MPP "control node" and store the value of the open file
*   descriptor in the global variable utilVirtualFd. 
*
* RESTRICTIONS:
* 
* RETURNS: 
*   TRUE    - A valid file descriptor can be determined.
*   FALSE   - A valid file descriptor cannot be determined.

* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/

BOOL
utilDetermineBusFd ( void )
{
	int mppFd = 0;
	
	if( ( mppFd = open(MPP_BUS_NODE, O_RDWR ) ) >= 0 )
	{
		utilVirtualFd = mppFd;
		if(ioctl(utilVirtualFd, MPP_IOCTL_GETPATHS, &utilMaxPaths) != INVALID)
		{

			utilDebug("Found mpp virtual bus\n");
			(void )ioctl(utilVirtualFd, MPP_IOCTL_GETLUNS, &utilMaxLuns);
			utilDebug("maxPaths = %d; maxLuns = %d\n", utilMaxPaths, utilMaxLuns);
			return(TRUE);
		}
		else
		{
			return(FALSE);
		}
	}
	else
	{
		return(FALSE);
	}
    
}

/***************************************************************************
* PROCEDURE:    utilFormatHandle
* SUMMARY:      utilFormatHandle - supply a string that is an appropriate
*               representation of an MPP handle.
*
* DESCRIPTION:
*   The function receives an MPP-HANDLE as input and returns a displayable
*   string representation of the handle as is appropriate.  On Linux platform
*   it is not applicable. It simply formats the string to "none". 
*
* NOTES:
*
*
* INPUT:
*   handle -
*
* OUTPUT:
*   buffer - representation of an MPP Handle
*
* RETURNS:
*   None
*
****************************************************************************/
VOID 
utilFormatHandle (MPP_HANDLE handle, MPP_STRING buffer)
{

	sprintf( buffer, "%s", "none");
}

/***************************************************************************
* PROCEDURE:    utilFormatVertex
* SUMMARY:      utilFormatVertex - supply a string that is an appropriate
*               representation of an MPP vertex.
*
* DESCRIPTION:
*   The function receives an MPP-HANDLE representing a "vertex" as input
*   and returns a displayable string representation of the handle. Linux 
*   does not have vertices as in SGI IRIX, so this routine can just return 
*   the string "none".
*
* NOTES:
*
*
* INPUT:
*   handle - representing the "vertex"
*
* OUTPUT:
*   buffer - string of "none"
*
* RETURNS:
*   None
*
****************************************************************************/
VOID 
utilFormatVertex (MPP_HANDLE handle, MPP_STRING buffer)
{
       if (handle)
               sprintf(buffer, "%s", "present");
       else
               sprintf(buffer, "%s", "none");

        return;
}


/***************************************************************************
* PROCEDURE:    utilFormatControllerAddress
* SUMMARY:      utilFormatControllerAdderess - supply a string that is an appropriate
*               representation of a ControllerAdderess.
*
* DESCRIPTION:
*   This function is invoked from utilPrintRdacInfo() . It is a part of the
*   system dependent code. It receives am argument of type controllerAddress_t
*   and a buffer and returns a displayable string representation. 
*   On Linux, controllerAddress_t has the following structure:
*    struct controllerAddress_t {
*       Int hostId;
*	Int targetId; 
*	Int channel;
*    } Address;
* NOTES:
*
*
* INPUT:
*   handle - representing the "vertex"
* OUTPUT:
*   buffer - string of "none"
*
* RETURNS:
*   None
*
*
****************************************************************************/
void 
utilFormatControllerAddress(ControllerAddress_t address, MPP_STRING buffer )
{
	sprintf( buffer, "hostId: %d, targetId: %d, channelId: %d\n", address.HostId, address.TargetId, address.ChannelId);
}

/***************************************************************************
* PROCEDURE:    utilFormatDevState
* SUMMARY:      - supply a string that is an appropriate representation of Lun State.
*
* DESCRIPTION:
*
* NOTES:
*
*
* INPUT:
*  mpp_device_state 
* OUTPUT:
*
* RETURNS:
*   string of ""
*
*
****************************************************************************/
char *utilFormatDevState (MPP_DEVICE_STATE PseudoDevState) {
	return "";

}




/***************************************************************************
* PROCEDURE:    utilGetPhysicalHostMaxTargetId
* SUMMARY:      utilGetPhysicalHostMaxTargetId - 
*
* DESCRIPTION: This function takes a hostid and sends an ioctl to the 
*  Upper level mpp driver that will cause the driver to go through the 
*  list of hosts maintained by the scsi middle level and return the 
*  max number of TargetIds supported by the host.
*
* NOTES:
*
*
* INPUT:
*     The HostId  of the host bus adaptor driver we are interested in
*
* OUTPUT:
*     Prints the Max number of Targets handled by the host bus adaptor driver 
*
* RETURNS:
*     True or False depending on the outcome of the ioctl
*
****************************************************************************/
BOOL
utilGetPhysicalHostMaxTargetId(LINT hostid)
{
	struct host_target_t host_target_id;

	if(!utilDetermineBusFd())
	{
		fprintf(stderr, "Unable to determine virtual bus!!\n");
		return(INVALID);
	}

	host_target_id.host_id = hostid;
	utilDebug("Send physical host max target parameter request\n");
	if(ioctl(utilVirtualFd, MPP_IOCTL_GET_PHYSICALHOST_MAX_TARGET, &host_target_id ) == INVALID)
	{
		fprintf(stderr, "Unable to send ioctl request - %s!\n", strerror(errno));
		return(FALSE);
	}

	printf("MaxTargetIdPerHostId = %d\n", host_target_id.max_target_id);
	return(TRUE);

}
/***********************************************************************
* PROCEDURE:    mppLuxUtil_reportSrcAttributes
* SUMMARY:      Report mpp driver source file attributes at the source release
*               time and driver build time 
*
*	DESCRIPTION: This function performs the following two operations:
*  1.	Prints out each file's versions and digests at the source package release time and
*     the driver build time
*  2.	Compare an attribute of its release time with its build time. Report true or false 
*     of the comparison result
* NOTES:
*
* INPUT: N/A
*     
* OUTPUT:
*     Reports file attributes to the standard output
* RETURNS:
*     Void
*
***********************************************************************/

void mppLnxUtil_reportSrcAttributes()
{
   int numOfModifiedFiles = 0;
   int i;
   int isSame = 0;
   printf("Linux mpp driver version: %s\n",MPPLNX_VERSION);
   printf("Linux mpp driver release time stamp: %s\n\n", MPPLNX_BUILDTIMESTAMP);
   printf("File Name            Version  Digest                              Changed\n");

   for( i = 0; i < MPPLNX_NUM_FILE_ENTRY; i++)
   {
      isSame = 1;
      if( strcmp(mppLnx_rls_version_attributes[i].version,
                 mppLnx_bld_version_attributes[i].version) != 0)
      {
         numOfModifiedFiles++;
         isSame = 0;
      }
      if( strcmp(mppLnx_rls_version_attributes[i].digest,
                 mppLnx_bld_version_attributes[i].digest) != 0)
      {
         if(isSame == 1)
         {  
            numOfModifiedFiles++;
         }
         isSame = 0;
      }

      printf("%-20.20s rls %-4.4s %-35.35s\n",
         mppLnx_rls_version_attributes[i].fileName,
         mppLnx_rls_version_attributes[i].version,
         mppLnx_rls_version_attributes[i].digest);
      printf("%-20.20s bld %-4.4s %-35.35s %s\n",
         " ",
         mppLnx_bld_version_attributes[i].version,
         mppLnx_bld_version_attributes[i].digest,
         isSame == 1?"No":"Yes");
      printf("\n\n");
   }

   if(numOfModifiedFiles == 0)
   {
      printf("No source files have been changed.\n");
   }
   else
   {
      printf("%d source file(s) have been modified.\n",numOfModifiedFiles);
   }

}
/***************************************************************************
* PROCEDURE:    mppSys_ProcessCmdLineArgs
* SUMMARY:      Process platform specific mppUtil command line options
*
	DESCRIPTION: The design of the mppUtil allows each platform to specify platform
   specific command line options. This gives a platform MPP driver freedom to add command
   line options that are only applicable to the specific platform. This function is 
   the entry pointer for processing platform specific function.
*
* NOTES:
	This function is invoked by the main() function of the mppUtil. The main() function
   of the mppUtil handles all common command line options. When this function is invoked, 
   the command line options should be the options that are specific to the Linux platform.
	A platform specific option should be in capital letter
*
* INPUT:
	@option - an option's ASCII value
	@argument the operand of the option. Null if the option has no argument
* 
*
* RETURNS:
*     True or False depending on the outcome of the operation
*
****************************************************************************/
BOOL
mppSys_ProcessCmdLineArgs(LINT option, MPP_STRING argument)
{
	BOOL status = FALSE;

	switch( option )
	{
		case 'I':
			if (argument ) {
				status = utilGetPhysicalHostMaxTargetId(atoi(argument));
				return(status);
			}
      case 'V':
         printf("Linux mpp driver version: %s\n", MPPLNX_VERSION);
         status = 0;
         break;
      case 'C':
         mppLnxUtil_reportSrcAttributes();
         status = 0;
         break;

		default:
			utilUsage(progname);
			return(INVALID);
	}	

	return(status);

}
